# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

"""
Flexible, multi-property feature model informed by the OGC GML simple features
profile specification

  http://portal.opengeospatial.org/files/?artifact_id=11266

which is Copyright (C) 2005 Open Geospatial Consortium, Inc. All Rights
Reserved.
"""

from zope.interface import implements

from cartography.data.interfaces import IPropertyType, IGeometryPropertyType
from cartography.data.interfaces import IFeature
from cartography.geometry.interfaces import IOGCSFGeometry    


# ===========================================================================
# Feature properties
# ===========================================================================

class FeatureProperty:
    """Feature property."""

    implements(IPropertyType)
    label = None
    
    def __init__(self, name):
        """Initialize."""
        self.name = name
        
    def totuple(self):
        """Return as a tuple."""
        return (self.name, self.label)
        
        
class IntegerProperty(FeatureProperty):
    """Integer number."""
    label = 'int'

class RealProperty(FeatureProperty):
    """Real number."""
    label = 'float'

class StringProperty(FeatureProperty):
    """String."""
    label = 'string'

class DateProperty(FeatureProperty):
    """Date and time."""
    label = 'datetime'
        
class BooleanProperty(FeatureProperty):
    """True or False."""
    label = 'bool'

class BinaryProperty(FeatureProperty):
    """Binary data string."""
    label = 'binary'

class ArrayProperty(FeatureProperty):
    """Sequence of something"""
    label = 'array'

class GeometryProperty(FeatureProperty):
    """A geometry.
    
    Attributes
    ----------
    name : string
        Property name
    label : string
        Label
    homotype : string
        Can be used to indicate to some applications (like MapServer) that all
        instances of this property are of one homogeneous geometry type, such
        as "Point".
    homosrs : SpatialReference
        As above, to indicate that all geometries are of a single reference
        system.
    """

    implements(IGeometryPropertyType)
    label = 'geometry'

    def __init__(self, name, homotype=None, homosrs=None):
        """Initialize."""
        self.name = name
        self.homotype = homotype
        self.homosrs = homosrs

    def _get_defaultsrs(self):
        # Return the default srs, or None
        return self.homosrs
    defaultsrs = property(_get_defaultsrs)

    
# Utility function
def isgeometry(p):
    if IGeometryPropertyType.providedBy(p):
        return True
    else:
        return False
    

class PropertyTypeCollection(object):

    """Collection of property types.

    Our schemas are instances of this class.
    """

    def __init__(self, propertytypes=[]):
        """Initialize.
        
        Parameters
        ----------
        propertytypes : iterable
            A sequence of ProperyTypes
        """
        self._names = [p.name for p in propertytypes]
        self._types = dict([(p.name, p) for p in propertytypes])

    def __getitem__(self, key):
        # Get a property by name
        return self._types[key]

    def __iter__(self):
        # Return iterator over all property types
        return iter([self._types[n] for n in self._names])
       
    def names(self):
        """Return a list of property names."""
        return self._names
        

# ===========================================================================
# Features
# ===========================================================================

class PropertyCollection(dict):
    
    """Collection of feature properties.
    """
    
    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value


class Feature(object):

    """A multi-property feature.
    """
    implements(IFeature)

    def __init__(self, id, bounds=None, pdict={}, **kw):
        """Initialize a feature.
        
        Parameters
        ----------
        id : str
            a unique identifier for this feature
        bounds : BoundingBox
            Bounding box for the feature. If None, will be determined lazily
            from the default geometry.
        pdict, kw : mapping
            feature properties
        """
        self.id = str(id)
        self._bounds = bounds
        skeys = self.__schema__.names()
        self.properties = PropertyCollection()
        for k, v in pdict.items() + kw.items():
            if k not in skeys:
                raise AttributeError, \
                "Feature type '%s' (%s) has no '%s' property" \
                % (self.typename(), str(self.__class__), k)
            self.properties[k] = v

    def typename(cls):
        """Return the typename of this feature class."""
        return cls.__typename__
    typename = classmethod(typename)
    
    def schema(cls):
        """Return the schema of this feature class as a dict."""
        return cls.__schema__
    schema = classmethod(schema)

    def schema_info(cls):
        """Return an information dictionary about the schema."""
        return dict([x.totuple() for x in cls.schema()])
    schema_info = classmethod(schema_info)
    
    def defaultgeometry(cls):
        """Return the default geometry property."""
        schema = cls.__schema__
        # If featuretype has a __g__ attribute, this is the default
        if hasattr(cls, '__g__'):
            for s in schema:
                if s.name == cls.__g__:
                    return s
        else:
            # Pick out the *first* geometry property, this is our default
            for s in schema:
                if isgeometry(s):
                    return s
        # Fall through
        return None
    defaultgeometry = classmethod(defaultgeometry)

    def defaultsrs(cls):
        """Return the default spatial reference, if defined"""
        geom = cls.defaultgeometry()
        if geom:
            if geom.homosrs: return geom.homosrs
        if hasattr(cls, '__srs__'): return cls.__srs__
        else: return None
    defaultsrs = classmethod(defaultsrs)

    def _getBounds(self):
        # Get feature bounds, possibly calculating and setting as a side effect
        if self._bounds:
            return self._bounds
        else:
            g = self.properties[self.defaultgeometry().name]
            e = g.envelope()
            self._setBounds(e)
            return self._bounds

    def _setBounds(self, box):
        self._bounds = box
    bounds = property(_getBounds, _setBounds)

    def info(self):
        """Return an informational dict."""
        prop_items = []
        for k, p in self.properties.items():
            if IOGCSFGeometry.providedBy(p):
                # XXX: for now we skip geometries
                pass
            else:
                prop_items.append((k, p))
        b = self.bounds
        return {'id': self.id, 'bounds': [b.minx, b.miny, b.maxx, b.maxy],
                'srs': str(b.srs), 'properties': dict(prop_items)}

