# =============================================================================
# Python Cartographic Library. Copyright (C) 2005 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
#
# $Id: interfaces.py 432 2006-05-10 15:30:53Z seang $
# =============================================================================

"""
PCL-Data is informed by the OGC GML simple features profile specification

  http://portal.opengeospatial.org/files/?artifact_id=11266

which is Copyright (C) 2005 Open Geospatial Consortium, Inc. All Rights
Reserved.
"""

from zope.interface import Interface, Attribute

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
except ImportError:
    pass


class IPropertyType(Interface):
    """A simple feature property."""
    
    name = Attribute("name of the property")
    label = Attribute("string label of the property type")
    
    def totuple():
        """Return as a tuple.

        For example: IntProperty('gid') -> ('gid', 'int')
        """


class IGeometryPropertyType(IPropertyType):
    """A simple feature geometry property."""

    homotype = Attribute("Optional string such as 'Point' used to indicate "
                         "that geometries are of a single type")
    homosrs = Attribute("Optional SpatialReference used to indicate that "
                        "geometries are of a single reference system")
    defaultsrs = Attribute("Default spatial reference system, should be None "
                           "in the case of mixed or heterogeneous features")


class IFeatureType(Interface):
    """A GIS feature type."""

    def schema_info():
        """Return information about the featuretype schema as a dict.
        
        Each dict item corresponds to a featuretype property. The key is the
        property name, the value is the string name of the property type. For
        example, a very generic feature type schema:

        {'name': 'string', 'category': 'int', 'location': 'Point'}
        """

    def schema():
        """Return feature type schema."""
        
    def typename():
        """Return feature type name."""
        
    def defaultgeometry():
        """Return default geometry property."""
        
    def defaultsrs():
        """Return default spatial reference system.
        
        Should be None in the case of mixed or heterogeneous features.
        """


class IFeature(Interface):
    """A simple GIS feature.
    
    In the manner of the OGC GML simple features profile, IFeature is a simple
    and flat container for properties, any number of which may be geometries.
    """
    
    id = Attribute("Unique identifier for the feature (string)")
    properties = Attribute("Collection of properties, may include geometries "
                           "(dict)")
    bounds = Attribute("Bounding box of the feature (IBoundingBox)")

    def info():
        """Return an information dict."""


class IFilteringFeatureIterator(Interface):
    """An iterable collection of features, with filtering capability."""

    def next():
        """Return the next feature"""
        
        
class IFeatureSource(Interface):
    """A cursor-like source of features of a type. 
    """

    featuretype = Attribute("The source's feature type")

    def iterfeatures(filter):
        """Returns IFilteringFeatureIterator"""

    def __call__(filter):
        """Returns IFilteringFeatureIterator"""


class IDataStore(Interface):
    """Feature or raster data store."""

    def typenames():
        """Return a list of feature typename strings"""

    def bounds(typename):
        """Return bounding box of all features of the type name"""
            
    def info(typename):
        """Returns an information dict."""


class IFeatureStore(IDataStore):
    """One or more types of features, on any of which we may tap a source."""

    def featuretype(typename):
        """Return the feature type"""

    def source(typename, bbox, properties, maxfeatures):
        """Tap a source on a feature type by name, with an optional query.
        Returns an object that implements IFeatureSource.
        
        Parameters
        ----------
        bbox : IBoundingBox
            Bounding box
        properties : string
            An SQL-like feature property filter
        maxfeatures : int
            Limit the number of features from the source
        """

    def count(typename):
        """Return feature count"""


class IOWSStore(Interface):
    """Marker interface for OGC Web Service stores."""


class IWFSStore(IOWSStore, IFeatureStore):
    """OGC Web Feature Service store."""


class IRasterStore(IDataStore):
    """Raster data store."""


class IWMSStore(IOWSStore, IRasterStore):
    """OGC Web Map Service store."""


