# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================
# Require PCL-Core
try:
    import pkg_resources
    try:	
        pkg_resources.require('PCL-Core')
    except pkg_resources.DistributionNotFound:
        pass
except ImportError:
    pass

import sys

from cartography.data.feature import Feature
from cartography.geometry import BoundingBox
from cartography.data.interfaces import IFeatureStore, IFeatureSource

from zope.interface import implements


class MemoryFeatureStore:

    """Implements IFeatureStore.
    """

    implements(IFeatureStore)

    def __init__(self, *args):
        self._featuretypes = {}
        self._features = {}
        self._bounds = {}
        for a in args:
            if issubclass(a, Feature):
                self._featuretypes[a.__typename__] = a
                self._features[a.__typename__] = []
                self._bounds[a.__typename__] = None

    def _expand_bounds(self, typename, feature):
        """Expands the bounds for the typename if the given
        feature is not contained within the current bounds.

        Parameters
        ----------
        typename : string
            Name of a feature type

        feature : object
            The new feature
        """
        fbounds = feature.bounds
        sbounds = self._bounds[typename]
        if sbounds is None:
            self._bounds[typename] = fbounds.copy()
        elif not sbounds.contains(fbounds):
            if fbounds.minx < sbounds.minx:
                minx = fbounds.minx
            else:
                minx = sbounds.minx
            if fbounds.miny < sbounds.miny:
                miny = fbounds.miny
            else:
                miny = sbounds.miny
            if fbounds.maxx > sbounds.maxx:
                maxx = fbounds.maxx
            else:
                maxx = sbounds.maxx
            if fbounds.maxy > sbounds.maxy:
                maxy = fbounds.maxy
            else:
                maxy = sbounds.maxy
            self._bounds[typename] = BoundingBox(minx, miny, maxx, maxy)
   
    def featuretype(self, typename):
        return self._featuretypes[typename]

    def source(self, typename, bbox=None, properties=None, maxfeatures=None):
        return MemoryFeatureSource(self, typename, bbox, maxfeatures)

    def count(self, typename):
        """Returns the feature count."""
        return len(self._features[typename])

    def typenames(self):
        """Returns the typenames available in this store."""
        return tuple(self._features.keys())

    def bounds(self, typename):
        """Returns a BoundingBox of all features for the given typename."""
        if self._bounds[typename] is None:
            return BoundingBox()
        else:
            return self._bounds[typename].copy()

    def info(self, typename):
        """Returns an information dictionary"""
        ft = self.featuretype(typename)
        p = ft.schema_info()
        n = self.count(typename)
        b = self.bounds(typename).totuple()
        srs = ft.defaultsrs()
        return {'typename': typename,
                'schema': p, 
                'srs': srs.tostring(),
                'bounds': b}

class MemoryFeatureSource:

    """Implements IFeatureSource.
    """
    implements(IFeatureSource)

    def __init__(self, store, typename, bbox, maxfeatures):
        self.store = store
        self.typename = typename
        self._featuretype = store._featuretypes[typename]
        self._bbox = bbox
        self._maxfeatures = maxfeatures or self.store.count(typename)

    def _get_featuretype(self):
        return self._featuretype
    featuretype = property(_get_featuretype)

    def schema(self):
        return self._featuretype.schema()
        
    def iterfeatures(self, query=None):
        i = 0
        N = self._maxfeatures
        candidates = self.store._features[self.typename]
        bbox = self._bbox
        if bbox:
            b1 = bbox.totuple()
            for feature in candidates:
                if i >= N:
                    raise StopIteration
                if bbox.intersects(feature.bounds):
                    i += 1
                    yield feature
        else:
            for feature in candidates:
                if i >= N:
                    raise StopIteration
                i += 1
                yield feature

    __call__ = iterfeatures

    def addfeature(self, feature):
        if isinstance(feature, self._featuretype):
            self.store._features[self.typename].append(feature)
            self.store._expand_bounds(self.typename, feature)


