# $Id: data.py,v 1.40 2005/05/19 19:46:29 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

"""
OWS -- WMS and WFS -- data stores.

Currently only WMS 1.1.0 and WFS 1.0.0 protocols are supported.
"""

import os

# Require PCL-Referencing
try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
    pkg_resources.require('OWSLib')
except:
    pass

from cartography.geometry import BoundingBox
from cartography.proj import SpatialReference
from owslib import wms
from owslib import wfs

#from cartography.interfaces import IOWSMapRequest
from cartography.data.interfaces import IFeatureStore, IWMSStore, IWFSStore

from zope.interface import implements

# ============================================================================
# Web Map Service
# ============================================================================

class WMSRasterStore:
    
    """An external web map service
    
    Instances are handed directly to the underlying mapping engine and so yet
    have no methods.
    """
    implements(IWMSStore)

    def __init__(self, url='', version='', incoming=None):
        """Initialize

        Parameters
        ----------
        url : string
            The service onlineresource url, the name is a layer name
        version : string
            WMS Version string, such as '1.1.0'
        incoming : string
            incoming folder for requested images
        """
        self.url = url
        self.version = version
        self._incoming = incoming
        self._wms = wms.WebMapService(self.url, self.version)
        
    def typenames(self):
        """Return a list of typenames."""
        return [layer.name for layer in self._wms.capabilities.contents]
     
    def bounds(self, name):
        """Returns bounds."""
        meta = self._wms.capabilities.getContentByName(name)
        if meta.boundingBox:
            b = meta.boundingBox
            return BoundingBox(b[0], b[1], b[2], b[3],
                    srs=SpatialReference(epsg=int(b[4].split(':')[1])))
        elif meta.boundingBoxWGS84:
            b = meta.boundingBoxWGS84
            return BoundingBox(b[0], b[1], b[2], b[3],
                    srs=SpatialReference(epsg=4326))
        else:
            return None
            
    def source(self, typename,  ows_srs=None, ows_format=None):
        return WMSRasterSource(self, typename, ows_srs, ows_format)
     
    def info(self, typename):
        """Source description"""
        meta = self._wms.capabilities.getContentByName(typename)
        return {'typename': typename, 'srs': meta.crsOptions}


class WMSRasterSource:
    """A layer within a web map service.
    """
    #implements(IOWSMapRequest)

    def __init__(self, store, typename, ows_srs=None, ows_format=None):
        """Initialize"""
        self.store = store
        self.typename = typename
        self.ows_srs = ows_srs
        self.ows_format = ows_format


class WFSFeatureStore:
    
    """Represents an external WFS layer.
    """
    implements(IWFSStore)

    def __init__(self, url='', version='', incoming=None):
        """Initialize

        Parameters
        ----------
        url : string
            The service onlineresource url, the name is a layer name
        version : string
            WFS Version string, such as '1.0.0'
        incoming : string
            incoming directory for requested features
        """
        self.url = url
        self.version = version
        self._incoming = incoming
        self._wfs = wfs.WebFeatureService(self.url, self.version)
        
    def typenames(self):
        """Return a list of typenames."""
        return [meta.name for meta in self._wfs.capabilities.contents]

    def bounds(self, name):
        """Returns bounds of a feature type. WFS bounds are always EPSG:4326.
        """
        meta = self._wfs.capabilities.getContentByName(name)
        b = meta.boundingBoxWGS84
        return BoundingBox(b[0], b[1], b[2], b[3], 
                srs=SpatialReference(epsg=4326))

    def source(self, typename, bbox=None, properties=None, maxfeatures=None,
               ows_srs=None, ows_format=None):
        """Returns a features source.
        """
        return WFSFeatureSource(self, typename, ows_srs, ows_format)
    
    def info(self, typename):
        """feature store description."""
        meta = self._wfs.capabilities.getContentByName(typename)
        return {'typename': meta.name, 'title': meta.title,
                'srs': meta.crsOptions}
        
    def count(self, typename):
        # WFS doesn't support this
        return -1

    def featuretype(self, typename):
        """Returns the feature type."""
        raise NotImplementedError


class WFSFeatureSource:
    #implements(IOWSMapRequest)

    def __init__(self, store, typename, ows_srs=None, ows_format=None):
        """Initialize"""
        self.store = store
        self.typename = typename
        self.ows_srs = ows_srs
        self.ows_format = ows_format

    def datatype(self):
        """Data type"""
        # for use with the mapserver rendering engine
        return 'Point'
        

