# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

"""
Implementation of IFeatureType, IFeatureSource, and IFeatureStore for PostGIS
version 1.x using psycopg.

Usage
-----

Here's an example using the well-known world_borders dataset loaded into a
PostgreSQL schema named 'postgis':

>>> store = PGFeatureStore("dbname=pcl_test")
>>> source = store.source("postgis.world_borders", properties="'fips_cntry = 'UK'")
>>> for feature in source():
...     # use the UK features

Here's another example using the uk_borders view (see the definition in 
PCL-Data/fixtures/makefile_postgis):

>>> source = store.source("postgis.uk_borders")
>>> for feature in source():
...     # use the features from the view, each is an instance of a 
...     # class derived from PGFeature

An example of a user-defined SQL selection as a feature type:

>>> class CustomFeature(postgis.PGSelectionFeature):
...     __typename__ = "custom"
...     __schema__ = PropertyTypeCollection([
...                        IntegerProperty('gid'),
...                        GeometryProperty('the_geom',homotype='MultiPolygon'),
...                        StringProperty('fips_cntry')
...                        ])
...     __sql_from__ = "world_borders"
...     __sql_select__ = "gid, the_geom, fips_cntry"
...     __sql_where__ = "fips_cntry = 'UK'"
...     __gid__ = "gid"
...     __srid__ = 4269
...
>>> self.store.addfeaturetype(CustomFeature)
>>> source = store.source("custom")
>>> for feature in source():
...     # use the features from the selection, each of which is an instance
...     # of CustomFeature
"""

import os
import re
import types

try:
    import psycopg
except ImportError:
    try:
        import psycopg2 as psycopg
    except ImportError:
        pass

# Require PCL-Core
try:
    import pkg_resources
    try:	
        pkg_resources.require('PCL-Core')
    except pkg_resources.DistributionNotFound:
        pass
except ImportError:
    pass


from cartography.proj import SpatialReference
from cartography.geometry import Geometry, BoundingBox

from cartography.data.feature import Feature, isgeometry, \
    PropertyTypeCollection
from cartography.data.feature import IntegerProperty, RealProperty, \
    StringProperty, GeometryProperty, ArrayProperty

from cartography.data.interfaces import IFeatureStore
from zope.interface import implements


# ============================================================================
# Utility functions
# ============================================================================

def _parse_typename(typename):
    # Split the typename string and return a (schema, object) tuple.
    # if no schema is specified, we assume 'public'
    name_parts = typename.split('.')
    nparts = len(name_parts)
    if nparts == 1:
        return ('public', name_parts[0])
    elif nparts == 2:
        return (name_parts[0], name_parts[1])
    else:
        raise ValueError, "Invalid typename '%s'" % (typename)
   
# ============================================================================
# Classes
# ============================================================================

class PGFeature(Feature):

    """A table or view based feature.

    Interesting behavior of a PGFeature class is that it can form SQL for use
    in querying a PostGIS backend about its instances.

    Attributes
    ----------
    __g__ : string
        Name of the default geometry property
    __srid__ : int
        SRID of the default geometry, and thereby features
    """

    # Override Feature.defaultsrs
    def defaultsrs(cls):
        """Return the default spatial reference, if defined"""
        geom = cls.defaultgeometry()
        if geom:
            if geom.homosrs: return geom.homosrs
        if hasattr(cls, '__srid__'):
            cls.__srs__ = SpatialReference(epsg=int(cls.__srid__))
            return cls.__srs__
        else: return None
    defaultsrs = classmethod(defaultsrs)

    def defaultsrid(cls):
        """Return the default SRID, if defined."""
        if hasattr(cls, '__srid__'):
            return int(cls.__srid__)
        else: return None
    defaultsrid = classmethod(defaultsrid)

    def estimated_extent_sql(cls):
        # Return SQL for fast, estimated extent
        # We have to use just a table name, not qualified name
        geoms = [s.name for s in cls.__schema__ if isgeometry(s)]
        selections = []
        items = cls.__typename__.split('.')
        if len(items) == 1:
            typename = items[0]
            for geom in geoms:
                s = "SELECT estimated_extent('%s', '%s'), area(estimated_extent('%s', '%s'))" % (typename, geom, typename, geom)
                selections.append(s)
        elif len(items) == 2:
            schema, typename = items
            for geom in geoms:
                s = "SELECT estimated_extent('%s', '%s', '%s'), area(estimated_extent('%s', '%s', '%s'))" % (schema, typename, geom, schema, typename, geom)
                selections.append(s)
        return ' UNION '.join(selections) + ' ORDER BY area DESC LIMIT 1'
    estimated_extent_sql = classmethod(estimated_extent_sql)
    
    def extent_sql(cls):
        # Return SQL for accurate extent
        typename = cls.__typename__
        geoms = [s.name for s in cls.__schema__ if isgeometry(s)]
        selections = []
        for geom in geoms:
            s = "SELECT extent(%s), area(extent(%s)) FROM %s" % (geom, geom, typename)
            selections.append(s)
        return ' UNION '.join(selections) + ' ORDER BY area DESC LIMIT 1'
    extent_sql = classmethod(extent_sql)

    def count_sql(cls):
        return "SELECT count(*) FROM %s" % (cls.__typename__)
    count_sql = classmethod(count_sql)

    def pkey_sql(cls):
        # Return primary key column 
        schema, objname = _parse_typename(cls.__typename__)
        return """
        SELECT column_name FROM information_schema.key_column_usage
        WHERE table_schema = '%s'
        AND constraint_name like '%%pkey'
        AND table_name = '%s' UNION
        SELECT column_name
        FROM information_schema.key_column_usage k,
             information_schema.view_table_usage v
        WHERE k.table_name = v.table_name
        AND view_schema = '%s'
        AND v.view_name = '%s'
        """ % (schema, objname, schema, objname)
    pkey_sql = classmethod(pkey_sql)
    
    def feature_sql(cls, pkey, bbox=None, srid=None, 
                    properties=None, maxfeatures=0, target_geometry=None):
        schema, objname = _parse_typename(cls.__typename__)
        geom_names = ['"%s"' % s.name for s in cls.__schema__ if isgeometry(s)]
        other_names = ['"%s"' % s.name 
            for s in cls.__schema__ if not isgeometry(s)]
        w_clause = ''
        if properties:
            w_clause = properties
        if bbox:
            w_clause += " %s && SetSRID('BOX3D(%f %f, %f %f)'::box3d, %d)" \
                     % (target_geometry or cls.defaultgeometry().name, 
                        bbox.minx, bbox.miny, bbox.maxx, bbox.maxy, srid)
        if w_clause:
            w_clause = 'WHERE ' + w_clause.strip()
        other_cols = ','.join(other_names)
        geom_cols = ','.join([g for g in geom_names])
        sql = 'SELECT %s, %s FROM %s %s' \
            % (geom_cols, other_cols, cls.__typename__, w_clause)
        if maxfeatures:
            sql += " LIMIT %d" % (int(maxfeatures))
        return sql
    feature_sql = classmethod(feature_sql)

    
class PGSelectionFeature(PGFeature):

    """An SQL selection based feature type.
    
    Interesting behavior of a PGFeature class is that it can form SQL for use
    in querying a PostGIS backend about its instances.

    Attributes
    ----------
    __typename__ : string
        A unique name for the feature store
    __schema__ : dict
        See Feature
    __sql_select__ : string
        SQL for field selection
    __sql_from__ : string
        SQL for table definition
    __sql_where__ : string
        SQL for where clause
    __sql_order__ : string
        Optional ORDER BY clause
    __gid__ : string
        field containing the unique identifier (must be part of the SQL)
    __g__ : string
        Name of the default geometry
    __srid__ : integer
        Spatial Reference ID
    """

    def estimated_extent_sql(cls):
        # Not implemented. Uses extent_sql()
        return cls.extent_sql()
    estimated_extent_sql = classmethod(estimated_extent_sql)
    
    def extent_sql(cls):
        # Return SQL for accurate extent
        typename = cls.__typename__
        geoms = [s.name for s in cls.__schema__ if isgeometry(s)]
        if hasattr(cls, '__sql_where__'):
            where = "WHERE %s" % (cls.__sql_where__)
        else:
            where = ""
        selections = []
        for geom in geoms:
            s = "SELECT extent(%s), area(extent(%s)) FROM %s %s" % (geom, geom, cls.__sql_from__, where)
            selections.append(s)
        return ' UNION '.join(selections) + ' ORDER BY area DESC LIMIT 1'
    extent_sql = classmethod(extent_sql)

    def count_sql(cls):
        if hasattr(cls, '__sql_where__'):
            where = "WHERE %s" % (cls.__sql_where__)
        else:
            where = ""
        return "SELECT count(*) FROM %s %s" % (cls.__sql_from__, where)
    count_sql = classmethod(count_sql)

    def pkey_sql(cls):
        return "SELECT '%s'" % (cls.__gid__)
    pkey_sql = classmethod(pkey_sql)

    def feature_sql(cls, pkey, bbox=None, srid=None,
                    properties=None, maxfeatures=0, target_geometry=None):
        geom_names = [s.name for s in cls.__schema__ if isgeometry(s)]
        where = ''
        if properties:
            where = properties
        if bbox:
            # find the element in __sql_select__ corresponding to the 
            # target_geometry
            target = target_geometry or cls.defaultgeometry().name
            for c in [c.strip() for c in cls.__sql_select__.split(',')]:
                if c.find(target) == 0:
                  where += " %s && SetSRID('BOX3D(%f %f, %f %f)'::box3d, %d)" \
                        % (c, bbox.minx, bbox.miny, bbox.maxx, bbox.maxy, srid)
        if hasattr(cls, '__sql_where__'):
            if where:
                where = ' AND '.join([where, cls.__sql_where__])
            else:
                where = cls.__sql_where__
        
        if where:
            where = 'WHERE ' + where.strip()

        if hasattr(cls, '__sql_order__'):
            order = cls.__sql_order__
        else:
            order = ""

        sql = "SELECT %s FROM %s %s %s" \
            % (cls.__sql_select__, cls.__sql_from__, where, order)
        if maxfeatures:
            sql += " LIMIT %d" % (int(maxfeatures))
        return sql
    feature_sql = classmethod(feature_sql)


class PGFeatureStore:
    """Interface to a PostGIS 1.x enabled PostgreSQL database.
    """
    
    implements(IFeatureStore)
    
    def __init__(self, dsn='', db_schemas=None, empty=False):
        """Initialize a new store.

        Parameters
        ----------
        dsn : string
            A PostgreSQL connection string like 'host=foo dbname=bar user=fred'
        db_schemas : list
            Restrict initial feature types to these specified schemas
        empty : boolean
            If True, intialize without any feature types
        """
        self.dsn = dsn
        self._featuretypes = {}
        if not empty:
            self._initialize_featuretypes()
        
    def _connect(self):
        # Connect to data source
        conxn = psycopg.connect(self.dsn)
        return conxn

    def _initialize_featuretypes(self, db_schemas=None):
        for dbname, schemaname, tablename, objname \
        in self._get_elligible_typenames(db_schemas):
            self._activate_relation(schemaname, tablename, objname)

    # Create feature type for a database relation (table or view)
    def _activate_relation(self, schemaname, tablename, objname):
        typename = '%s.%s' % (schemaname, objname)
        if typename in self._featuretypes.keys():
            raise Exception, "Featuretype '%s' is already active" % (typename)
        
        # Query to determine the schema of this type
        conxn = self._connect()
        cur = conxn.cursor()
        # Relation is a table
        if tablename == objname:
            cur.execute("""
            SELECT DISTINCT c.column_name, c.data_type, c.table_catalog, g.type
            FROM information_schema.columns c
            FULL OUTER JOIN geometry_columns g
            ON c.column_name = g.f_geometry_column
            AND c.table_name = g.f_table_name
            WHERE c.table_schema = '%s'
            AND c.table_name = '%s'
            """ % (schemaname, tablename))
            
        # Relation is a view
        else:
            cur.execute("""
            SELECT column_name, data_type, table_catalog, type
            FROM (
            SELECT c.column_name, c.data_type, c.table_catalog
            FROM information_schema.columns c,
                  information_schema.view_column_usage u
            WHERE c.column_name = u.column_name
            AND c.table_name = u.table_name
            AND u.view_schema = '%s'
            AND u.view_name = '%s'
            ) AS foo       
            FULL OUTER JOIN (
            SELECT type, f_geometry_column
            FROM information_schema.view_column_usage u, geometry_columns g
            WHERE u.table_name = g.f_table_name
            AND u.column_name = g.f_geometry_column
            AND u.view_schema = '%s'
            AND u.view_name = '%s'
            ) AS bar
            ON column_name = f_geometry_column
            """ % (schemaname, objname, schemaname, objname))

        schema = []
        dbname = ''
        while 1:
            row = cur.fetchone()
            if not row:
                break
            # Convert PostgreSQL types to PCL types. Let's find a better
            # way to do this ASAP!
            name, t = row[0:2]
            if t in ('integer', 'bigint'):
                si = IntegerProperty(name)
            elif t in ('double precision', 'numeric'):
                si = RealProperty(name)
            elif t in ('character varying', 'text'):
                si = StringProperty(name)
            elif t in ('boolean'): 
 	            si = BooleanProperty(name)
            elif t in ('timestamp', 'timestamp with time zone'): 
 	            si = DateProperty(name) 
            elif t.lower() in ('array'):
                si = ArrayProperty(name)
            elif row[3]:
                # Homogeneous SRS value will be provided below
                si = GeometryProperty(name, homotype=str(row[3]), 
                                      homosrs=None)

            schema.append(si)
            # Get database name from the query. A bit redundant, but not
            # terribly so
            dbname = row[2]
        
        # Determine the default geometry, this is the first one with an
        # index, or if there are no indexes, just the first one we encounter.
        first_geom = None
        first_srid = None
        default_geom = None
        default_srid = None
        for s in [s for s in schema if isgeometry(s)]:
            k = s.name
            cur.execute("""
            SELECT srid, indexdef
            FROM (
                SELECT srid, f_table_name, f_geometry_column 
                FROM geometry_columns g, information_schema.tables t
                WHERE g.f_table_name = t.table_name
                AND t.table_schema = '%s'
                AND g.f_table_name = '%s'
                AND g.f_geometry_column = '%s'
            ) AS foo
            LEFT OUTER JOIN pg_indexes x
            ON foo.f_table_name = x.tablename
            AND x.indexdef LIKE '%%gist (' || foo.f_geometry_column  || ')'
            UNION
            SELECT srid, indexdef
            FROM (
                SELECT srid, f_table_name, f_geometry_column
                FROM geometry_columns g, information_schema.view_table_usage v
                WHERE v.table_name = g.f_table_name
                AND v.view_schema = '%s'
                AND v.view_name = '%s'
                AND g.f_geometry_column = '%s'
            ) AS bar
            LEFT OUTER JOIN pg_indexes x
            ON bar.f_table_name = x.tablename
            AND x.indexdef LIKE '%%gist (' || bar.f_geometry_column  || ')'
            """ % (schemaname, objname, k, schemaname, objname, k))
            row = cur.fetchone()
            if not row:
                raise Exception, "Could not find SRID for geometry '%s'" % k
            s.homosrs = SpatialReference(epsg=int(row[0]))
            if row[1] and not default_geom:
                default_geom, default_srid = (k, int(row[0]))
            if first_geom is None:
                first_geom = k
                first_srid = int(row[0])

        if not default_geom:
            default_geom = first_geom
            default_srid = first_srid

        # Dynamically create a new class for this feature type
        ftype = type("%s_%s_%sFeature" % (dbname, schemaname, objname),
                     (PGFeature,),
                     dict(__schema__=PropertyTypeCollection(schema),
                          __typename__=typename,
                          __g__=default_geom, __srid__=default_srid),
                          )
        self._featuretypes[typename] = ftype

    # ----------------------------------------------------------------------
    # _get_elligible_typenames()
    #
    # Get list of all tables with geometry columns, and views that use these
    # tables, restricting results to specified database schemas. These
    # relations are elligible to be PG feature types.
    
    def _get_elligible_typenames(self, db_schemas=None):
        # Return list of feature type names
        conxn = self._connect()
        cur = conxn.cursor()
        if not db_schemas:
            cur.execute("""
            SELECT DISTINCT table_catalog, table_schema, table_name, table_name
            FROM information_schema.tables, geometry_columns
            WHERE table_name = f_table_name
            """)
        else:
            # Cat together quoted schema names like "'schema1','schema2',..."
            schemas_csv = ','.join(["'%s'" % (s) for s in db_schemas])
            cur.execute("""
            SELECT DISTINCT table_catalog, table_schema, table_name, table_name
            FROM information_schema.tables, geometry_columns
            WHERE table_name = f_table_name
            AND table_schema IN '(%s)'
            """ % schemas_csv)
        typenames = []
        while 1:
            row = cur.fetchone()
            if not row:
                break
            typenames.append(row)
        return typenames

    def activate(self, typename):
        """Activate an elligible feature type"""
        if typename in self._featuretypes.keys():
            raise KeyError, "Featuretype '%s' is already active" % (typename)
        schema, name = _parse_typename(typename)
        self._activate_relation(schema, name, name)
      
    def addfeaturetype(self, type):
        """Add a feature type"""
        self._featuretypes[type.typename()] = type

    def deactivate(self, typename):
        """Deactivate a feature type"""
        if typename not in self._featuretypes.keys():
            raise KeyError, "No such featuretype '%s'" % (typename)
        del self._featuretypes[typename]
        
    def typenames(self):
        """Return list of feature type names."""
        return self._featuretypes.keys()
        
    def featuretype(self, typename):
        """Return the feature type object by its name.
        
        Parameters
        ----------
        typename : string
            Name of a feature type
        """
        return self._featuretypes[typename]
        
    def source(self, typename, bbox=None, properties=None, maxfeatures=0,
               geometry=None):
        """Return a PGFeatureSource tapped into a feature type.
        
        Parameters
        ----------
        typename : string
            Name of a feature type
        """
        return PGFeatureSource(self, typename, bbox, properties, maxfeatures,
                               geometry)
        
    def count(self, typename):
        """Return number of features of a type in the store.
        
        Parameters
        ----------
        typename : string
            Name of a feature type
        """
        conxn = self._connect()
        cur = conxn.cursor()
        sql = self._featuretypes[typename].count_sql()
        cur.execute(sql)
        row = cur.fetchone()
        n = int(row[0])
        return n

    def bounds(self, typename, fastest=False):
        """Return bounds of all features of a type.
        
        Parameters
        ----------
        typename : string
            Name of a feature type
        fastest : boolean
            If True, use the PostGIS estimated_extent function to get the
            fastest, but less accurate, estimate of the bounds.
        """
        try_fastest = fastest
        conxn = self._connect()
        cur = conxn.cursor()
       
        ftype = self.featuretype(typename)
        if try_fastest == True:
            sql = ftype.estimated_extent_sql()
            cur.execute(sql)
            row = cur.fetchone()
            if not row or not row[0]:
                # try again, the slow way
                try_fastest = False
            else:
                e = row[0]
        if try_fastest == False:
            sql = ftype.extent_sql()
            cur.execute(sql)
            row = cur.fetchone()
            e = row[0]
            
        if not e:
            return None
        
        # Parse PostGIS response (WKT)
        start = e.find('(')
        e = e[start+1:-1]
        min, max = e.split(',')
        minx, miny = [float(v) for v in min.strip().split()]
        maxx, maxy = [float(v) for v in max.strip().split()]
        return BoundingBox(minx, miny, maxx, maxy)

    def info(self, typename):
        """Return an informational dict about a feature type.
                
        Parameters
        ----------
        typename : string
            Name of a feature type

        Returns a dict like 
        { 'typename': 'street', 
          'schema': {'name': 'string', 'type': 'int'},
          'srs': 'init=epsg:26913',
          'bounds': (400000, 4000000, 400400, 4000400) }
        """
        ft = self.featuretype(typename)
        p = ft.schema_info()
        n = self.count(typename)
        b = self.bounds(typename).totuple()
        srs = ft.defaultsrs()
        return {'typename': typename, 'schema': p, 
                'srs': srs.tostring(), 'bounds': b}


class PGFeatureSource:

    """A read-only source for features of a single type from a store.
    """
    
    def __init__(self, store, typename, bbox=None, properties=None, 
                 maxfeatures=0, geometry=None):
        """Initialize a Source.

        Parameters
        ----------
        store : FeatureStore
            A PGFeatureStore
        typename : string
            Name of one of the types in the store
        bbox : iterable
            List or tuple like [minx, miny, maxx, maxy]. Only features that
            intersect with this bounding box will issue from the source.
        properties : string
            An SQL-like WHERE clause in terms of feature properties such as
            "fname = 'foo' AND ftype = 2".
        maxfeatures : int
            Maximum number of features to return.
        geometry : string
            Name of the geometry which will be compared to the bbox.
        """
        self.store = store
        self.typename = typename
        self._bbox = bbox
        self._properties = properties
        self._maxfeatures = maxfeatures
        self._featuretype = store.featuretype(typename)
        self._geometry = geometry
        
        self.conxn = None
        if not self.conxn:
            self._connect()
        
        ftype = self._featuretype
        if hasattr(ftype, '__gid__'):
            self.pkey_column = ftype.__gid__
        else:
            cur = self.conxn.cursor()
            sql = ftype.pkey_sql()
            cur.execute(sql)
            self.pkey_column = cur.fetchone()[0]

    def _connect(self):
        # Connect to data source
        self.conxn = self.store._connect()

    def _get_featuretype(self):
        # Return the feature type object
        return self._featuretype
    featuretype = property(_get_featuretype)
      
    def iterfeatures(self, filter=None, **params):
        """Return iterator over features, applying optional front-end filter.
        
        Parameters
        ----------
        filter : string
            A Python expression
        params : mapping
            Environment in which to evaluate the filter expression

        Usage
        -----
        >>> for feature in a_source():
        >>>     # use feature     
        """
        if not self.conxn:
            self._connect()
        cur = self.conxn.cursor()
        sql = self.featuretype.feature_sql(self.pkey_column, self._bbox,
            self.featuretype.defaultsrid(), self._properties, self._maxfeatures,
            self._geometry)
        try:
            cur.execute(sql)
        except:
            print sql
            raise
        # Pass this cursor along with the front_query to the iterator
        return PGFeatureIterator(self, cur, filter, **params)

    __call__ = iterfeatures


class PGFeatureIterator:

    """Iterates over selected features from a source using a psycopg cursor.
    """
    
    def __init__(self, source, cursor, filter=None, **params):
        """Initialize.
        
        Parameters
        ----------
        source : PGFeatureSource
            The source
        cursor : -
            A cursor created by the source
        """
        self._cur = cursor
        self._table = source.typename
        self._ftype = source.featuretype
        self._colnames = [x[0] for x in self._cur.description]
        self.pkey_column = source.pkey_column
        
        try:
            self._gidx = [self._colnames.index(s.name) for s in self._ftype.schema() if isgeometry(s)]
        except:
            print self._colnames, s.name
            raise
        self.hasfrontfilter = 0
        if filter is not None:
            self._compiled_filter = compile(filter, 'string', 'eval')
            self._locals = {}
            self._locals.update(params)
            self.hasfrontfilter = 1
            
    def __iter__(self):
        # Return self, following the Python iterator protocol
        return self

    def next(self):
        """Returns next feature, applying optional front-end filter."""
        if not self.hasfrontfilter:
            row = self._cur.fetchone()
            if not row:
                raise StopIteration
            # Make and return a Feature
            data = []
            for i in range(len(row)):
                if i in self._gidx:
                    wkb = row[i].decode('hex')
                    data.append(Geometry.fromWKB(wkb))
                else:
                    data.append(row[i])
            d = dict(zip(self._colnames, data))
            ft = self._ftype
            return ft(str(d[self.pkey_column]), pdict=d)
        else: 
            while 1:
                row = self._cur.fetchone()
                if not row:
                    raise StopIteration
                # Evaluate the front filter
                data = []
                for i in range(len(row)):
                    if i in self._gidx:
                        wkb = row[i].decode('hex')
                        data.append(Geometry.fromWKB(wkb))
                    else:
                        data.append(row[i])
                d = dict(zip(self._colnames, data))
                ft = self._ftype
                feature = ft(str(d[self.pkey_column]), pdict=d)
                g = feature.properties[ft.__g__]
                self._locals.update({'f': feature.properties,
                                     'g': g, 'id': feature.id})
                if not eval(self._compiled_filter, globals(), self._locals):
                    continue
                else:
                    return feature

            

