/*
# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================
*/

#include "Python.h"
#include "structmember.h"

#include "geos_c.h"
#include "proj_api.h"

static PyObject *GEOSError;

typedef struct {
    PyObject_HEAD
    GEOSCoordSeq seq;
    /* Flag for whether seq is an owned (1) or borrowed (0) reference to
       allocated memory */
    int owns;
} CoordSeq;

staticforward PyTypeObject CoordinateSequenceType;

typedef struct {
    PyObject_HEAD
    GEOSGeom geom;
    /* Flag for whether seq is an owned (1) or borrowed (0) reference to
       allocated memory */
    int owns;
} Geometry;

staticforward PyTypeObject GeometryType;

/*
==============================================================================
 CoordinateSequence
==============================================================================
*/

/* === Alloc/dealloc === */

PyObject *
CoordSeq_New(GEOSCoordSeq s)
{
    CoordSeq *self;
    
    if (!s)
        return NULL;
    
    self = PyObject_NEW(CoordSeq, &CoordinateSequenceType);
    if (self == NULL)
    {
        GEOSCoordSeq_destroy(self->seq);
        return NULL;
    }
    
    self->seq = s;
    self->owns = 1;
    return (PyObject *) self;
}

static PyObject *
_createCoordSeq(PyObject *self, PyObject *args)
{
    int size=1;
    int dims=3;
    GEOSCoordSeq s;
    
    if (!PyArg_ParseTuple(args, "i|i", &size, &dims))
        return NULL;
        
    if (size < 1)
    {
        PyErr_SetString(PyExc_ValueError, "Sequence must have size >=1");
        return NULL;
    }
    if (dims < 2 || dims > 3)
    {
        PyErr_SetString(PyExc_ValueError, "Dimensions must be 2 or 3");
        return NULL;
    }
    
    s = (GEOSCoordSeq) GEOSCoordSeq_create(size, dims);
    if (!s)
        return NULL;
    
    return CoordSeq_New(s);
}

static void
CoordSeq_dealloc(CoordSeq *self)
{
    if (self->owns && self->seq != NULL)
    {
        GEOSCoordSeq_destroy(self->seq);
    }
    self->seq = NULL;
    self->ob_type->tp_free((PyObject*) self);
}

static PyObject *
CoordSeq_clone(CoordSeq *self)
{
    CoordSeq *s;
    GEOSCoordSeq new;

    s = (CoordSeq *) self;
    new = GEOSCoordSeq_clone(s->seq);
    if (!new)
        return NULL;

    return (PyObject *) CoordSeq_New(new);
}

/* === Methods === */

static PyObject *
CoordSeq_getSize(CoordSeq *self)
{
    unsigned int size;
    if (GEOSCoordSeq_getSize(self->seq, &size) == 0)
    {
        PyErr_SetString(PyExc_Exception, "Failed to get sequence size");
        return NULL;
    }    
    return Py_BuildValue("i", size);
}

static PyObject *
CoordSeq_getDimensions(CoordSeq *self)
{
    unsigned int dims;
    if (!GEOSCoordSeq_getDimensions(self->seq, &dims))
    {
        PyErr_SetString(PyExc_Exception, "Failed to get sequence dimensions");
        return NULL;
    }    
    return Py_BuildValue("i", dims);
}

static PyObject *
CoordSeq_getX(CoordSeq *self, PyObject *item)
{
    unsigned int num;
    double value;
    long n;
    
    if (!PyInt_Check(item))
    {
		PyErr_SetString(PyExc_TypeError, "index must be an integer");
		return NULL;
	}
    
    GEOSCoordSeq_getSize(self->seq, &num);
    if (num == 0)
    {
        PyErr_SetString(PyExc_Exception, "Sequence has no coordinates");
        return NULL;
    }

    n = PyInt_AS_LONG(item);
    if (n < 0 || n >= num)
    {
        PyErr_Format(PyExc_IndexError,
                "Index is beyond valid range [0:%d]", num);
        return NULL;
    }
    
    if (!GEOSCoordSeq_getX(self->seq, (int)n, &value))
    {
        PyErr_SetString(PyExc_Exception, "Failed to get X");
        return NULL;
    }
    return Py_BuildValue("d", value);
}

static PyObject *
CoordSeq_getY(CoordSeq *self, PyObject *item)
{
    unsigned int num;
    double value;
    long n;
    
    if (!PyInt_Check(item))
    {
		PyErr_SetString(PyExc_TypeError, "index must be an integer");
		return NULL;
	}
    
    GEOSCoordSeq_getSize(self->seq, &num);
    if (num == 0)
    {
        PyErr_SetString(PyExc_Exception, "Sequence has no coordinates");
        return NULL;
    }

    n = PyInt_AS_LONG(item);
    if (n < 0 || n >= num)
    {
        PyErr_Format(PyExc_IndexError,
                "Index is beyond valid range [0:%d]", num);
        return NULL;
    }
    
    if (!GEOSCoordSeq_getY(self->seq, (int)n, &value))
    {
        PyErr_SetString(PyExc_Exception, "Failed to get Y");
        return NULL;
    }
    return Py_BuildValue("d", value);
}

static PyObject *
CoordSeq_getZ(CoordSeq *self, PyObject *item)
{
    unsigned int num;
    double value;
    long n;
    
    if (!PyInt_Check(item))
    {
		PyErr_SetString(PyExc_TypeError, "index must be an integer");
		return NULL;
	}
    
    GEOSCoordSeq_getSize(self->seq, &num);
    if (num == 0)
    {
        PyErr_SetString(PyExc_Exception, "Sequence has no coordinates");
        return NULL;
    }

    n = PyInt_AS_LONG(item);
    if (n < 0 || n >= num)
    {
        PyErr_Format(PyExc_IndexError,
                "Index is beyond valid range [0:%d]", num);
        return NULL;
    }
    
    if (!GEOSCoordSeq_getZ(self->seq, (int)n, &value))
    {
        PyErr_SetString(PyExc_Exception, "Failed to get Z");
        return NULL;
    }
    return Py_BuildValue("d", value);
}

static PyObject *
CoordSeq_setX(CoordSeq *self, PyObject *args)
{
    int n;
    unsigned int num;
    double value;
    if (!PyArg_ParseTuple(args, "id", &n, &value))
    {
		return NULL;
	}
    
    GEOSCoordSeq_getSize(self->seq, &num);
    if (num == 0)
    {
        PyErr_SetString(PyExc_Exception, "Sequence has no coordinates");
        return NULL;
    }

    if (n < 0 || n >= num)
    {
        PyErr_Format(PyExc_IndexError,
                "Index is beyond valid range [0:%d]", num);
        return NULL;
    }
    
    if (!GEOSCoordSeq_setX(self->seq, (int)n, value))
    {
        PyErr_SetString(PyExc_Exception, "Failed to set X");
        return NULL;
    }
    
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
CoordSeq_setY(CoordSeq *self, PyObject *args)
{
    int n;
    unsigned int num;
    double value;
    if (!PyArg_ParseTuple(args, "id", &n, &value))
    {
		return NULL;
	}
    
    GEOSCoordSeq_getSize(self->seq, &num);
    if (num == 0)
    {
        PyErr_SetString(PyExc_Exception, "Sequence has no coordinates");
        return NULL;
    }
    
    if (n < 0 || n >= num)
    {
        PyErr_Format(PyExc_IndexError,
                     "Index is beyond valid range [0:%d]", num);
        return NULL;
    }
    
    if (!GEOSCoordSeq_setY(self->seq, (int)n, value))
    {
        PyErr_SetString(PyExc_Exception, "Failed to set Y");
        return NULL;
    }
    
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
CoordSeq_setZ(CoordSeq *self, PyObject *args)
{
    int n;
    unsigned int num;
    double value;
    if (!PyArg_ParseTuple(args, "id", &n, &value))
    {
		return NULL;
	}
    
    GEOSCoordSeq_getSize(self->seq, &num);
    if (num == 0)
    {
        PyErr_SetString(PyExc_Exception, "Sequence has no coordinates");
        return NULL;
    }
    
    if (n < 0 || n >= num)
    {
        PyErr_Format(PyExc_IndexError,
                     "Index is beyond valid range [0:%d]", num);
        return NULL;
    }
    
    if (!GEOSCoordSeq_setZ(self->seq, (int)n, value))
    {
        PyErr_SetString(PyExc_Exception, "Failed to set Z");
        return NULL;
    }
    
    Py_INCREF(Py_None);
    return Py_None;
}

/* === Define Methods === */

static PyMethodDef coordseq_methods[] = {
    {"getSize", (PyCFunction)CoordSeq_getSize, METH_NOARGS, "Returns number of coordinates in the sequence"},
    {"getDimensions", (PyCFunction)CoordSeq_getDimensions, METH_NOARGS, "Returns number of coordinate dimensions"},
    {"clone", (PyCFunction)CoordSeq_clone, METH_NOARGS, "Returns clone"},
    {"getX", (PyCFunction)CoordSeq_getX, METH_O, "Get Nth X, index starting at 0"},
    {"getY", (PyCFunction)CoordSeq_getY, METH_O, "Get Nth Y, index starting at 0"},
    {"getZ", (PyCFunction)CoordSeq_getZ, METH_O, "Get Nth Z, index starting at 0"},
    {"setX", (PyCFunction)CoordSeq_setX, METH_VARARGS, "Set Nth X, index starting at 0"},
    {"setY", (PyCFunction)CoordSeq_setY, METH_VARARGS, "Set Nth Y, index starting at 0"},
    {"setZ", (PyCFunction)CoordSeq_setZ, METH_VARARGS, "Set Nth Z, index starting at 0"},
    {NULL}
};

/* === Define Type === */

static PyTypeObject CoordinateSequenceType = {
    PyObject_HEAD_INIT(NULL)
    0,                              /*ob_size*/
    "GEOSCoordinateSequence", /*tp_name*/
    sizeof(CoordSeq),               /*tp_basicsize*/
    0,                              /*tp_itemsize*/
    (destructor)CoordSeq_dealloc,   /*tp_dealloc*/
    0,                              /*tp_print*/
    0,                              /*tp_getattr*/
    0,                              /*tp_setattr*/
    0,                              /*tp_compare*/
    0,                              /*tp_repr*/
    0,                              /*tp_as_number*/
    0,                              /*tp_as_sequence*/
    0,                              /*tp_as_mapping*/
    0,                              /*tp_hash */
    0,                              /*tp_call*/
    0,                              /*tp_str*/
    0,                              /*tp_getattro*/
    0,                              /*tp_setattro*/
    0,                              /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "A GEOS Coordinate Sequence",           /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    coordseq_methods,             /* tp_methods */
    0,                     /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    0,      /* tp_init */
    0,                         /* tp_alloc */
    0,                 /* tp_new */
};

/*
==============================================================================
 Geometry
==============================================================================
*/

/* === Alloc/dealloc === */

static void
Geometry_dealloc(Geometry *self)
{
    if (self->owns && self->geom != NULL)
    {
        GEOSGeom_destroy(self->geom);
    }
    self->geom = NULL;
    self->ob_type->tp_free((PyObject*) self);
}

PyObject *
Geometry_New(GEOSGeom g)
{
    Geometry *self;
    
    if (!g)
        return NULL;
    
    self = PyObject_NEW(Geometry, &GeometryType);
    if (self == NULL)
    {
        GEOSGeom_destroy(g);
        return NULL;
    }
    
    self->geom = g;
    self->owns = 1;
    return (PyObject *) self;
}

static PyObject *
Geometry_clone(PyObject *self)
{
    Geometry *g;
    GEOSGeom new;

    g = (Geometry *) self;
    if (!g->geom)
    {
        PyErr_SetString(PyExc_Exception, "NULL GEOSGeom");
        return NULL;
    }
    
    new = GEOSGeom_clone(g->geom);
    if (!new)
        return NULL;

    return (PyObject *) Geometry_New(new);
}

static PyObject *
_createGeomFromWKB(PyObject *self, PyObject *args)
{
    unsigned char *data=NULL;
    int size;
    GEOSGeom g;
    
    if (!PyArg_ParseTuple(args, "s#", &data, &size))
        return NULL;

    g = GEOSGeomFromWKB_buf(data, (size_t)size);
    if (!g)
    {
        /* Rely on exception set by GEOS through error_handler */
        return NULL;
    }
        
    return Geometry_New(g);
}

static PyObject *
_updateGeomFromWKB(PyObject *self, PyObject *args)
{
    unsigned char *data=NULL;
    int size;
    GEOSGeom g;
    Geometry *gs;
    
    if (!PyArg_ParseTuple(args, "s#", &data, &size))
        return NULL;

    g = GEOSGeomFromWKB_buf(data, (size_t)size);
    if (!g)
    {
        /* Rely on exception set by GEOS through error_handler */
        return NULL;
    }

    gs = (Geometry *) self;
    if (gs->geom)
        GEOSGeom_destroy(gs->geom);
    gs->geom = g;
    gs->owns = 1;
    
    Py_INCREF(Py_None);
    return Py_None;
}
    
static PyObject *
_createGeomFromWKT(PyObject *self, PyObject *args)
{
    char *data;
    GEOSGeom g;
    
    if (!PyArg_ParseTuple(args, "s", &data))
        return NULL;
        
    g = GEOSGeomFromWKT(data);
    if (!g)
    {
        /* Rely on exception set by GEOS through error_handler */
        return NULL;
    }
    
    return Geometry_New(g);
}

static PyObject *
_updateGeomFromWKT(PyObject *self, PyObject *args)
{
    char *data;
    GEOSGeom g;
    Geometry *gs;

    if (!PyArg_ParseTuple(args, "s", &data))
        return NULL;
        
    g = GEOSGeomFromWKT(data);
    if (!g)
    {
        /* Rely on exception set by GEOS through error_handler */
        return NULL;
    }

    gs = (Geometry *) self;
    if (gs->geom)
        GEOSGeom_destroy(gs->geom);
    gs->geom = g;
    gs->owns = 1;
    
    Py_INCREF(Py_None);
    return Py_None;
}
    
static PyObject *
_createGeomFromCoordSeq(PyObject *self, PyObject *args)
{
    PyObject *seq=NULL;
    int typeid;
    GEOSGeom g;
    CoordSeq *s;
    PyObject *geom;
    unsigned int size=0;
    
    if (!PyArg_ParseTuple(args, "O!i", &CoordinateSequenceType, &seq,  &typeid))
        return NULL;

    s = (CoordSeq *) seq;

    GEOSCoordSeq_getSize(s->seq, &size);
    if (size < 1)
    {
        PyErr_SetString(PyExc_ValueError, "Geometry can not be created from a zero length sequence");
        return NULL;
    }
    
    switch (typeid)
    {
        case GEOS_POINT:
            if (size > 1)
            {
                PyErr_SetString(PyExc_ValueError, "Sequence too long for Point type (1)");
                return NULL;
            }
            g = GEOSGeom_createPoint(s->seq);
            break;
        case GEOS_LINESTRING:
            g = GEOSGeom_createLineString(s->seq);
            break;
        case GEOS_LINEARRING:
            g = GEOSGeom_createLinearRing(s->seq);
            break;
        default:
            PyErr_Format(PyExc_ValueError, "Invalid typeid %d", typeid);
            return NULL;
    }
            
    if (!g)
        return NULL;
        
    geom = Geometry_New(g);
    if (!geom)
        return NULL;
    
    s->owns = 0;
    return geom;        
}

static PyObject *
Geometry_copyFromCoordSeq(PyObject *self, PyObject *args)
{
    PyObject *seq=NULL;
    int typeid;
    GEOSGeom new;
    GEOSCoordSeq cseq;
    Geometry *gs;
    CoordSeq *s;
    unsigned int size=0;
    
    if (!PyArg_ParseTuple(args, "O!i", &CoordinateSequenceType, &seq,  &typeid))
        return NULL;

    s = (CoordSeq *) seq;

    GEOSCoordSeq_getSize(s->seq, &size);
    if (size < 1)
    {
        PyErr_SetString(PyExc_ValueError, "Geometry can not be created from a zero length sequence");
        return NULL;
    }
    
    cseq = GEOSCoordSeq_clone(s->seq);
    if (!cseq)
    {
        PyErr_SetString(PyExc_Exception, "Failed to clone coordinate sequence");
        return NULL;
    }
        
    switch (typeid)
    {
        case GEOS_POINT:
            if (size > 1)
            {
                PyErr_SetString(PyExc_ValueError, "Sequence too long for Point type (1)");
                return NULL;
            }
            new = GEOSGeom_createPoint(cseq);
            break;
        case GEOS_LINESTRING:
            new = GEOSGeom_createLineString(cseq);
            break;
        case GEOS_LINEARRING:
            new = GEOSGeom_createLinearRing(cseq);
            break;
        default:
            PyErr_Format(PyExc_ValueError, "Invalid typeid %d", typeid);
            return NULL;
    }
            
    if (!new)
    {
        /* count on exception from GEOS */
        return NULL;
    }
      
    gs = (Geometry *) self;
    if (gs->geom)
        GEOSGeom_destroy(gs->geom);
    gs->geom = new;
    
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
Geometry_copyFromGeom(PyObject *self, PyObject *args)
{
    PyObject *g2;
    Geometry *g;
    GEOSGeom new;
    Geometry *gs;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;

    g = (Geometry *) g2;
    if (!g->geom)
    {
        PyErr_SetString(PyExc_Exception, "NULL GEOSGeom");
        return NULL;
    }
    
    new = GEOSGeom_clone(g->geom);
    if (!new)
    {
        /* exception provided by GEOS */
        return NULL;
    }
      
    gs = (Geometry *) self;
    if (gs->geom)
        GEOSGeom_destroy(gs->geom);
    gs->geom = new;
    gs->owns = 1;
    
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
_createPolygon(PyObject *self, PyObject *args)
{
    GEOSGeom g;
    PyObject *shell=NULL;
    PyObject *holes=NULL;
    Geometry *gs;
    Geometry *gh;
    GEOSGeom *holes_ptr;
    int i, nholes=0;
    PyObject *geom;
    
    if (!PyArg_ParseTuple(args, "O!|O", &GeometryType, &shell, &holes))
        return NULL;

    gs = (Geometry *) shell;
    if (!holes)
    {
        holes_ptr = (GEOSGeom*) malloc(sizeof(GEOSGeom)*nholes);
        g = GEOSGeom_createPolygon(gs->geom, holes_ptr, nholes);
        free(holes_ptr);
    }
    else
    {
        if (!PyList_Check(holes))
        {
            PyErr_SetString(PyExc_TypeError, "holes argument must be a list of rings");
            return NULL;
        }
            
        nholes = PyList_Size(holes);
        holes_ptr = (GEOSGeom*) malloc(sizeof(GEOSGeom)*nholes);
        for (i=0; i<nholes; i++)
        {
            gh = (Geometry *) PyList_GetItem(holes, i);
            holes_ptr[i] = (GEOSGeom) gh->geom;
        }
            
        g = GEOSGeom_createPolygon(gs->geom, holes_ptr, nholes);
        free(holes_ptr);
    }
    if (!g)
    return NULL;
        
    geom = Geometry_New(g);
    if (!geom)
        return NULL;
    
    gs->owns = 0;
    for (i=0; i<nholes; i++)
    {
        gh = (Geometry *) PyList_GetItem(holes, i);
        gh->owns = 0;
    }   
    return geom;         
}

static PyObject *
_createCollection(PyObject *self, PyObject *args)
{
    int typeid;
    PyObject *geoms;
    GEOSGeom new;
    Geometry *g;
    GEOSGeom *geoms_ptr;
    int i, ngeoms=0;
    PyObject *geom;
    
    if (!PyArg_ParseTuple(args, "iO", &typeid, &geoms))
        return NULL;

    if (!PyList_Check(geoms))
    {
        PyErr_SetString(PyExc_TypeError, "geoms arg must be a list of geoms");
        return NULL;
    }
            
    ngeoms = PyList_Size(geoms);
    geoms_ptr = (GEOSGeom *) malloc(sizeof(GEOSGeom)*ngeoms);
    for (i=0; i<ngeoms; i++)
    {
        g = (Geometry *) PyList_GetItem(geoms, i);
        geoms_ptr[i] = (GEOSGeom) g->geom;
    }
            
    new = GEOSGeom_createCollection(typeid, geoms_ptr, ngeoms);
    free(geoms_ptr);
    if (!new)
        return NULL;
        
    geom = Geometry_New(new);
    if (!geom)
        return NULL;
    
    for (i=0; i<ngeoms; i++)
    {
        g = (Geometry *) PyList_GetItem(geoms, i);
        g->owns = 0;
    }   
    
    return geom;         
}

/* === Utility methods === */

static PyObject *
Geometry_toWKB(Geometry *self)
{
    char *buf;
    size_t size;
    PyObject *wkb;
    buf = GEOSGeomToWKB_buf(self->geom, &size);
    wkb = PyString_FromStringAndSize(buf, (int)size);
    free(buf);
    return wkb;
}

static PyObject *
to_wkt(Geometry *self)
{
    char *buf;
    PyObject *wkt;
    buf = (char *)GEOSGeomToWKT(self->geom);
    wkt = PyString_FromString(buf);
    free(buf);
    return wkt;
}

static PyObject *
Geometry_getTypeId(Geometry *self)
{
    return Py_BuildValue("i", (int)GEOSGeomTypeId(self->geom));
}

#if defined(GEOS_CAPI_VERSION_MAJOR) && defined(GEOS_CAPI_VERSION_MINOR) && GEOS_CAPI_VERSION_MAJOR >= 1 && GEOS_CAPI_VERSION_MINOR >= 1
static PyObject *
Geometry_getGeomType(Geometry *self)
{
    char *buf;
    PyObject *name;
    buf = (char *)GEOSGeomType(self->geom);
    name = Py_BuildValue("s", buf);
    free(buf);
    return name;
}
#endif /* GEOS C API version 1.0.2 only */

/* === Unary Predicates === */

static PyObject *
Geometry_isEmpty(Geometry *self)
{
    int result;
    result = GEOSisEmpty(self->geom);
    if (result > 1)
        return NULL;
    return Py_BuildValue("i", result);
}

static PyObject *
Geometry_isValid(Geometry *self)
{
    int result;
    result = GEOSisValid(self->geom);
    if (result > 1)
        return NULL;
    return Py_BuildValue("i", result);}

static PyObject *
Geometry_isSimple(Geometry *self)
{
    int result;
    result = GEOSisSimple(self->geom);
    if (result > 1)
        return NULL;
    return Py_BuildValue("i", result);
}

static PyObject *
Geometry_isRing(Geometry *self)
{
    int result;
    result = GEOSisRing(self->geom);
    if (result > 1)
        return NULL;
    return Py_BuildValue("i", result);}

static PyObject *
Geometry_hasZ(Geometry *self)
{
    int result;
    result = GEOSHasZ(self->geom);
    if (result > 1)
        return NULL;
    return Py_BuildValue("i", result);}

/* === Topology === */

#if defined(GEOS_CAPI_VERSION_MAJOR) && defined(GEOS_CAPI_VERSION_MINOR) && GEOS_CAPI_VERSION_MAJOR >= 1 && GEOS_CAPI_VERSION_MINOR >= 1
static PyObject *
Geometry_envelope(Geometry *self)
{
    GEOSGeom new;

    new = GEOSEnvelope(self->geom);
    if (!new)
    {
        return NULL;
    }
    
    return Geometry_New(new);
}
#else
static PyObject *
Geometry_envelope(Geometry *self)
{
    GEOSCoordSeq s, envseq=NULL;
    GEOSGeom new=NULL, geom, shell, ring;
    GEOSGeom holes_ptr[1];
    int i, k, gtype, dims, num;
    unsigned int size;
    double minx=1.0e10, miny=1.0e10, maxx=-1.0e10, maxy=-1.0e10;
    double x, y;
        
    gtype = GEOSGeomTypeId(self->geom);
    dims = GEOSGeom_getDimensions(self->geom);
    
    switch (gtype)
    {
        case GEOS_POINT:
            /* In this case, the envelope will be a point */
            envseq = (GEOSCoordSeq) GEOSCoordSeq_create(1, dims);

            /* Get values from the source sequence */
            s = GEOSGeom_getCoordSeq(self->geom);
            GEOSCoordSeq_getX(s, 0, &x);
            GEOSCoordSeq_getY(s, 0, &y);

            /* Set them in the envelope sequence */
            GEOSCoordSeq_setX(envseq, 0, x);
            GEOSCoordSeq_setY(envseq, 0, y);

            new = GEOSGeom_createPoint(envseq);
            break;
        case GEOS_LINESTRING:
            envseq = (GEOSCoordSeq) GEOSCoordSeq_create(5, dims);
            
            /* Get bounding values from the source sequence */
            s = GEOSGeom_getCoordSeq(self->geom);
            GEOSCoordSeq_getSize(s, &size);

            for (k=0; k<size; k++)
            {
                GEOSCoordSeq_getX(s, k, &x);
                GEOSCoordSeq_getY(s, k, &y);
                if (x < minx) minx = x;
                if (x > maxx) maxx = x;
                if (y < miny) miny = y;
                if (y > maxy) maxy = y;
            }

            /* Set the envelope sequence */
            GEOSCoordSeq_setX(envseq, 0, minx);
            GEOSCoordSeq_setY(envseq, 0, miny);
            GEOSCoordSeq_setX(envseq, 1, minx);
            GEOSCoordSeq_setY(envseq, 1, maxy);
            GEOSCoordSeq_setX(envseq, 2, maxx);
            GEOSCoordSeq_setY(envseq, 2, maxy);
            GEOSCoordSeq_setX(envseq, 3, maxx);
            GEOSCoordSeq_setY(envseq, 3, miny);
            GEOSCoordSeq_setX(envseq, 4, minx);
            GEOSCoordSeq_setY(envseq, 4, miny);

            shell = GEOSGeom_createLinearRing(envseq);
            new = GEOSGeom_createPolygon(shell, holes_ptr, 0);
            break;
        case GEOS_LINEARRING:
            envseq = (GEOSCoordSeq) GEOSCoordSeq_create(5, dims);
            
            /* Get bounding values from the source sequence */
            s = GEOSGeom_getCoordSeq(self->geom);
            GEOSCoordSeq_getSize(s, &size);

            for (k=0; k<size; k++)
            {
                GEOSCoordSeq_getX(s, k, &x);
                GEOSCoordSeq_getY(s, k, &y);
                if (x < minx) minx = x;
                if (x > maxx) maxx = x;
                if (y < miny) miny = y;
                if (y > maxy) maxy = y;
            }

            /* Set the envelope sequence */
            GEOSCoordSeq_setX(envseq, 0, minx);
            GEOSCoordSeq_setY(envseq, 0, miny);
            GEOSCoordSeq_setX(envseq, 1, minx);
            GEOSCoordSeq_setY(envseq, 1, maxy);
            GEOSCoordSeq_setX(envseq, 2, maxx);
            GEOSCoordSeq_setY(envseq, 2, maxy);
            GEOSCoordSeq_setX(envseq, 3, maxx);
            GEOSCoordSeq_setY(envseq, 3, miny);
            GEOSCoordSeq_setX(envseq, 4, minx);
            GEOSCoordSeq_setY(envseq, 4, miny);

            shell = GEOSGeom_createLinearRing(envseq);
            new = GEOSGeom_createPolygon(shell, holes_ptr, 0);
            break; 
        case GEOS_POLYGON:
            envseq = (GEOSCoordSeq) GEOSCoordSeq_create(5, dims);
            
            /* Get bounding values from the source sequence */
            ring = GEOSGetExteriorRing(self->geom);
            s = GEOSGeom_getCoordSeq(ring);
            GEOSCoordSeq_getSize(s, &size);

            for (k=0; k<size; k++)
            {
                GEOSCoordSeq_getX(s, k, &x);
                GEOSCoordSeq_getY(s, k, &y);
                if (x < minx) minx = x;
                if (x > maxx) maxx = x;
                if (y < miny) miny = y;
                if (y > maxy) maxy = y;
            }

            /* Set the envelope sequence */
            GEOSCoordSeq_setX(envseq, 0, minx);
            GEOSCoordSeq_setY(envseq, 0, miny);
            GEOSCoordSeq_setX(envseq, 1, minx);
            GEOSCoordSeq_setY(envseq, 1, maxy);
            GEOSCoordSeq_setX(envseq, 2, maxx);
            GEOSCoordSeq_setY(envseq, 2, maxy);
            GEOSCoordSeq_setX(envseq, 3, maxx);
            GEOSCoordSeq_setY(envseq, 3, miny);
            GEOSCoordSeq_setX(envseq, 4, minx);
            GEOSCoordSeq_setY(envseq, 4, miny);

            shell = GEOSGeom_createLinearRing(envseq);
            new = GEOSGeom_createPolygon(shell, holes_ptr, 0);
            break; 
        case GEOS_MULTIPOINT:
            envseq = (GEOSCoordSeq) GEOSCoordSeq_create(5, dims);
            num = GEOSGetNumGeometries(self->geom);
            for (i=0; i<num; i++)
            {
                geom = GEOSGetGeometryN(self->geom, i);
                
                /* Get values from the source sequence */
                s = GEOSGeom_getCoordSeq(geom);
                GEOSCoordSeq_getX(s, 0, &x);
                GEOSCoordSeq_getY(s, 0, &y);

                if (x < minx) minx = x;
                if (x > maxx) maxx = x;
                if (y < miny) miny = y;
                if (y > maxy) maxy = y;
            }
            /* Set the envelope sequence */
            GEOSCoordSeq_setX(envseq, 0, minx);
            GEOSCoordSeq_setY(envseq, 0, miny);
            GEOSCoordSeq_setX(envseq, 1, minx);
            GEOSCoordSeq_setY(envseq, 1, maxy);
            GEOSCoordSeq_setX(envseq, 2, maxx);
            GEOSCoordSeq_setY(envseq, 2, maxy);
            GEOSCoordSeq_setX(envseq, 3, maxx);
            GEOSCoordSeq_setY(envseq, 3, miny);
            GEOSCoordSeq_setX(envseq, 4, minx);
            GEOSCoordSeq_setY(envseq, 4, miny);

            shell = GEOSGeom_createLinearRing(envseq);
            new = GEOSGeom_createPolygon(shell, holes_ptr, 0);
            break;
        case GEOS_MULTILINESTRING:
            envseq = (GEOSCoordSeq) GEOSCoordSeq_create(5, dims);
            num = GEOSGetNumGeometries(self->geom);
            for (i=0; i<num; i++)
            {
                geom = GEOSGetGeometryN(self->geom, i);
                
                /* Get values from the source sequence */
                s = GEOSGeom_getCoordSeq(geom);
                GEOSCoordSeq_getSize(s, &size);

                for (k=0; k<size; k++)
                {
                    GEOSCoordSeq_getX(s, k, &x);
                    GEOSCoordSeq_getY(s, k, &y);
                    if (x < minx) minx = x;
                    if (x > maxx) maxx = x;
                    if (y < miny) miny = y;
                    if (y > maxy) maxy = y;
                }
            }
            
            /* Set the envelope sequence */
            GEOSCoordSeq_setX(envseq, 0, minx);
            GEOSCoordSeq_setY(envseq, 0, miny);
            GEOSCoordSeq_setX(envseq, 1, minx);
            GEOSCoordSeq_setY(envseq, 1, maxy);
            GEOSCoordSeq_setX(envseq, 2, maxx);
            GEOSCoordSeq_setY(envseq, 2, maxy);
            GEOSCoordSeq_setX(envseq, 3, maxx);
            GEOSCoordSeq_setY(envseq, 3, miny);
            GEOSCoordSeq_setX(envseq, 4, minx);
            GEOSCoordSeq_setY(envseq, 4, miny);

            shell = GEOSGeom_createLinearRing(envseq);
            new = GEOSGeom_createPolygon(shell, holes_ptr, 0);
            break;
        case GEOS_MULTIPOLYGON:
            envseq = (GEOSCoordSeq) GEOSCoordSeq_create(5, dims);
            num = GEOSGetNumGeometries(self->geom);
            for (i=0; i<num; i++)
            {
                geom = GEOSGetGeometryN(self->geom, i);
                
                /* Get values from the source sequence */
                ring = GEOSGetExteriorRing(geom);
                s = GEOSGeom_getCoordSeq(ring);
                GEOSCoordSeq_getSize(s, &size);

                for (k=0; k<size; k++)
                {
                    GEOSCoordSeq_getX(s, k, &x);
                    GEOSCoordSeq_getY(s, k, &y);
                    if (x < minx) minx = x;
                    if (x > maxx) maxx = x;
                    if (y < miny) miny = y;
                    if (y > maxy) maxy = y;
                }
            }
            
            /* Set the envelope sequence */
            GEOSCoordSeq_setX(envseq, 0, minx);
            GEOSCoordSeq_setY(envseq, 0, miny);
            GEOSCoordSeq_setX(envseq, 1, minx);
            GEOSCoordSeq_setY(envseq, 1, maxy);
            GEOSCoordSeq_setX(envseq, 2, maxx);
            GEOSCoordSeq_setY(envseq, 2, maxy);
            GEOSCoordSeq_setX(envseq, 3, maxx);
            GEOSCoordSeq_setY(envseq, 3, miny);
            GEOSCoordSeq_setX(envseq, 4, minx);
            GEOSCoordSeq_setY(envseq, 4, miny);

            shell = GEOSGeom_createLinearRing(envseq);
            new = GEOSGeom_createPolygon(shell, holes_ptr, 0);
            break;
        case GEOS_GEOMETRYCOLLECTION:
            envseq = (GEOSCoordSeq) GEOSCoordSeq_create(5, dims);
            num = GEOSGetNumGeometries(self->geom);
            for (i=0; i<num; i++)
            {
                geom = GEOSGetGeometryN(self->geom, i);
                              
                if (GEOSGeomTypeId(geom) == GEOS_POINT)
                {
                    s = GEOSGeom_getCoordSeq(geom);
                    GEOSCoordSeq_getX(s, 0, &x);
                    GEOSCoordSeq_getY(s, 0, &y);
                    if (x < minx) minx = x;
                    if (x > maxx) maxx = x;
                    if (y < miny) miny = y;
                    if (y > maxy) maxy = y;
                }
                else if (GEOSGeomTypeId(geom) < 3)
                {
                    s = GEOSGeom_getCoordSeq(geom);
                    GEOSCoordSeq_getSize(s, &size);
                    for (k=0; k<size; k++)
                    {
                        GEOSCoordSeq_getX(s, k, &x);
                        GEOSCoordSeq_getY(s, k, &y);
                        if (x < minx) minx = x;
                        if (x > maxx) maxx = x;
                        if (y < miny) miny = y;
                        if (y > maxy) maxy = y;
                    }
                }
                else
                {
                    ring = GEOSGetExteriorRing(geom);
                    s = GEOSGeom_getCoordSeq(ring);
                    GEOSCoordSeq_getSize(s, &size);
                    for (k=0; k<size; k++)
                    {
                        GEOSCoordSeq_getX(s, k, &x);
                        GEOSCoordSeq_getY(s, k, &y);
                        if (x < minx) minx = x;
                        if (x > maxx) maxx = x;
                        if (y < miny) miny = y;
                        if (y > maxy) maxy = y;
                    }   
                }
            }
            
            /* Set the envelope sequence */
            GEOSCoordSeq_setX(envseq, 0, minx);
            GEOSCoordSeq_setY(envseq, 0, miny);
            GEOSCoordSeq_setX(envseq, 1, minx);
            GEOSCoordSeq_setY(envseq, 1, maxy);
            GEOSCoordSeq_setX(envseq, 2, maxx);
            GEOSCoordSeq_setY(envseq, 2, maxy);
            GEOSCoordSeq_setX(envseq, 3, maxx);
            GEOSCoordSeq_setY(envseq, 3, miny);
            GEOSCoordSeq_setX(envseq, 4, minx);
            GEOSCoordSeq_setY(envseq, 4, miny);

            shell = GEOSGeom_createLinearRing(envseq);
            new = GEOSGeom_createPolygon(shell, holes_ptr, 0);
            break;
        default:
            PyErr_Format(PyExc_ValueError, "Invalid typeid %d", gtype);
            return NULL;
    }

    if (!new)
        return NULL;
            
    return Geometry_New(new);
}
#endif /* GEOS C API version 1.0.2 only */

static PyObject *
Geometry_intersection(Geometry *self, PyObject *args)
{
    PyObject *g2;
    Geometry *g;
    GEOSGeom new;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    new = GEOSIntersection(self->geom, g->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate intersection");
        return NULL;
    }
    return Geometry_New(new);
}

static PyObject *
Geometry_buffer(Geometry *self, PyObject *args)
{
    double width;
    int quadsegs=16;
    GEOSGeom new;
    
    if (!PyArg_ParseTuple(args, "d|i", &width, &quadsegs))
        return NULL;
    
    new = GEOSBuffer(self->geom, width, quadsegs);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate buffer");
        return NULL;
    }
    return Geometry_New(new);
}

static PyObject *
Geometry_convexHull(Geometry *self)
{
    GEOSGeom new;
    
    new = GEOSConvexHull(self->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate convex hull");
        return NULL;
    }
    
    return Geometry_New(new);
}

static PyObject *
Geometry_difference(Geometry *self, PyObject *args)
{
    PyObject *g2;
    Geometry *g;
    GEOSGeom new;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    new = GEOSDifference(self->geom, g->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate difference");
        return NULL;
    }
    
    return Geometry_New(new);
}

static PyObject *
Geometry_symDifference(Geometry *self, PyObject *args)
{
    PyObject *g2;
    Geometry *g;
    GEOSGeom new;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;    
    new = GEOSSymDifference(self->geom, g->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate symmetric difference");
        return NULL;
    }
    
    return Geometry_New(new);
}

static PyObject *
Geometry_boundary(Geometry *self)
{
    GEOSGeom new;
    
    new = GEOSBoundary(self->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate boundary");
        return NULL;
    }
    
    return Geometry_New(new);
}

static PyObject *
Geometry_union(Geometry *self, PyObject *args)
{
    PyObject *g2;
    Geometry *g;
    GEOSGeom new;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;    
    new = GEOSUnion(self->geom, g->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate union");
        return NULL;
    }
    
    return Geometry_New(new);
}

static PyObject *
Geometry_pointOnSurface(Geometry *self)
{
    GEOSGeom new;
    
    new = GEOSPointOnSurface(self->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate boundary");
        return NULL;
    }
    
    return Geometry_New(new);
}

static PyObject *
Geometry_getCentroid(Geometry *self)
{
    GEOSGeom new;
    
    new = GEOSGetCentroid(self->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate centroid");
        return NULL;
    }
    
    return Geometry_New(new);
}

/* TODO: Polygonize and LineMerge? Geometry Info? */

/* === Binary Predicates === */

static PyObject *
Geometry_disjoint(Geometry *self, PyObject *args)
{
    char result;
    PyObject *g2;
    Geometry *g;

    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    result = GEOSDisjoint(self->geom, g->geom);
    if (result == 2)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate disjoint");
        return NULL;
    }
    return Py_BuildValue("i", (int)result);
}

static PyObject *
Geometry_touches(Geometry *self, PyObject *args)
{
    char result;
    PyObject *g2;
    Geometry *g;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    result = GEOSTouches(self->geom, g->geom);
    if (result == 2)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate touches");
        return NULL;
    }
    return Py_BuildValue("i", (int)result);
}

static PyObject *
Geometry_intersects(Geometry *self, PyObject *args)
{
    char result;
    PyObject *g2;
    Geometry *g;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    result = GEOSIntersects(self->geom, g->geom);
    if (result == 2)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate intersects");
        return NULL;
    }
    return Py_BuildValue("i", (int)result);
}

static PyObject *
Geometry_crosses(Geometry *self, PyObject *args)
{
    char result;
    PyObject *g2;
    Geometry *g;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    result = GEOSCrosses(self->geom, g->geom);
    if (result == 2)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate crosses");
        return NULL;
    }
    return Py_BuildValue("i", (int)result);}

static PyObject *
Geometry_within(Geometry *self, PyObject *args)
{
    char result;
    PyObject *g2;
    Geometry *g;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    result = GEOSWithin(self->geom, g->geom);
    if (result == 2)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate within");
        return NULL;
    }
    return Py_BuildValue("i", (int)result);
}

static PyObject *
Geometry_contains(Geometry *self, PyObject *args)
{
    char result;
    PyObject *g2;
    Geometry *g;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    result = GEOSContains(self->geom, g->geom);
    if (result == 2)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate contains");
        return NULL;
    }
    return Py_BuildValue("i", (int)result);
}

static PyObject *
Geometry_overlaps(Geometry *self, PyObject *args)
{
    char result;
    PyObject *g2;
    Geometry *g;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    result = GEOSOverlaps(self->geom, g->geom);
    if (result == 2)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate overlaps");
        return NULL;
    }
    return Py_BuildValue("i", (int)result);
}

static PyObject *
Geometry_equals(Geometry *self, PyObject *args)
{
    char result;
    PyObject *g2;
    Geometry *g;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    result = GEOSEquals(self->geom, g->geom);
    if (result == 2)
    {
        PyErr_SetString(PyExc_Exception, "Failed to evaluate equals");
        return NULL;
    }
    return Py_BuildValue("i", (int)result);
}

/* === Info === */

static PyObject *
Geometry_getDimensions(Geometry *self)
{
    return Py_BuildValue("i", GEOSGeom_getDimensions(self->geom));
}

/* === Other === */

#if defined(GEOS_CAPI_VERSION_MAJOR) && defined(GEOS_CAPI_VERSION_MINOR) && GEOS_CAPI_VERSION_MAJOR >= 1 && GEOS_CAPI_VERSION_MINOR >= 1

static PyObject *
Geometry_area(Geometry *self)
{
    double area;
    
    if (GEOSArea(self->geom, &area) != 1)
    {
        PyErr_SetString(PyExc_Exception, "Failed to calculate area");
        return NULL;
    }    
    return Py_BuildValue("d", area);
}

static PyObject *
Geometry_length(Geometry *self)
{
    double length;
    
    if (GEOSLength(self->geom, &length) != 1)
    {
        PyErr_SetString(PyExc_Exception, "Failed to calculate length");
        return NULL;
    }    
    return Py_BuildValue("d", length);
}
#endif /* C API version 1.0.2 */

static PyObject *
Geometry_distance(Geometry *self, PyObject *args)
{
    double distance;
    PyObject *g2;
    Geometry *g;
    
    if (!PyArg_ParseTuple(args, "O!", &GeometryType, &g2))
        return NULL;
    
    g = (Geometry *) g2;
    if (!GEOSDistance(self->geom, g->geom, &distance))
    {
        PyErr_SetString(PyExc_Exception, "Failed to calculate distance");
        return NULL;
    }    
    return Py_BuildValue("d", distance);
}

/* === geometry access === */

static PyObject *
geom_numGeometries(Geometry *self)
{
    if (GEOSGeomTypeId(self->geom) < 4)
        return Py_BuildValue("i", -1);
    else
        return Py_BuildValue("i", GEOSGetNumGeometries(self->geom));
}

static PyObject *
geom_getGeometryN(Geometry *self, PyObject *item)
{
    Geometry *g;
    GEOSGeom new;
    PyObject *geom;
    char *name;
    int num;
    long n;
    
    if (!PyInt_Check(item))
    {
		PyErr_SetString(PyExc_TypeError, "index must be an integer");
		return NULL;
	}
    
    num = GEOSGetNumGeometries(self->geom);
    if (num < 0)
    {
        name = (char *)GEOSGeomType(self->geom);
        PyErr_Format(PyExc_TypeError,
            "Geometry type %s cannot have children", name);
        return NULL;
    }

    if (num == 0)
    {
        PyErr_SetString(PyExc_Exception, "Geometry has no children");
        return NULL;
    }

    n = PyInt_AS_LONG(item);
    if (n < 0 || n > num)
    {
        PyErr_Format(PyExc_IndexError,
                "Index is beyond valid range 0 to %d", num);
        return NULL;
    }
    
    new = GEOSGetGeometryN(self->geom, (int)n);
    if (!new)
    {
        PyErr_Format(PyExc_Exception,
            "Failed to get child geometry at index %d", (int)n);
        return NULL;
    }
    
    geom = Geometry_New(new);
    g = (Geometry *) geom;
    g->owns = 0;
    return geom;
}

static PyObject *
geom_getExteriorRing(Geometry *self)
{
    Geometry *g;
    PyObject *geom;
    GEOSGeom new;
    
    /* Bail out immediately if not a polygon */
    if (GEOSGeomTypeId(self->geom) != GEOS_POLYGON)
    {
        PyErr_SetString(PyExc_TypeError,
                "Only polygon geometries have interior rings");
        return NULL;
    }

    new = GEOSGetExteriorRing(self->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to get exterior ring");
        return NULL;
    }

    geom = Geometry_New(new);
    g = (Geometry *) geom;
    g->owns = 0;
    return geom;    
}

static PyObject *
geom_numInteriorRings(Geometry *self)
{
    if (GEOSGeomTypeId(self->geom) != GEOS_POLYGON)
    {
        PyErr_SetString(PyExc_TypeError,
                "Only polygon geometries have interior rings");
        return NULL;
    }
    return Py_BuildValue("i", GEOSGetNumInteriorRings(self->geom));
}

static PyObject *
geom_getInteriorRingN(Geometry *self, PyObject *item)
{
    Geometry *g;
    PyObject *geom;
    GEOSGeom new;
    int num;
    long n;
    
    if (!PyInt_Check(item))
    {
		PyErr_SetString(PyExc_TypeError, "index must be an integer");
		return NULL;
	}
    
    if (GEOSGeomTypeId(self->geom) != GEOS_POLYGON)
    {
        PyErr_SetString(PyExc_TypeError,
                "Only polygon geometries have interior rings");
        return NULL;
    }

    num = GEOSGetNumInteriorRings(self->geom);
    if (num == 0)
    {
        PyErr_SetString(PyExc_Exception, "Geometry has no interior rings");
        return NULL;
    }

    n = PyInt_AS_LONG(item);
    if (n < 0 || n >= num)
    {
        PyErr_Format(PyExc_IndexError,
                "Index is beyond valid range 0 to %d", num-1);
        return NULL;
    }
    
    new = GEOSGetInteriorRingN(self->geom, (int)n);
    if (!new)
    {
        PyErr_Format(PyExc_Exception,
            "Failed to get interior ring at index %d", (int)n);
        return NULL;
    }
    
    geom = Geometry_New(new);
    g = (Geometry *) geom;
    g->owns = 0;
    return geom;    
}

static PyObject *
geom_getCoordSeq(Geometry *self)
{
    CoordSeq *s;
    PyObject *seq;
    GEOSCoordSeq new;
    
    new = GEOSGeom_getCoordSeq(self->geom);
    if (!new)
    {
        PyErr_SetString(PyExc_Exception, "Failed to get coordinate sequence");
        return NULL;
    }
    
    seq = CoordSeq_New(new);
    s = (CoordSeq *) seq;
    s->owns = 0;
    return seq;
}

/* === Transform === */

static int
transform_coordseq(GEOSCoordSeq s, projPJ *dst, projPJ *src)
{
    int error, i;
    unsigned int size;
    double *x, *y, *z;
    
    GEOSCoordSeq_getSize(s, &size);
            
    x = (double *) malloc(size*sizeof(double));
    y = (double *) malloc(size*sizeof(double));
    z = (double *) malloc(size*sizeof(double));
            
    /* Prep arrays */
    for (i=0; i<size; i++)
    {
        GEOSCoordSeq_getX(s, i, x+i);
        GEOSCoordSeq_getY(s, i, y+i);
        GEOSCoordSeq_getZ(s, i, z+i);
        //z[i] = 0.0;
        
        if (pj_is_latlong(src))
        {
            *(x+i) *= DEG_TO_RAD;
            *(y+i) *= DEG_TO_RAD;
        }
    } 

    /* Project, catching errors */
    error = pj_transform(src, dst, size, 0, x, y, z);
    if (error)
        return error;

    /* Update the coordinate sequence */
    for (i=0; i<size; i++)
    {
        if (pj_is_latlong(dst))
        {
            *(x+i) *= RAD_TO_DEG;
            *(y+i) *= RAD_TO_DEG;
        }
        GEOSCoordSeq_setX(s, i, *(x+i));
        GEOSCoordSeq_setY(s, i, *(y+i));
        GEOSCoordSeq_setZ(s, i, *(z+i));
    } 

    /* free arrays */
    free(x);
    free(y);
    free(z);
    
    return 0;
}

static PyObject *
geom_transform(Geometry *self, PyObject *args, PyObject *kwds)
{
    GEOSCoordSeq s;
    GEOSGeom g, shell;
    GEOSGeom hole;
    PyObject *transform=NULL;
    PyObject *tmp;
    char *dst_proj=NULL;
    char *src_proj=NULL;
    projPJ *dst, *src;
    int gtype, error=0, num=0, nholes=0, j, k;
   
    static char *kwlist[] = {"transform", "dst_proj", "src_proj", NULL};
    
    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|Oss", kwlist,
            &transform, &dst_proj, &src_proj))
        return NULL;
        
    /* If we have a ProjTransform, get proj_dst and proj_src */
    if (transform)
    {
        tmp = PyObject_GetAttrString(transform, "dst_proj");
        dst_proj = PyString_AsString(tmp);
        tmp = PyObject_GetAttrString(transform, "src_proj");
        src_proj = PyString_AsString(tmp);
    }
   
    /* Raise exception if we don't have enough info */
    if (!dst_proj || !src_proj)
    {
        PyErr_SetString(PyExc_ValueError, "Destination and source projections must be specified, either through a ProjTransform instance or dst_proj and src_proj strings");
        return NULL;
    }
    
    /* Make PJ from destination and source */
    dst = pj_init_plus(dst_proj);
    src = pj_init_plus(src_proj);
    
    /* Check geometries now, and treat as valid afterwards */
    if (!GEOSisValid(self->geom))
        return NULL;
    
    gtype = GEOSGeomTypeId(self->geom);
    
    switch (gtype)
    {
        case GEOS_POINT:
            s = GEOSGeom_getCoordSeq(self->geom);
            error = transform_coordseq(s, dst, src);
            break;
        case GEOS_LINESTRING:
            s = GEOSGeom_getCoordSeq(self->geom);
            error = transform_coordseq(s, dst, src);
            break;
        case GEOS_LINEARRING:
            s = GEOSGeom_getCoordSeq(self->geom);
            error = transform_coordseq(s, dst, src);
            break;
        case GEOS_POLYGON:
            shell = GEOSGetExteriorRing(self->geom);
            s = GEOSGeom_getCoordSeq(shell);
            error = transform_coordseq(s, dst, src);
            
            nholes = GEOSGetNumInteriorRings(self->geom);
            for (j=0; j<nholes; j++)
            {
                hole = GEOSGetInteriorRingN(self->geom, j);
                s = GEOSGeom_getCoordSeq(hole);
                error = transform_coordseq(s, dst, src);
            }
            break;
        case GEOS_MULTIPOINT:
            num = GEOSGetNumGeometries(self->geom);
            for (k=0; k<num; k++)
            {
                g = GEOSGetGeometryN(self->geom, k);
                s = GEOSGeom_getCoordSeq(g);
                error = transform_coordseq(s, dst, src);
            }
            break;
        case GEOS_MULTILINESTRING:
            num = GEOSGetNumGeometries(self->geom);
            for (k=0; k<num; k++)
            {
                g = GEOSGetGeometryN(self->geom, k);
                s = GEOSGeom_getCoordSeq(g);
                error = transform_coordseq(s, dst, src);
            }
            break;
        case GEOS_MULTIPOLYGON:
            num = GEOSGetNumGeometries(self->geom);
            for (k=0; k<num; k++)
            {
                g = GEOSGetGeometryN(self->geom, k);
                shell = GEOSGetExteriorRing(g);
                s = GEOSGeom_getCoordSeq(shell);
                error = transform_coordseq(s, dst, src);
                nholes = GEOSGetNumInteriorRings(g);
                for (j=0; j<nholes; j++)
                {
                    hole = GEOSGetInteriorRingN(g, j);
                    s = GEOSGeom_getCoordSeq(hole);
                    error = transform_coordseq(s, dst, src);
                }
            }
            break;
        case GEOS_GEOMETRYCOLLECTION:
            num = GEOSGetNumGeometries(self->geom);
            for (k=0; k<num; k++)
            {
                g = GEOSGetGeometryN(self->geom, k);
                if (GEOSGeomTypeId(g) < 3)
                {
                    s = GEOSGeom_getCoordSeq(g);
                    error = transform_coordseq(s, dst, src);
                }
                else
                {
                    shell = GEOSGetExteriorRing(g);
                    s = GEOSGeom_getCoordSeq(shell);
                    error = transform_coordseq(s, dst, src);
                    nholes = GEOSGetNumInteriorRings(g);
                    for (j=0; j<nholes; j++)
                    {
                        hole = GEOSGetInteriorRingN(g, j);
                        s = GEOSGeom_getCoordSeq(hole);
                        error = transform_coordseq(s, dst, src);
                    }
                }
            }
            break;
    }
            
    if (error)
    {
        pj_free(dst);
        pj_free(src);
        PyErr_Format(PyExc_Exception,
            "Failed to transform coordinate arrays, PJ error %d", error);
        return NULL;
    }
    
    pj_free(dst);
    pj_free(src);

    Py_INCREF(Py_None);
    return Py_None;
}

/* === Define Methods === */

static PyMethodDef module_methods[] = {
    {"getTypeId", (PyCFunction)Geometry_getTypeId, METH_NOARGS,
        "Returns GEOS geometry type id"},
    {"toWKB", (PyCFunction)Geometry_toWKB, METH_NOARGS,
     "Return Well-known binary representation"},
    {"toWKT", (PyCFunction)to_wkt, METH_NOARGS,
     "Return Well-known binary representation"},
    {"isEmpty", (PyCFunction)Geometry_isEmpty, METH_NOARGS, "See GEOS"},
    {"isValid", (PyCFunction)Geometry_isValid, METH_NOARGS, "See GEOS"},
    {"isSimple", (PyCFunction)Geometry_isSimple, METH_NOARGS, "See GEOS"},
    {"isRing", (PyCFunction)Geometry_isRing, METH_NOARGS, "See GEOS"},
    {"hasZ", (PyCFunction)Geometry_hasZ, METH_NOARGS, "See GEOS"},
    {"intersection", (PyCFunction)Geometry_intersection, METH_VARARGS, "See GEOS"},
    {"buffer", (PyCFunction)Geometry_buffer, METH_VARARGS, "See GEOS"},
    {"convexHull", (PyCFunction)Geometry_convexHull, METH_VARARGS, "See GEOS"},
    {"difference", (PyCFunction)Geometry_difference, METH_VARARGS, "See GEOS"},
    {"symDifference", (PyCFunction)Geometry_symDifference, METH_VARARGS, "See GEOS"},
    {"boundary", (PyCFunction)Geometry_boundary, METH_VARARGS, "See GEOS"},
    {"union", (PyCFunction)Geometry_union, METH_VARARGS, "See GEOS"},
    {"pointOnSurface", (PyCFunction)Geometry_pointOnSurface, METH_VARARGS, "See GEOS"},
    {"getCentroid", (PyCFunction)Geometry_getCentroid, METH_VARARGS, "See GEOS"},
    {"disjoint", (PyCFunction)Geometry_disjoint, METH_VARARGS, "See GEOS"},
    {"touches", (PyCFunction)Geometry_touches, METH_VARARGS, "See GEOS"},
    {"intersects", (PyCFunction)Geometry_intersects, METH_VARARGS, "See GEOS"},
    {"crosses", (PyCFunction)Geometry_crosses, METH_VARARGS, "See GEOS"},
    {"within", (PyCFunction)Geometry_within, METH_VARARGS, "See GEOS"},
    {"contains", (PyCFunction)Geometry_contains, METH_VARARGS, "See GEOS"},
    {"overlaps", (PyCFunction)Geometry_overlaps, METH_VARARGS, "See GEOS"},
    {"equals", (PyCFunction)Geometry_equals, METH_VARARGS, "See GEOS"},
    {"_transform", (PyCFunction)geom_transform, METH_VARARGS | METH_KEYWORDS, "Transform coordinates between spatial reference systems"},
    {"equals", (PyCFunction)Geometry_equals, METH_VARARGS, "See GEOS"},
    {"getDimensions", (PyCFunction)Geometry_getDimensions, METH_NOARGS,
     "Returns number of coordinate dimensions"},
    {"envelope", (PyCFunction)Geometry_envelope, METH_VARARGS, "See GEOS"},
#if defined(GEOS_CAPI_VERSION_MAJOR) && defined(GEOS_CAPI_VERSION_MINOR) && GEOS_CAPI_VERSION_MAJOR >= 1 && GEOS_CAPI_VERSION_MINOR >= 1
    {"getGeomType", (PyCFunction)Geometry_getGeomType, METH_NOARGS,
        "Returns GEOS geometry type name"},
    {"area", (PyCFunction)Geometry_area, METH_NOARGS,
     "Returns area of geometry in its own coordinate units"}, 
    {"length", (PyCFunction)Geometry_length, METH_NOARGS,
     "Returns length or perimeter of geometry in its own coordinate units"},
#endif
    {"distance", (PyCFunction)Geometry_distance, METH_VARARGS,
     "Returns distance between geometries in own coordinate units"}, 
    {"numGeometries", (PyCFunction)geom_numGeometries, METH_NOARGS,
     "Returns the number of child geometries"},
    {"getGeometryN", (PyCFunction)geom_getGeometryN, METH_O,
     "Returns the Nth child geometry from the collection"},
    {"getExteriorRing", (PyCFunction)geom_getExteriorRing, METH_NOARGS,
     "Returns the exterior ring of a Polygon geometry"},
    {"numInteriorRings", (PyCFunction)geom_numInteriorRings, METH_NOARGS,
     "Returns the number of interior rings in a Polygon geometry"},
    {"getInteriorRingN", (PyCFunction)geom_getInteriorRingN, METH_O,
     "Returns the Nth interior ring of a Polygon geometry"},
    {"getCoordSeq", (PyCFunction)geom_getCoordSeq, METH_NOARGS,
     "Returns the coordinate sequence of the geometry"},
    {"createCoordSeq", (PyCFunction)_createCoordSeq, METH_VARARGS, "Return a GEOSCoordinateSequence"},
    {"createGeomFromWKB", (PyCFunction)_createGeomFromWKB, METH_VARARGS, "Return a GEOSGeometry"},
    {"createGeomFromWKT", (PyCFunction)_createGeomFromWKT, METH_VARARGS, "Return a GEOSGeometry"},
    {"_updateGeomFromWKB", (PyCFunction)_updateGeomFromWKB, METH_VARARGS, "Return a GEOSGeometry"},
    {"_updateGeomFromWKT", (PyCFunction)_updateGeomFromWKT, METH_VARARGS, "Return a GEOSGeometry"},
    {"createGeomFromCoordSeq", (PyCFunction)_createGeomFromCoordSeq, METH_VARARGS, "Return a GEOSGeometry"},
    {"clone", (PyCFunction)Geometry_clone, METH_NOARGS, "Return a clone"},
    {"_copyFromCoordSeq", (PyCFunction)Geometry_copyFromCoordSeq, METH_VARARGS, "Return a GEOSGeometry"},
    {"_copyFromGeom", (PyCFunction)Geometry_copyFromGeom, METH_VARARGS, "Return a GEOSGeometry"},
    {"createPolygon", (PyCFunction)_createPolygon, METH_VARARGS, "Return a GEOSGeometry"},
    {"createCollection", (PyCFunction)_createCollection, METH_VARARGS, "Return a GEOSGeometry"},
    {NULL}
};

/* === Define Type === */

static PyTypeObject GeometryType = {
    PyObject_HEAD_INIT(NULL)
    0,                              /*ob_size*/
    "GEOSGeometry",               /*tp_name*/
    sizeof(Geometry),               /*tp_basicsize*/
    0,                              /*tp_itemsize*/
    (destructor)Geometry_dealloc,   /*tp_dealloc*/
    0,                              /*tp_print*/
    0,                              /*tp_getattr*/
    0,                              /*tp_setattr*/
    0,                              /*tp_compare*/
    0,                              /*tp_repr*/
    0,                              /*tp_as_number*/
    0,                              /*tp_as_sequence*/
    0,                              /*tp_as_mapping*/
    0,                              /*tp_hash */
    0,                              /*tp_call*/
    0,                              /*tp_str*/
    0,                              /*tp_getattro*/
    0,                              /*tp_setattro*/
    0,                              /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Geometry objects",           /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    module_methods,             /* tp_methods */
    0, /*Noddy_members,              tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    0,      /* tp_init */
    0,                         /* tp_alloc */
    PyType_GenericNew,                 /* tp_new */
};

/*
==============================================================================
 Initialization
==============================================================================
*/

void
notice_handler(const char *fmt, ...) {
    /* TODO: decide whether we want notices to stderr */
	va_list ap;
    fprintf(stderr, "NOTICE: ");
    va_start (ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, "\n" );
}

void
error_handler(const char *fmt, ...) {
	va_list ap;
    va_start (ap, fmt);
    va_end(ap);
    PyErr_Format(GEOSError, fmt, ap);
}

#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
init_geom(void) 
{
    PyObject* m;

    if (PyType_Ready(&GeometryType) < 0)
        return;
    if (PyType_Ready(&CoordinateSequenceType) < 0)
        return;

    m = Py_InitModule3("_geom", module_methods,
                       "New geom module based on GEOS");

    if (m == NULL)
      return;

    Py_INCREF(&GeometryType);
    PyModule_AddObject(m, "GEOSGeometry", (PyObject *)&GeometryType);
    Py_INCREF(&CoordinateSequenceType);
    PyModule_AddObject(m, "GEOSCoordinateSequence",
                       (PyObject *)&CoordinateSequenceType);

    PyModule_AddIntConstant(m, "GEOS_POINT", 0);
    PyModule_AddIntConstant(m, "GEOS_LINESTRING", 1);
    PyModule_AddIntConstant(m, "GEOS_LINEARRING", 2);
    PyModule_AddIntConstant(m, "GEOS_POLYGON", 3);
    PyModule_AddIntConstant(m, "GEOS_MULTIPOINT", 4);
    PyModule_AddIntConstant(m, "GEOS_MULTILINESTRING", 5);
    PyModule_AddIntConstant(m, "GEOS_MULTIPOLYGON", 6);
    PyModule_AddIntConstant(m, "GEOS_GEOMETRYCOLLECTION", 7);
    
    GEOSError = PyErr_NewException("_geom.GEOSError", NULL, NULL);
    Py_INCREF(GEOSError);
    PyModule_AddObject(m, "GEOSError", GEOSError);
    
    initGEOS(notice_handler, error_handler);
    Py_AtExit(finishGEOS);
}

