# =============================================================================
# Python Cartographic Library (C) 2006 Sean C. Gillies, Howard Butler
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com, hobu@iastate.edu
#
# This geometry module implements the Open Geospatial Consortium's Simple
# Features Access specification:
#
#   http://portal.opengeospatial.org/files/?artifact_id=13227
#
# The Simple Features Access specification is Copyright (C) 2005 Open
# Geospatial Consortium, Inc. All Rights Reserved.
# =============================================================================

__docformat__ = 'restructuredtext en'

import struct

from zope.interface import implements

from cartography.geometry.interfaces import IGeometry, IPoint, ILineString, \
    ILinearRing, IPolygon, IGeometryCollection, IMultiPoint, IMultiLineString, \
    IMultiPolygon, IBoundingBox

from cartography.proj import SpatialReference
from cartography.geometry._geom import GEOSGeometry, \
    createGeomFromWKB, createGeomFromWKT, createGeomFromCoordSeq, \
    createCoordSeq, createCollection, createPolygon , \
    GEOS_POINT, GEOS_LINESTRING, GEOS_LINEARRING, GEOS_POLYGON, \
    GEOS_MULTIPOINT, GEOS_MULTILINESTRING, GEOS_MULTIPOLYGON, \
    GEOSError


class SRSError(Exception):

    """An exception concerning the spatial reference systems of geometries.
    """


class Geometry(GEOSGeometry):

    """Implements IGeometry.

    Attributes
    ----------
    srs : SpatialReference
        Spatial reference system
    """
    
    implements(IGeometry)
   
    def __init__(self):
        """Initialize a geometry."""
        GEOSGeometry.__init__(self)
        self._updateGeomFromWKT('GEOMETRYCOLLECTION EMPTY')
        self.srs = None
        
    def __str__(self):
        return self.toWKT()        
               
    # __reduce__ and __setstate__ are for support of pickling. See ticket #71
    def __reduce__(self):
        return (self.__class__, (), self.toWKB())

    def __setstate__(self, state):
        self._updateGeomFromWKB(state)

    def fromWKT(cls, wkt, srs=None):
        """Create and initialize a geometry from a well-known text string.

        Parameters
        ----------
        wkt : string
            WKT string like 'POINT (-105.5 40.0)'
        srs : SpatialReference
            The spatial reference system of the geometry
        """
        g = Geometry()
        g._updateGeomFromWKT(wkt)
        g.__class__ = CLASS_TYPE[g.getTypeId()]
        g.srs = srs
        return g
    fromWKT = classmethod(fromWKT)
    
    def fromWKB(cls, wkb, srs=None):
        """Create and initialize a geometry from a well-known binary string.

        Parameters
        ----------
        wkb : string
            String of bytes like struct.pack('>bldd', 1, 1, -105.5, 40.0)
        srs : SpatialReference
            The spatial reference system of the geometry
        """
        g = Geometry()
        g._updateGeomFromWKB(wkb)
        g.__class__ = CLASS_TYPE[g.getTypeId()]
        g.srs = srs
        return g
    fromWKB = classmethod(fromWKB)

    def envelope(self):
        """Returns minimum bounding rectangle as a polygon."""
        x = GEOSGeometry.envelope(self)
        gtypid = x.getTypeId()
        if gtypid == GEOS_POINT:
            seq = x.getCoordSeq()
            return BoundingBox(seq.getX(0), seq.getY(0),
                               None, None, self.srs)
        elif gtypid == GEOS_POLYGON:
            ring = x.getExteriorRing()
            seq = ring.getCoordSeq()
            return BoundingBox(seq.getX(0), seq.getY(0),
                               seq.getX(2), seq.getY(2), self.srs)
        else:
            raise TypeError, "envelope() returned other than Point or Polygon"

    def transform(self, transform=None, dst_proj='', src_proj=''):
        """Transform geometry in-place."""
        if transform:
            self._transform(transform)
        else:
            self._transform(dst_proj=dst_proj, src_proj=src_proj)
        
    def _hasSameSRS(self, o):
        # return True if geometries have the same SRS, else False
        try:
            assert self.srs == o.srs or self.srs.equals(o.srs)
            return True
        except:
            return False
            
    def _binary_geometry_operation(self, op, o):
        # wrapper for GEOSGeometry operations
        # fail immediately if geometries have different srs
        if self._hasSameSRS(o):
            f = getattr(GEOSGeometry, op)
            x = f(self, o)
            g = CLASS_TYPE[x.getTypeId()].__call__()
            g._copyFromGeom(x)
            return g
        else:
            raise SRSError, \
                "Operations can not be made between geometries of different spatial reference systems"

    def _unary_geometry_operation(self, op, *args):
        # wrapper for GEOSGeometry operations
        f = getattr(GEOSGeometry, op)
        x = f(self, *args)
        g = CLASS_TYPE[x.getTypeId()].__call__()
        g._copyFromGeom(x)
        g.srs = self.srs
        return g

    def boundary(self):
        """Return boundary as a Python Geometry."""
        return self._unary_geometry_operation('boundary')
       
    def buffer(self, distance, n):
        """Return buffer as a Python Geometry."""
        return self._unary_geometry_operation('buffer', distance, n)

    def convexHull(self):
        """Return intersection as a Python Geometry."""
        return self._unary_geometry_operation('convexHull')

    def pointOnSurface(self):
        """Return a point on surface."""
        return self._unary_geometry_operation('pointOnSurface')

    def centroid(self):
        """Return a point on surface."""
        return self._unary_geometry_operation('getCentroid')
        
    def intersection(self, o):
        """Return intersection as a Python Geometry."""
        return self._binary_geometry_operation('intersection', o)
       
    def union(self, o):
        """Return union as a Python Geometry."""
        return self._binary_geometry_operation('union', o)
       
    def difference(self, o):
        """Return difference as a Python Geometry."""
        return self._binary_geometry_operation('difference', o)
       
    def symDifference(self, o):
        """Return symmetric difference as a Python Geometry."""
        return self._binary_geometry_operation('symDifference', o)
       
    def copy(self):
        """Return a new copy of this geometry."""
        return Geometry.fromWKB(self.toWKB(), self.srs)

    def asText(self):
        """Return the geometry as a well-known text string."""
        return self.toWKT()
        
    def asBinary(self):
        """Return the geometry as a well-known binary byte string."""
        return self.toWKB()
        
    def geometryType(self):
        """Return geometry type as a string."""
        return self.geometrytype


class Point(Geometry):

    """Implements IPoint.

    Attributes
    ----------
    x, y, z : float
        Coordinate values.
    """
    
    implements(IPoint)
    geometrytype = 'Point'
    
    def __init__(self, x=None, y=None, z=0.0, srs=None):
        """Initialize a point.
        
        Parameters
        ----------
        x : float
            easting
        y : float
            northing
        z : float
            elevation
        srs : SpatialReference
            Spatial reference system of the coordinate values.
        """
        Geometry.__init__(self)
        self.srs = srs
        if (x is not None and y is not None): 
            s = createCoordSeq(1)
            s.setX(0, float(x))
            s.setY(0, float(y))
            s.setZ(0, float(z))
            self._copyFromCoordSeq(s, GEOS_POINT)
        
    def getX(self):
        s = self.getCoordSeq()
        return s.getX(0)
    def setX(self, x):
        s = self.getCoordSeq()
        s.setX(0, float(x))
    x = property(getX, setX)
    
    def getY(self):
        s = self.getCoordSeq()
        return s.getY(0)
    def setY(self, y):
        s = self.getCoordSeq()
        s.setY(0, float(y))
    y = property(getY, setY)

    def getZ(self):
        s = self.getCoordSeq()
        return s.getZ(0)
    def setZ(self, z):
        s = self.getCoordSeq()
        s.setZ(0, float(z))
    z = property(getZ, setZ)


class LineString(Geometry):

    """Implements ILineString.
    """
    
    implements(ILineString)
    geometrytype = 'LineString'
    
    def __init__(self, points=[], srs=None):
        """Initialize a LineString.
        
        Parameters
        ----------
        points : list
            list of Point objects
        srs : SpatialReference
            Spatial reference system of the coordinate values.
        """
        Geometry.__init__(self)
        self.srs = srs
        if points:
            s = createCoordSeq(len(points))
            for i in range(len(points)):
                p = points[i]
                s.setX(i, p.x)
                s.setY(i, p.y)
                s.setZ(i, p.z)
            self._copyFromCoordSeq(s, GEOS_LINESTRING)

    def numPoints(self):
        """Return number of points."""
        return self.getCoordSeq().getSize()
        
    def pointN(self, N):
        """Return the Nth point, starting at 0."""
        s = self.getCoordSeq()
        return Point(s.getX(N), s.getY(N), s.getZ(N), self.srs)
        
    def length(self):
        """Return the length."""
        length = 0.0
        for i in range(1, self.numPoints()):
            current = self.pointN(i)
            previous = self.pointN(i-1)
            length = length + current.distance(previous)
        return length
    
    def startPoint(self):
        """Return the first point."""
        return self.pointN(0)
    
    def endPoint(self):
        """Return the last point."""
        return self.pointN(self.numPoints()-1)
    
    def isClosed(self):
        """Return True if the linestring is closed, else False."""
        return self.startPoint().equals(self.endPoint())


class LinearRing(LineString):

    """Implements ILinearRing.
    """
    
    implements(ILinearRing)
    geometrytype = 'LinearRing'
    
    def __init__(self, points=[], srs=None):
        """Initialize a LineString.
        
        Parameters
        ----------
        points : list
            list of Point objects
        srs : SpatialReference
            Spatial reference system of the coordinate values.
        """
        Geometry.__init__(self)
        self.srs = srs
        if points:
            s = createCoordSeq(len(points))
            for i in range(len(points)):
                p = points[i]
                s.setX(i, p.x)
                s.setY(i, p.y)
                s.setZ(i, p.z)
            self._copyFromCoordSeq(s, GEOS_LINEARRING)

      
class Polygon(Geometry):

    """Implements IPolygon.
    """
    
    implements(IPolygon)
    geometrytype = 'Polygon'

    def __init__(self, shell=None, holes=[], srs=None):
        """Initialize a Polygon.

        Parameters
        ----------
        shell : LinearRing
            Exterior ring
        holes : list
            List of interior rings
        srs : SpatialReference
            Spatial reference system of the coordinate values.
        """
        Geometry.__init__(self)
        self.srs = srs
        if shell:
            if not shell.isRing():
                raise ValueError, "Shell is not a ring"
            if not holes:
                x = createPolygon(shell)
            else:
                x = createPolygon(shell, holes)
            self._copyFromGeom(x)

    def exteriorRing(self):
        """Return a point on surface."""
        return self._unary_geometry_operation('getExteriorRing')

    def interiorRingN(self, N):
        """Return a point on surface."""
        return self._unary_geometry_operation('getInteriorRingN', N)
    

class GeometryCollection(Geometry):

    """Implements IGeometryCollection.
    """
    
    implements(IGeometryCollection)
    geometrytype = 'GeometryCollection'
    
    def geometryN(self, N):
        """Returns the Nth geometry in this GeometryCollection."""
        x = self.getGeometryN(N)
        g = CLASS_TYPE[x.getTypeId()].__call__()
        g._copyFromGeom(x)
        g.srs = self.srs
        return g


class MultiPoint(GeometryCollection):
    
    """Implements IMultiPoint.
    """

    implements(IMultiPoint)
    geometrytype = 'MultiPoint'
    
    def __init__(self, points=[], srs=None):
        """Initialize.
        
        Parameters
        ----------
        points : list
            List of Points
        srs : SpatialReference
            Spatial reference system of this collection.
        """
        Geometry.__init__(self)
        self.srs = srs
        if points:
            x = createCollection(GEOS_MULTIPOINT, points)
            self._copyFromGeom(x)
  

class MultiLineString(GeometryCollection):

    """Implements IMultiLineString.
    """

    implements(IMultiLineString)
    geometrytype = 'MultiLineString'
    
    def __init__(self, lines=[], srs=None):
        """Initialize.

        Parameters
        ----------
        lines : list
            list of LineStrings
        srs : SpatialReference
            Spatial reference system of this collection.
        """
        Geometry.__init__(self)
        self.srs = srs
        if lines:
            x = createCollection(GEOS_MULTILINESTRING, lines)
            self._copyFromGeom(x)


class MultiPolygon(GeometryCollection):

    """Implements IMultiPolygon.
    """

    implements(IMultiPolygon)
    geometrytype = 'MultiPolygon'
    
    def __init__(self, polygons=[], srs=None):
        """Initialize.

        Parameters
        ----------
        polygons : list
            list of Polygons
        srs : SpatialReference
            Spatial reference system of this collection
        """
        Geometry.__init__(self)
        self.srs = srs
        if polygons:
            x = createCollection(GEOS_MULTIPOLYGON, polygons)
            self._copyFromGeom(x)


class BoundingBox(Polygon):

    """A Polygon defined by four minimum and maximum values.

    Attributes
    ----------
    minx, miny, maxx, maxy : float
        Bounding coordinate values of the box
    srs : SpatialReference
        Spatial reference system
    dimension : int
        0: point-like bbox, 2: polygon bbox
    """
    
    implements(IBoundingBox)
    geometrytype = 'Polygon'

    def __init__(self, minx=None, miny=None, maxx=None, maxy=None, srs=None):
        """Initialize a new bounding box."""
        if (minx is not None and miny is None) \
        or (miny is not None and minx is None):
            raise ValueError, \
                "Invalid bounding values: minx and maxx must be specified"
        if maxx is not None and maxx < minx:
            raise ValueError, \
                "Invalid bounding values: minx > maxx"
        if maxy is not None and maxy < miny:
            raise ValueError, \
                "Invalid bounding values: miny > maxy"
        self._minx = minx
        self._miny = miny
        self._maxx = maxx
        self._maxy = maxy
 
        Geometry.__init__(self)
        self.srs = srs
        try:
            self._update_geom()
        except:
            pass
        
    def _update_geom(self):
        if self._maxx is not None or self._maxy is not None:
            wkb = struct.pack('>bllldddddddddd', 0, 3, 1, 5,
            self.minx, self.miny,
            self.minx, self.maxy,
            self.maxx, self.maxy,
            self.maxx, self.miny,
            self.minx, self.miny)
        else:
            wkb = struct.pack('>bldd', 0, 1, self.minx, self.miny)
        x = createGeomFromWKB(wkb)
        self._copyFromGeom(x)

    def copy(self):
        """Return a copy of this bounding box."""
        # Overrides Geometry.copy()
        return BoundingBox(self._minx, self._miny, self._maxx, self._maxy,
                           self.srs)
       
    def _get_dimension(self):
        if self._maxx is None and self._maxy is None:
            return 0
        else:
            return 2
    dimension = property(_get_dimension)
    
    # Bounding value properties
    def _get_minx(self):
        return self._minx
    def _set_minx(self, x):
        self._minx = float(x)
        self._update_geom()
    minx = property(_get_minx, _set_minx)
    
    def _get_miny(self):
        return self._miny
    def _set_miny(self, y):
        self._miny = float(y)
        self._update_geom()
    miny = property(_get_miny, _set_miny)
    
    def _get_maxx(self):
        if self._maxx is None:
            return self._minx
        return self._maxx
    def _set_maxx(self, x):
        self._maxx = float(x)
        if self._maxy is None: self._maxy = self._miny
        self._update_geom()
    maxx = property(_get_maxx, _set_maxx)
    
    def _get_maxy(self):
        if self._maxy is None:
            return self._miny
        return self._maxy
    def _set_maxy(self, y):
        self._maxy = float(y)
        if self._maxx is None: self._maxx = self._minx
        self._update_geom()
    maxy = property(_get_maxy, _set_maxy)

    # Bounding box size
    def _get_width(self):
        """Returns the width of the bounding box."""
        return abs(self.maxx - self.minx)
    width = property(_get_width)
    
    def _get_height(self):
        """Returns the height of the bounding box."""
        return abs(self.maxy - self.miny)
    height = property(_get_height)

    def totuple(self):
        """Return bounding box values as (minx,miny,maxx,may) tuple"""
        return (self.minx, self.miny, self.maxx, self.maxy)

    def transform(self, transform=None, dst_proj='', src_proj=''):
        """Transform geometry in-place."""
        if transform:
            self._transform(transform)
        else:
            self._transform(dst_proj=dst_proj, src_proj=src_proj)
        ring = self.getExteriorRing()
        seq = ring.getCoordSeq()
        self._minx = seq.getX(0)
        self._miny = seq.getY(0)
        self._maxx = seq.getX(2)
        self._maxy = seq.getY(2)
        
        
CLASS_TYPE = [Point, LineString, LinearRing, Polygon, MultiPoint,
              MultiLineString, MultiPolygon, GeometryCollection]



