# =============================================================================
# Python Cartographic Library (C) 2006 Sean C. Gillies, and Howard Butler
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com, hobu@iastate.edu
#
# $Id: interfaces.py 544 2006-10-20 20:47:35Z seang $
# =============================================================================

"""
See the Open Geospatial Consortium's Simple Features Access specification

  http://portal.opengeospatial.org/files/?artifact_id=13227

All IOGCSF* interfaces shall follow the details of specification 1.1.0 with the
exception that, to follow Python convention, method names begin with lower case
characters.

The Simple Features Access specification is Copyright (C) 2005 Open Geospatial
Consortium, Inc. All Rights Reserved.
"""

from zope.interface import Interface, Attribute


class IOGCSFGeometry(Interface):

    """Two-dimensional, topologically-closed geometry.
    """

    # Basic Methods
    # -------------
    
    def dimension():
        """Returns the inherent dimension of this geometry (int)."""

    def geometryType():
        """Returns name of the geometry type (string).

        Example: 'Polygon'
        """

    def envelope():
        """Returns minimum bounding box as a polygon (Polygon)."""

    def asText():
        """Exports to a well-known text representation (string)."""

    def asBinary():
        """Exports to a well-known binary representation (binary)."""

    def isEmpty():
        """Returns 1 if this geometry is the empty point set, else 0 (int)."""

    def isSimple():
        """Returns 1 if this geometry has no anomalies, such as self-
        intersection, else 0 (int)."""

    def boundary():
        """Returns the boundary (Geometry)."""

    def copy():
        """Returns a copy of the geometry (Geometry)."""

    # Spatial Relations
    # -----------------

    def equals(anotherGeometry):
        """Returns 1 if this geometry is spatially equal to another,
        else 0 (int)."""
        
    def disjoint(anotherGeometry):
        """Returns 1 if this geometry is spatially disjoint from another,
        else 0 (int)."""
        
    def intersects(anotherGeometry):
        """Returns 1 if this geometry intersects another, else 0 (int)."""

    def touches(anotherGeometry):
        """Returns 1 if this geometry touches another, else 0 (int)."""

    def crosses(anotherGeometry):
        """Returns 1 if this geometry crosses another, else 0 (int)."""

    def within(anotherGeometry):
        """Returns 1 if this geometry is within another, else 0 (int)."""

    def contains(anotherGeometry):
        """Returns 1 if this geometry contains another, else 0 (int)."""

    def overlaps(anotherGeometry):
        """Returns 1 if this geometry overlaps another, else 0 (int)."""

    def relate(anotherGeometry, intersectionPatternMatrix):
        """Returns 1 if this geometry relates to another as specified by
        the intersection pattern matrix (string), else 0 (int)."""

    # Spatial Analysis
    # ----------------

    def distance(anotherGeometry):
        """Returns the shortest distance between two geometries in the
        coordinate reference system of this geometry (float)."""

    def buffer(distance):
        """Returns a geometry representing all points whose distance away
        from this geometry is less than or equal to distance (float), 
        measured in the coordinate reference system of this geometry
        (Geometry)."""
        
    def convexHull():
        """Returns a geometry that represents the convex hull of this 
        geometry (Geometry)."""

    def intersection(anotherGeometry):
        """Returns a geometry that represents the point set intersection of
        this geometry with another (Geometry)."""
        
    def union(anotherGeometry):
        """Returns a geometry that represents the point set union of
        this geometry with another (Geometry)."""
        
    def difference(anotherGeometry):
        """Returns a geometry that represents the point set difference of
        this geometry with another (Geometry)."""
        
    def symDifference(anotherGeometry):
        """Returns a geometry that represents the point set symmetric
        difference of this geometry with another (Geometry)."""
        
      
class ISpatiallyReferenced(Interface):

    """A spatially referenced geometry.
    """
    
    Attribute('srs',
    'Spatial reference system for this geometry and all of its components')
  

class IGeometry(IOGCSFGeometry, ISpatiallyReferenced):

    """A spatially referenced generic geometry.
    """


class IOGCSFGeometryCollection(IOGCSFGeometry):

    """A Geometry that is a collection of one or more geometries.
    """

    def numGeometries():
        """Returns the number of sub-geometries (int)."""

    def geometryN():
        """Returns the Nth geometry, index beginning with 0 (Geometry)."""


class IGeometryCollection(IOGCSFGeometryCollection, ISpatiallyReferenced):

    """A spatially referenced Polygon.
    """


class IOGCSFPoint(IOGCSFGeometry):

    """A 0-dimensional geometry.
    """

    def X():
        """Returns x coordinate (float)."""

    def Y():
        """Returns y coordinate (float)."""

    def Z():
        """Returns z coordinate (float)."""


class IPoint(IOGCSFPoint, ISpatiallyReferenced):

    """A more Python-friendly point."""
    x = Attribute("X coordinate value (float)")
    y = Attribute("Y coordinate value (float)")
    z = Attribute("Z coordinate value (float)")
    

class IOGCSFMultiPoint(IOGCSFGeometryCollection):

    """A collection, unordered, exclusively of points.
    
    Details
    -------
    isSimple : if any two points are equal, the collection is not simple.
    boundary : a multipoint is bounded by the empty set.
    """

    
class IMultiPoint(IOGCSFMultiPoint, ISpatiallyReferenced):

    """A spatially referenced MultiPoint.
    """


class IOGCSFCurve(IOGCSFGeometry):

    """One-dimensional geometric object, sub-typed by LineString.

    Details
    -------
    isSimple: if it does not pass through the same point twice
    boundary: empty set if curve is closed, else the two end points
    """

    def length():
        """Returns the length of this geometry in units of its own coordinate
        reference system (float)."""
    
    def startPoint():
        """Returns starting point of this curve (IOGCSFPoint)."""
   
    def endPoint():
        """Returns end point of this curve (IOGCSFPoint)."""

    def isClosed():
        """Returns 1 of the start and end points are equal, else 0 (int)."""

    def isRing():
        """Returns 1 if is closed and is simple, else 0 (int)."""


class IOGCSFLineString(IOGCSFCurve):

    """A curve with implicit linear interpolation between points.
    """

    def numPoints():
        """Returns the number of points (int)."""

    def pointN():
        """Returns the Nth point, index starting at 0 (IOGCSFPoint)."""


class ILineString(IOGCSFLineString, ISpatiallyReferenced):

    """Spatially referenced LineString.
    """

        
class IOGCSFLine(IOGCSFLineString):

    """A linestring with exactly 2 points.
    """


class IOGCSFLinearRing(IOGCSFLineString):

    """A linestring that is both closed and simple.
    """


class ILinearRing(IOGCSFLinearRing, ISpatiallyReferenced):

    """Spatially referenced LinearRing.
    """


class IOGCSFMultiCurve(IOGCSFGeometry):

    """A one-dimensional collection of curves.

    Details
    -------
    isSimple: if all members are simple
    boundary: see description of "mod 2" union rule in OGC SFSQL spec
    """

    def isClosed():
        """Returns 1 if all members are closed, else 0 (int)."""

    def length():
        """Returns sum of the length of all members (float)."""


class IMultiLineString(IOGCSFMultiCurve, ISpatiallyReferenced):

    """Spatially referenced MultiLineString.
    """


class IOGCSFSurface(IOGCSFGeometry):

    """A two-dimensional geometry.

    Basically, a patch with one exterior boundary (shell) and zero or more
    interior boundaries (holes).
    """

    def area():
        """Returns the area of the surface in units of its own coordinate
        reference system (float)."""

    def centroid():
        """Returns the mathematic centroid, not necessarily on the surface
        (IOGCSFPoint)."""

    def pointOnSurface():
        """Returns a point guaranteed to be on the surface (IOGCSFPoint)."""


class IOGCSFPolygon(IOGCSFSurface):

    """A restricted form of a surface.
    """

    def exteriorRing():
        """Returns the exterior ring, or shell, of this polygon
        (IOGCSFLinearRing)."""

    def numInteriorRings():
        """Returns the number of interior rings (int)."""

    def interiorRingN():
        """Returns the Nth interior ring, index starting at 1
        (IOGCSFLinearRing)."""


class IPolygon(IOGCSFPolygon, ISpatiallyReferenced):

    """A spatially referenced Polygon.
    """

    
class IOGCSFMultiSurface(IOGCSFSurface):

    """Two-dimensional collection of surfaces, whose interiors do not 
    intersect.
    """


class IMultiPolygon(IOGCSFMultiSurface, ISpatiallyReferenced):

    """A spatially referenced MultiPolygon.
    """


class IBoundingBox(IPolygon):
    
    """2D spatial box."""
    
    minx = Attribute("The lesser X coordinate")
    
    miny = Attribute("The lesser Y coordinate")
    
    maxx = Attribute("The greater X coordinate")
    
    maxy = Attribute("The greater Y coordinate")

    width = Attribute("The width of the bounding box")
    
    height = Attribute("The height of the bounding box")

    def totuple():
        """Return bounding box values as (minx,miny,maxx,may) tuple"""


