# $Id: srs.py,v 1.16 2005/04/22 03:21:35 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2005 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

__docformat__ = 'restructuredtext en'

try:
    from cartography.proj._osrtranslate import _translate
except ImportError:
    # OSR translation not available
    _translate = None


class SpatialReference:
    """A spatial reference system based on PROJ.4, the leading open source
    library for cartographic projections

      http://proj.maptools.org

    Refer to SpatialReferenceUserStory in the wiki for more discussion.
    """

    def __init__(self, proj4=None, epsg=None):
        """Initialize a SpatialReference

        Parameters
        ---------
        proj4 : string
            PROJ.4 parameter string such as '+proj=latlong datum=WGS84'
        epsg : int
            a numerical EPSG code such as 4326

        Example
        -------
        >>> wgs84 = SpatialReference('proj=latlong datum=WGS84')
        >>> epsg26913 = SpatialReference(epsg=26913)
        """
        self._proj4 = None
        self._epsg = None
        
        # PROJ.4 input takes precedence
        if proj4:
            params = []
            # Our preferred usage is string input
            try:
                for p in proj4.split():
                    p = p.strip()
                    if not p.startswith('+'):
                        p = '+' + p
                    params.append(p)
            # the backwards compatibility case where proj4 is a list
            except AttributeError:
                for p in proj4:
                    p = p.strip()
                    if not p.startswith('+'):
                        p = '+' + p
                    params.append(p)
            self._proj4 = ' '.join(params)
        elif epsg:
            # future: assert that it's a valid epsg
            self._epsg = epsg

    def tostring(self):
        """Return a PROJ.4 string representation of the srs
        
        Example
        -------
        >>> epsg26913 = SpatialReference(epsg=26913)
        >>> epsg26913.tostring()
        '+init=epsg:26913'
        """
        # As always, any saved PROJ.4 takes precedence
        if self._proj4:
            return self._proj4
        elif self._epsg:
            return '+init=epsg:%d' % (self._epsg)
        else:
            raise ValueError, "Spatial reference system is undefined"

    def __str__(self):
        """Use tostring() to generate a string representation"""
        return self.tostring()

    def equals(self, srs):
        """Returns the True if this SpatialReference equals the
        given one.

        Note: without the PCL-GDAL package, we can't compare EPSG form
        to PROJ.4 form. These will appear to be different reference
        systems, possibly triggering unnecessary reprojection of data.
        """
        def equals_proj4(p1, p2):
            """Compares two proj4 definitions."""
            params1 = [param.lower() for param in p1.split()]
            params2 = [param.lower() for param in p2.split()]
            params1.sort()
            params2.sort()
            return params1 == params2

        if not isinstance(srs, SpatialReference):
            raise ValueError, "srs parameter is not a spatial reference"
            
        # Both using EPSG codes
        if self._epsg and srs._epsg:
            return self._epsg == srs._epsg
        # Both using PROJ.4 parameters
        elif self._proj4 and srs._proj4:
            return equals_proj4(self._proj4, srs._proj4)
        
        elif _translate:    
            # EPSG to PROJ.4 comparison
            if self._epsg and srs._proj4:
                proj4 = _translate(str(self._epsg), 'epsg', 'proj4')
                return equals_proj4(proj4, srs._proj4)
            # PROJ.4 to EPSG comparison
            elif self._proj4 and srs._epsg:
                proj4 = _translate(str(srs._epsg), 'epsg', 'proj4')
                return equals_proj4(self._proj4, proj4)
        
        # finally
        return False

    def __eq__(self, srs):
        """Returns True if srs equals this SpatialReference, False otherwise."""
        return self.equals(srs)

    def __ne__(self, srs):
        """Returns False if srs equals this SpatialReference, True otherwise."""
        return not self.equals(srs)

