/*
# =============================================================================
# Python Cartographic Library. Copyright (C) 2005 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================
*/

#include "Python.h"
#include <proj_api.h>

#define ERROR_BUFFER_LENGTH 200

/*
===============================================================================
 _proj4_transform_points

 Transform a list of coordinate tuples from one coordinate system to another.
 
 Parameters: "OOO"
    1: the destination coordinate system, specified by a string of proj4
       parameters
    2: the source coordinate system
    3: a sequence of X,Y coordinate tuples
    
 Example:

    results = _proj4.transform_points('+init=epsg:27582', '+init=epsg:4326',
                                      [(0.0, 47.0)])

===============================================================================
*/
   
static PyObject *
_proj4_transform_points(PyObject *self, PyObject *args)
{
    //char error_msg[ERROR_BUFFER_LENGTH];
    PyObject *arg1, *arg2, *arg3;
    PyObject *coords; 
    PyObject *results = NULL;
    projPJ *dst, *src;
    int i, error;
    double x, y, z;
    
    /* Parse for the three expected args */
    if (!PyArg_ParseTuple(args, "OOO:proj4_transform_points",
            &arg1, &arg2, &arg3))
        goto fail;
   
    /* Make PJ from destination and source */
    dst = pj_init_plus(PyString_AsString(arg1));
    if (!dst) goto fail;
    src = pj_init_plus(PyString_AsString(arg2));
    if (!src) goto fail;

    /* Transform all input points and pop into results list */
    results = PyList_New(0);
    for (i=0; i<PyList_Size(arg3); i++)
    {
        coords = PyList_GetItem(arg3, i);
        
        x = PyFloat_AsDouble(PyTuple_GetItem(coords, 0));
        y = PyFloat_AsDouble(PyTuple_GetItem(coords, 1));
        if (pj_is_latlong(src))
        {
            x *= DEG_TO_RAD;
            y *= DEG_TO_RAD;
        }

        z = 0.0;
        error = pj_transform(src, dst, 1, 0, &x, &y, &z);
        if (error || x == HUGE_VAL || y == HUGE_VAL) goto fail;

        if (pj_is_latlong(dst))
        {
            x *= RAD_TO_DEG;
            y *= RAD_TO_DEG;
        }

        error = PyList_Append(results, Py_BuildValue("(dd)", x, y));
        if (error < 0) goto fail;
    }
   
    pj_free(dst);
    pj_free(src);
    return results;
    
    fail:
    return NULL;
}

/*
===============================================================================
Initialization
===============================================================================
*/
static PyMethodDef Proj4Methods[] = {
    { "transform_points", _proj4_transform_points, METH_VARARGS,
      "Transform points to a coordinate system from another."},
    {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC
init_proj4(void)
{
    (void) Py_InitModule("_proj4", Proj4Methods);
}

