# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
#
# $Id: interfaces.py 544 2006-10-20 20:47:35Z seang $
# =============================================================================

try:
    from zope.interface import Interface, Attribute
except ImportError:
    class Interface:
        pass
    class Attribute:
        def __init__(self, *args):
            pass

# =============================================================================
# Web Service Interfaces
# =============================================================================

class IOWSDescriptor(Interface):
    """Describes an OGC Web Service."""
    
    url = Attribute("Base URL for requests (string)")

    version = Attribute("Protocol version, such as '1.1.1' (string)")

class IOWSMapRequest(Interface):
    """Encapsulate an WMS getmap request."""

    ows_format = Attribute("Request format, like 'image/jpeg' (string)")

    ows_srs = Attribute("Request SRS, like 'EPSG:4326' (string)")


class ILayer(Interface):
    """A cartographic layer."""

    datastore = Attribute("Object that implements IDataStore")

    typename = Attribute("Name of a type within the datastore (string)")

    filter = Attribute("A feature property-based filter expression (string)")

    extent = Attribute("A spatial extent-based feature constraint")
    
    classifier = Attribute("0 if classification is carried out internally "
        "in Python, or 1 if classification is carried out externally by the "
        "rendering engine.")
    
class IStyle(Interface):
    """Style interface."""

    name = Attribute("Name of style.")

    rules = Attribute("Sequence of objects that implement ISLDRule (iterable)")

    maxrules = Attribute("Maximum number of rules.")


# TODO: complete symbolizer interfaces
class ISLDSymbolizer(Interface):
    """Symbolization of features."""


class ISLDStroke(Interface):
    """Stroke interface."""
    
    stroke = Attribute("Object that implements IStrokeProperties")


class ISLDFill(Interface):
    """Fill interface."""
    
    fill = Attribute("Object that implement IFillProperties")


class IStrokeProperties(Interface):
    """Stroke properties"""
    color = Attribute("Stroke color")

    opacity = Attribute("Stroke opacity")

    width = Attribute("Stroke width")

    dasharray = Attribute("Stroke dasharray")

    #graphicfill = Attribute("Graphic fill")
    #graphicstroke = Attribute("Graphic stroke")


class IFillProperties(Interface):
    """Fill properties"""
    color = Attribute("Stroke color")

    opacity = Attribute("Stroke opacity")

    #graphicfill = Attribute("Graphic fill")


class IFontProperties(Interface):
    """Font properties"""
    name = Attribute("Font name")
    
    size = Attribute("Font size")


class IHaloProperties(ISLDFill):
    """Halo properties."""
    
    radius = Attribute("Halo radius")
    

class IPlacementProperties(Interface):
    """Placement properties"""
    anchorpoint = Attribute("Anchor point defined as object implementing IPoint")
    
    displacement = Attribute("Displacement defined as object implementing IPoint")
    
    rotation = Attribute("Rotation")
    
    padding = Attribute("Padding")
    
    mindupedistance = Attribute("Minimum duplicate distance")


class ISLDMark(Interface):
    """A little vector picture with a well-known name."""

    name = Attribute("name of this mark, such as 'circle' or 'square'")
    stroke = Attribute("stroke properties")
    fill = Attribute("fill properties")


class ISLDGraphic(Interface):
    """A little picture."""
    
    externalgraphic = Attribute("""A pixmap file object""")
    mark = Attribute("""An implementation of ISLDMark""")
    size = Attribute("""Pixel height of the graphic""")
    opacity = Attribute("""Graphic opacity (0-1)""")
   
    
class ISLDPointSymbolizer(ISLDSymbolizer, ISLDStroke, ISLDFill):
    """Point symbolization."""
    
    mark = Attribute("Name of a predefined vector mark")
    
    size = Attribute("Size of the symbolizer")
    
    graphic = Attribute("Symbolizer graphic")

    
class ISLDLineSymbolizer(ISLDSymbolizer, ISLDStroke):
    """Line symbolization."""

    
class ISLDPolygonSymbolizer(ISLDSymbolizer, ISLDStroke, ISLDFill):
    """Polygon symbolization."""


class ISLDTextSymbolizer(ISLDSymbolizer, ISLDStroke, ISLDFill):
    """Text symbolization."""
    
    label = Attribute("The label")
    
    font = Attribute("Object that implements IFontProperties")
    
    halo = Attribute("Object that implements IHaloProperties")
    
    placement = Attribute("Object that implements IPlacementProperties")

    minfeaturesize = Attribute("The minimum feature size")
    
    partials = Attribute("Partial label rendering (boolean)")
    
    force = Attribute("Forced label rendering (boolean)")
    
    autominfeaturesize = Attribute("Automatic minimum feature size (boolean)")


class ISLDRasterSymbolizer(ISLDSymbolizer):
    """Raster symbolization.
    """
    opacity = Attribute("Raster opacity")
   

class ISLDRule(Interface):
    """OGC SLD Rule.
    
    A rule maps classes of features to one or more cartographic symbolizers.
    Feature classes are produced by applying a filter to a feature iterator.
    A rule can be constrained to be active for a range of cartographic scales
    by setting the minimum and maximum bounding scale denominators. For 
    example:

      minscaledenominator = 10000
      maxscaledenominator = 100000

    produces a rule that is only active at scales between 1:10000 and
    1:100000.
    """

    name = Attribute("Identifier unique within style context (string)")

    title = Attribute("Title of rule to be used in map legends (string)")

    abstract = Attribute("Detailed information about rule (string)")
    
    legendgraphic = Attribute("File-like object")
    
    filter = Attribute("Object that implements ISLDFilter")
    
    minscaledenominator = Attribute(
    "The minimum of a rule's two bounding scale denominators. Features "
    "will not be displayed at scales larger than (more zoomed in than) "
    "1/minscaledenominator" )
    
    maxscaledenominator = Attribute(
    "The maximum of a rule's two bounding scale denominators. Features "
    "will not be displayed at scales smaller than (more zoomed out than) "
    "1/maxscaledenominator" )
    
    symbolizer = Attribute("Sequence of objects that implement ISLDSymbolizer")

    
class ISLDFilter(Interface):
    """Interface specification for filters."""

    expression = Attribute("Filter expression")


class IMapRenderer(Interface):
    """Map rendering interface.  Make a map from a set of layers."""

    def render(layering, view, format, size, bgcolor, transparent):
        """Makes map from layers, returning an object specified by the
        format and size parameters.
        
        Arguments
        ---------
        layering : iterable
            A list of (Layer, Style) tuples in "painting" order
            
        view : View
            spatial extent and reference system of output map
            
        format : string
            A named format specific to the implementor
            
        size : iterable
            image type implementors should require a width, height
            pair specifying the pixel dimensions of the output map
            
        bgcolor : string
            background color for image type maps
            
        transparent : boolean
            transparent background or not for image type maps
        """

    def draw(rule, format, size=[16, 16], bgcolor='#FFFFFF', transparent=False):
        """Makes an icon-sized image of a symbolizer.

        See the render method for explanation of arguments
        """

class IViewport(Interface):
    pass


