# $Id: mapping.py,v 1.16 2005/04/08 19:13:54 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004, 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

from zope.interface import implements
from cStringIO import StringIO

from cartography.styles.interfaces import ILayer, IStyle
from cartography.styles.interfaces import IFontProperties, IStrokeProperties, \
     IFillProperties, IHaloProperties, IPlacementProperties
from cartography.styles.interfaces import ISLDPointSymbolizer \
    , ISLDLineSymbolizer, ISLDPolygonSymbolizer, ISLDTextSymbolizer \
    , ISLDRasterSymbolizer, ISLDGraphic, ISLDMark


class Image:
    
    """Map imagery.
    
    Attributes
    ----------
    mimetype : string
        'image/png', 'image/jpeg', 'image/gif'
    data : StringIO
        the image data.
    size : []
        image dimensions in pixels, [width, height]
    """

    def __init__(self, mimetype=''):
        """Initialize an Image.  The mimetype should be like 'image/png'."""
        self.mimetype = mimetype
        self.data = StringIO()
        self.size = []

    def getdata(self):
        """Return image data"""
        return self.data.getvalue()


class Layer:
    """A Layer inhabits a discrete cartographic "z" level.
    """
    implements(ILayer)

    def __init__(self, datastore, typename=None, filter=None, extent=None,
                 maxfeatures=None, classifier='INTERNAL'):
        """Initialize.
        
        Parameters
        ----------
        datastore : IDataStore
            A data store
        typename : string
            Name of one of the feature types in the store
        filter : string
            An SQL-ish filter applied by the data store
        extent : ?
            Extent
        classifier : string
            Either 'INTERNAL' for classification carried out in Python, or
            'EXTERNAL' for classification carried out by the rendering engine.
        """
        self.datastore = datastore
        self.typename = typename
        self.filter = filter
        self.extent = extent
        self.maxfeatures = maxfeatures
        if classifier == 'INTERNAL':
            self.classifier = 0
        else:
            self.classifier = 1
        

class Rule:

    """A rule maps classes of features to one or more cartographic symbolizers.
    Feature classes are produced by applying a filter to a feature iterator.
    A rule can be constrained to be active for a range of cartographic scales
    by setting the minimum and maximum bounding scale denominators. For 
    example:

      minscaledenominator = 10000
      maxscaledenominator = 100000

    produces a rule that is only active at scales between 1:10000 and
    1:100000.

    Attributes
    ----------
    name : string
        Rule name
    symbolizer : Symbolizer | list of Symbolizers
        Reference to Line, Polygon, Point, or TextSymbolizer
    filter : filter.Filter
        instance of Filter
    minscaledenominator : float
        Minimum value of map scale denominator.  Zooming in past this value
        turns off the Rule.
    maxscaledenominator : float
        maximum value of map scale denominator.  Zooming out past this value
        turns off the Rule.
    params : mapping
        environment for evaluation of filter
    """

    def __init__(self, name, symbolizer, filter=None, 
                 minscaledenominator=0, maxscaledenominator=1e+20, **params):
        """Initialize a Rule."""
        self.name = str(name)
        self.filter = filter
        self.minscaledenominator = minscaledenominator
        self.maxscaledenominator = maxscaledenominator
        if isinstance(symbolizer, (tuple,list)):
            self.symbolizer = symbolizer
        else:
            self.symbolizer = [symbolizer]
        self.params = params.copy()


class Style:
    """Describes the presentation/visualization of a Layer.

    Implements IStyle
    
    Attributes
    ----------
    name : string
        optional name
    rules : []
        list of Rules in order of decreasing priority
    maxrules : int
        maximum number of rules that will be symbolized.
    """
    implements(IStyle)

    def __init__(self, name='', maxrules=0):
        """Initialize a Style."""
        self.name = str(name)
        self.rules = []
        self.maxrules = maxrules


class ExtrinsicStyle(Style):

    """Contains Rules that use extrinsic values in expressions.

    Rules of a Style class must use expressions only in terms of features'
    intrinsic attributes.  The Rules of an ExtrinsicStyle are not constrained
    in this way.
    """


class Stroke(object):
    """A CSS-like model for stroke properties.
    """
    implements(IStrokeProperties)

    def __init__(self, color='#000000', opacity=1.0, width=1.0, dasharray=[],
        antialiasing=False):
        """Initialize.
        
        Parameters
        ----------
        color : string
            Hex color definition.
        opacity : float
            Range: [0-1]
        width : float
            Width in pixels.
        dasharray : [int]
            A list of integers specify the dot/dash line style. [1, 1] is the
            simplest dotted line, one pixel on, one pixel off:

              - - - - - - - - - ...
              
            [3, 1, 1, 1] creates a repeating pattern: 3 pixels on, one off, and
            one on:

              --- - --- - --- - --- - --- ...
        
        antialiasing : bool
            Whether to render using antialiasing. This depends on capabilities
            of the renderer.
        """
        self.color = str(color)
        self.opacity = float(opacity)
        self.width = float(width)
        self.dasharray = dasharray
        self.antialiasing = False

    def todict(self):
        """Convert to a dict."""
        return {'stroke':           self.color,
                'stroke-opacity':   self.opacity,
                'stroke-width':     self.width,
                'stroke-dasharray': self.dasharray,
                'stroke-antialiasing': self.antialiasing,
                'graphicfill':      None,
                'graphicstroke':    None
                }


class Fill(object):
    """A CSS-like model for fill properties.
    """
    implements(IFillProperties)

    def __init__(self, color='#808080', opacity=0.5):
        """Initialize.
        
        Parameters
        ----------
        color : string
            Hex color definition.
        opacity : float
            Fill opacity [0-1].
        """
        self.color = str(color)
        self.opacity = float(opacity)
        
    def todict(self):
        """Convert to a dict."""
        return {'fill':         self.color,
                'fill-opacity': self.opacity,
                'graphicfill':  None
                }


class Mark(object):
    """A named vector type point symbol.
    """
    implements(ISLDMark)
    
    def __init__(self, name, stroke=0, fill=0):
        """Initialize a graphical mark.

        Parameters
        ----------
        name : string
            Name of a mark, must be provided by the renderer.
        stroke : ISLDStroke
            Stroke, as though the mark were a polygon.
        fill : ISLDFill
            Fill, as though the mark were a polygon.
        """
        self.name = str(name)
        if stroke == 0: self.stroke = Stroke()
        else: self.stroke = stroke
        if fill == 0: self.fill = Fill()
        else: self.fill = fill


class Graphic(object):
    """A little picture used in a point symbolizer.
    """
    implements(ISLDGraphic)
    
    def __init__(self, externalgraphic=None, mark=None, opacity=1.0, size=6.0):
        """Initialize a graphic from a pixmap or a graphical mark.

        Parameters
        ----------
        externalgraphic : string or Image
            Image instance, or path/URL to a pixmap.
        mark : ISLDMark
            A graphical mark.

        The externalgraphic parameter takes precedence over the mark.
        
        opacity : float
            Range [0-1].
        size : float
            Rendered size of the graphic in pixels. The renderer should scale
            pixmaps or marks to this size.
        """
        self.mark = mark
        self.size = float(size)
        self.opacity = float(opacity)
        if type(externalgraphic) in [type('foo'), type(u'foo')]:
            f = file(externalgraphic, 'rb')
            imdata = f.read()
            f.close()
            self.externalgraphic = Image()
            self.externalgraphic.data.write(imdata)
        else:
            self.externalgraphic = externalgraphic
        if not externalgraphic and not mark:
            self.mark = Mark('circle')


class PointSymbolizer:
    """Describes how point features are rendered onto a map.
    """
    implements(ISLDPointSymbolizer)

    def __init__(self, graphic=0, geometry=None):
        """Initialize a PointSymbolizer.
        
        Parameters
        ----------
        graphic : ISLDGraphic
            A graphic
        geometry : string
            Optional name of a feature geometry property. If None, the 
            feature's default geometry property will be used.
        """
        if graphic == 0: self.graphic = Graphic()
        else: self.graphic = graphic
        self.geometry = geometry


class LineSymbolizer:
    """Describes how line features are rendered onto a map.
    """
    implements(ISLDLineSymbolizer)

    def __init__(self, stroke=0, geometry=None):
        """Initialize.

        Parameters
        ----------
        stroke : ISLDStroke
            Line stroke.
        geometry : string
            Optional name of a feature geometry property. If None, the 
            feature's default geometry property will be used.
        """
        if stroke == 0: self.stroke = Stroke()
        else: self.stroke = stroke
        self.geometry = geometry

    
class PolygonSymbolizer:
    """Describes how polygon features are rendered onto a map.
    """
    implements(ISLDPolygonSymbolizer)

    def __init__(self, stroke=0, fill=0, geometry=None):
        """Initialize.

        Parameters
        ----------
        stroke : ISLDStroke
            Polygon stroke (outline).
        fill : ISLDFill
            Polygon fill (interior).
        geometry : string
            Optional name of a feature geometry property. If None, the 
            feature's default geometry property will be used.
        """
        if stroke == 0: self.stroke = Stroke()
        else: self.stroke = stroke
        if fill == 0: self.fill = Fill()
        else: self.fill = fill
        self.geometry = geometry


class Halo(object):
    """Describes the outlining of text in a label.
    """
    implements(IHaloProperties)

    def __init__(self, radius=0, fill=0):
        """Initialize.
        
        Parameters
        ----------
        radius: float
            Halo radius in pixels.
        fill : ISLDFill
            Halo color.
        """
        self.radius=float(radius)
        if fill == 0: self.fill = Fill('#FFFFFF')
        else: self.fill = fill


class Font(object):
    """Describes the font properties.
    """
    implements(IFontProperties)
    
    def __init__(self, name='', size=10.0):
        """Initialize.
        
        Parameters
        ----------
        name : string
            Name of a font, must be provided by a renderer.
        size : float
            Font size in pixels.
        """
        self.name = str(name)
        self.size = float(size)

    def todict(self):
        """Convert to a dict."""
        return {'font-family':  self.name,
                'font-style':   '',
                'font-weight':  '',
                'font-size':    self.size
                }

        
class Placement(object):
    """Describes the label placement properties.
    """
    implements(IPlacementProperties)
    
    # XXX: an offset parameter was in the original param list, but we've
    # never had a renderer that implemented it -- SG
    def __init__(self, anchorpoint=None, displacement=None, rotation=None,
                 padding=0.0, mindupedistance=0.0):
        """Initialize.
        
        Parameters
        ----------
        anchorpoint : tuple or IPoint
            Specifies how the text symbolizer is anchored to the feature. A
            tuple argument should contain 2 values: (X, Y). IPoint is easily
            implemented for this purpose:

                class XY:
                    def __init__(self, x, y):
                        self.x = x
                        self.y = y

                p = Placement(anchorpoint=XY(0,0))
            
            or you could use cartography.spatial.geometry.Point.
            
            (-1, 1) means that text appears to the upper left of
            the feature. (0, 0) is centered on the feature. (1, -1) means that
            text appears to the lower right of the feature.

            A value of None (the default) shall be understood by the renderer
            to be a "best" label placement.
        displacement : tuple or IPoint
            x and y displacement of the text from the feature. A value of None
            means no additional displacement of text from its anchor point.
        rotation : float
            Rotation angle of text. If None, text will be auto-aligned, 
            most useful with line features.
        padding : float
            Padding around text. A renderer may use this to exclude other text
            for the sake of map legibility.
        mindupedistance : float
            Prevent duplicate labels within this distance.
        """
        class XY:
            def __init__(self, args):
                self.x = args[0]
                self.y = args[1]
        
        if isinstance(anchorpoint, tuple):
            self.anchorpoint = XY(anchorpoint)
        else:
            self.anchorpoint = anchorpoint
        
        if isinstance(displacement, tuple):
            self.displacement = XY(displacement)
        else:
            self.displacement = displacement
        
        self.rotation = rotation
        self.padding = float(padding)
        self.mindupedistance = float(mindupedistance)

    def todict(self):
        """Convert to a dict."""
        return {'anchorpoint':      self.anchorpoint,
                'displacement':     self.displacement,
                'rotation':         self.rotation,
                'padding':          self.padding,
                'mindupedistance':  self.mindupedistance
                }

                
class TextSymbolizer:
    """Describes how features are labeled on a map.
    """
    implements(ISLDTextSymbolizer)

    def __init__(self, label='', font=None, fill=0, placement=None,
                 halo=None, minfeaturesize=-1, autominfeaturesize=False,
                 partials=False, force=False, geometry=None):
        """Initialize.
        
        Parameters
        ----------
        label : string
            Name of feature property to use in label.
        font : Font
            Font for text.
        fill : ISLDFill
            Font color.
        placement : Placement
            Placement of text.
        halo : Halo
            Text halo (outline).
        minfeaturesize : float
            Only features larger (in pixels) than this value will be labeled.
        autominfeaturesize : bool
            If True, only features larger than the text box will be labeled.
        partials : bool
            Whether or not to render partial text at map edges.
        force : bool
            Turn off label collision avoidance (on by default) and render all
            labels.
        geometry : string
            Name of feature geometry to which the text will be anchored.
        """
        self.label = label
        self.font = font or Font(size=10)
        self.placement = placement or Placement()
        if fill == 0: self.fill = Fill('#000000')
        else: self.fill = fill
        self.halo = halo or Halo()
        self.minfeaturesize = minfeaturesize
        self.partials = partials
        self.force = force
        self.autominfeaturesize = autominfeaturesize
        self.geometry = geometry
      
      
class RasterSymbolizer:
    """Raster symbolization.
    """
    implements(ISLDRasterSymbolizer)

    def __init__(self, opacity=1.0, geometry=None):
        """Initialize.
        Parameters
        ----------
        opacity : float
            Range: [0-1]
	"""
        self.geometry = None
        self.opacity = opacity


