# =============================================================================
# Python Cartographic Library. Copyright (C) 2005 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import sys
import unittest

import pkg_resources
pkg_resources.require('PCL-Core')

#import pdb; pdb.set_trace()

from cartography.data.feature import *
from cartography.geometry import Point, BoundingBox
from cartography.proj import SpatialReference


class PropertyTest(unittest.TestCase):
    def test_int(self):
        p = IntegerProperty('p')
        self.assertEqual(p.totuple(), ('p', 'int'))
    def test_real(self):
        p = RealProperty('p')
        self.assertEqual(p.totuple(), ('p', 'float'))
    def test_string(self):
        p = StringProperty('p')
        self.assertEqual(p.totuple(), ('p', 'string'))
    def test_date(self):
        p = DateProperty('p')
        self.assertEqual(p.totuple(), ('p', 'datetime'))
    def test_bool(self):
        p = BooleanProperty('p')
        self.assertEqual(p.totuple(), ('p', 'bool'))

class GeometryTest(unittest.TestCase):
    def test_homogeneous(self):
        p = GeometryProperty('p', homosrs=SpatialReference(epsg=26913),
                             homotype='Point')
        self.assertEqual(p.totuple(), ('p', 'geometry'))
        self.assertEqual(p.homosrs.tostring(), '+init=epsg:26913')
        self.assertEqual(p.homotype, 'Point')
        self.assertEqual(p.defaultsrs.tostring(), '+init=epsg:26913')
       

class NaiveTypeTest(unittest.TestCase):
    # Traditional single geometry, single srs, not aware of the GML simple
    # features profile.
    def setUp(self):
        epsg4326 = SpatialReference(epsg=4326)
        class F(Feature):
            # A simple feature type for testing
            __schema__ = PropertyTypeCollection([
                            StringProperty('name'),
                            IntegerProperty('category'),
                            GeometryProperty('position', homosrs=epsg4326, 
                                homotype='Point')
                            ])
            __typename__ = 'F'
        self.ft = F
    def test_schema(self):
        s = self.ft.schema()
        n = s.names()
        n.sort()
        self.assertEqual(n, ['category', 'name', 'position'])
    def test_info(self):
        s = self.ft.schema_info()
        self.assertEqual(s,
            {'name': 'string', 'category': 'int', 'position': 'geometry'})
    def test_props(self):
        self.assertEqual(self.ft.typename(), 'F')
        self.assertEqual(self.ft.defaultgeometry().name, 'position')
        self.assertEqual(self.ft.defaultsrs().tostring(), '+init=epsg:4326')
        
class TypeTest(unittest.TestCase):
    # A feature type that is aware of GMLSFP
    def setUp(self):
        class F(Feature):
            # A simple feature type for testing
            __schema__ = PropertyTypeCollection([
                            StringProperty('name'),
                            IntegerProperty('category'),
                            GeometryProperty('position'),
                            GeometryProperty('domain') 
                            ])
            __typename__ = 'F'
            __g__ = 'domain'
        self.ft = F
    def test_schema(self):
        s = self.ft.schema()
        n = s.names()
        n.sort()
        self.assertEqual(n, ['category', 'domain', 'name', 'position'])
    def test_info(self):
        s = self.ft.schema_info()
        self.assertEqual(s,
            {'name': 'string', 'category': 'int', 'position': 'geometry',
             'domain': 'geometry'})
    def test_props(self):
        self.assertEqual(self.ft.typename(), 'F')
        self.assertEqual(self.ft.defaultgeometry().name, 'domain')
        self.assertEqual(self.ft.defaultsrs(), None)

class FeatureTest(unittest.TestCase):
    def setUp(self):
        epsg4326 = SpatialReference(epsg=4326)
        class F(Feature):
            # A simple feature type for testing
            __schema__ = PropertyTypeCollection([
                            StringProperty('name'),
                            IntegerProperty('category'),
                            GeometryProperty('position', homosrs=epsg4326, 
                                homotype='Point')
                            ])
            __typename__ = 'F'
        self.ft = F
        self.f = F('f1', name='feature one', category=0, position=Point(1,1))
    def test_iface(self):
        self.assert_(IFeature.providedBy(self.f))
    def test_id(self):
        self.assert_(self.f.id == 'f1')
    def test_properties(self):
        self.assert_(hasattr(self.f, 'properties') == True)
        self.assert_(self.f.properties.name == 'feature one')
        self.assert_(self.f.properties.category == 0)
        self.assertAlmostEqual(self.f.properties.position.x, 1.0)
    def test_isfeature(self):
        self.assert_(isinstance(self.f, Feature) == True)
    def test_bounds(self):
        self.assertEqual(self.f.bounds.totuple(), (1.0, 1.0, 1.0, 1.0))
    def test_bounds_set(self):
        box = BoundingBox(0.0, 0.0, 2.0, 2.0, SpatialReference(epsg=4326))
        f = self.ft('f1', bounds=box, name='feature one', category=0,
            position=Point(1,1))
        self.assertEqual(f.bounds.totuple(), (0.0, 0.0, 2.0, 2.0))
        self.assertEqual(f.bounds.srs.tostring(), '+init=epsg:4326')
    def test_bogusname(self):
        self.assertRaises(AttributeError, self.ft, 'f1', nom='feature one',
            category=0)
    def test_info(self):
        self.assertEqual(self.f.info(),
        {'srs': 'None', 'properties': {'category': 0, 'name': 'feature one'}, 'id': 'f1', 'bounds': [1.0, 1.0, 1.0, 1.0]})



# =============================================================================
if __name__ == '__main__':
    unittest.main()

