# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import struct
import unittest

import pkg_resources
pkg_resources.require('PCL-Core')

from cartography.proj import SpatialReference
from cartography.proj import ProjTransform
from cartography.geometry import *

from _testing import *

epsg26913 = SpatialReference('+init=epsg:26913')
epsg4326 = SpatialReference('+init=epsg:4326')
TRANSFORM = ProjTransform(epsg26913, epsg4326)
TRANSFORM2 = ProjTransform(SpatialReference('+proj=tmerc +lon_0=0.0 +lat_0=0.0 +ellps=GRS80'), epsg4326)


class Test(unittest.TestCase):
    def test_point_kw(self):
        p = Point(-8.0, 50.0)
        p.transform(dst_proj='+init=epsg:29900', src_proj='+init=epsg:4326')
        self.assertAlmostEqual(p.x, 200000.0, 6)
        self.assertAlmostEqual(p.y, -139462.3036031, 6)
    def test_point_trans(self):
        p = Point(-105.5, 40.0)
        p.transform(TRANSFORM)
        self.assertAlmostEqual(p.x, 457320.0549729, 6)
        self.assertAlmostEqual(p.y, 4427876.9243054, 6)
    def test_line(self):
        l = LineString([Point(-105.5, 40.0), Point(-105.5, 40.5)])
        self.assertAlmostEqual(l.length(), 0.5, 6)
        l.transform(TRANSFORM)
        self.assertAlmostEqual(l.length(), 55498.7488926, 6)
        self.assertAlmostEqual(l.startPoint().x, 457320.0549729, 6)
        self.assertAlmostEqual(l.startPoint().y, 4427876.9243054, 6)
    def test_poly(self):
        p = Geometry.fromWKT('POLYGON ((-105.5 40,-105.5 40.5,-105 40.5,-105 40,-105.5 40))')
        self.assertAlmostEqual(p.area(), 0.25, 6)
        p.transform(TRANSFORM)
        self.assertAlmostEqual(p.area(), 2359973622.4577727, 6)
    def test_poly_holes(self):
        p = Geometry.fromWKT('POLYGON ((-105.5 40,-105.5 40.5,-105 40.5,-105 40,-105.5 40), (-105.3 40.3,-105.3 40.4,-105.2 40.4,-105.2 40.3,-105.3 40.3))')
        self.assertAlmostEqual(p.area(), 0.24, 6)
        p.transform(TRANSFORM)
        self.assertAlmostEqual(p.area(), 2265711450.5888467, 6)
    def test_multipoint(self):
        g = Geometry.fromWKB(MULTIPOINT)
        g.transform(TRANSFORM2)
    def test_multiline(self):
        g = Geometry.fromWKB(MULTILINESTRING)
        g.transform(TRANSFORM2)
    def test_multipoly(self):
        g = Geometry.fromWKB(MULTIPOLYGON)
        self.assertAlmostEqual(g.area(), 3.25)
        g.transform(TRANSFORM2)
        self.assertAlmostEqual(g.area(), 40008483598.414803)


# ============================================================================
if __name__ == '__main__':
    unittest.main()

