# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import struct
import unittest

import pkg_resources
pkg_resources.require('PCL-Core')

from cartography.geometry._geom import *

from _testing import *

# Base class for geometry tests
class GeometryTest:
    def test_unary(self):
        g = self.g
        self.assertEqual(self.g.isEmpty(), self.isempty, g.isEmpty())
        self.assertEqual(g.isValid(), self.isvalid, g.isValid())
        self.assertEqual(g.isSimple(), self.issimple, g.isSimple())
        self.assertEqual(g.isRing(), self.isring, g.isRing())
        self.assertEqual(g.hasZ(), self.hasZ, g.hasZ())
    def test_computed(self):
        g = self.g
        self.assertAlmostEqual(g.area(), self.area, 6, g.area())
    def test_info(self):
        g = self.g
        self.assertEqual(g.getTypeId(), self.typeid, g.getTypeId())
        self.assertEqual(g.getDimensions(), self.dims, g.getDimensions())
        self.assertEqual(g.numGeometries(), self.num, g.numGeometries())

# Testing geometries used by these are at the bottom of this file
class MultiPolygonTest(unittest.TestCase, GeometryTest):
    def setUp(self):
        self.g = createGeomFromWKB(MULTIPOLYGON)
        self.typeid = 6
        self.isempty = 0
        self.isvalid = 1
        self.issimple = 1
        self.isring = 0
        self.hasZ = 1
        self.dims = 3
        self.area = 3.25
        self.num = 2

class PolygonTest(unittest.TestCase, GeometryTest):
    def setUp(self):
        self.g = createGeomFromWKB(POLYGON)
        self.typeid = 3
        self.isempty = 0
        self.isvalid = 1
        self.issimple = 1
        self.isring = 0
        self.hasZ = 1
        self.dims = 3
        self.area = 3
        self.num = -1
        
class MultiLineStringTest(unittest.TestCase, GeometryTest):
    def setUp(self):
        self.g = createGeomFromWKB(MULTILINESTRING)
        self.typeid = 5
        self.isempty = 0
        self.isvalid = 1
        self.issimple = 0
        self.isring = 0
        self.hasZ = 1
        self.numgeoms = 1
        self.dims = 3
        self.area = 0.0
        self.num = 2

class LineStringTest(unittest.TestCase, GeometryTest):
    def setUp(self):
        self.g = createGeomFromWKB(LINESTRING)
        self.typeid = 1
        self.isempty = 0
        self.isvalid = 1
        self.issimple = 1
        self.isring = 0
        self.hasZ = 1
        self.dims = 3
        self.area = 0.0
        self.num = -1

class PointTest(unittest.TestCase, GeometryTest):
    def setUp(self):
        self.g = createGeomFromWKB(POINTA)
        self.typeid = 0
        self.isempty = 0
        self.isvalid = 1
        self.issimple = 1
        self.isring = 0
        self.hasZ = 0
        self.dims = 3
        self.area = 0.0
        self.num = -1

class MultiPointTest(unittest.TestCase, GeometryTest):
    def setUp(self):
        self.g = createGeomFromWKB(MULTIPOINT)
        self.typeid = 4
        self.isempty = 0
        self.isvalid = 1
        self.issimple = 1
        self.isring = 0
        self.hasZ = 0
        self.dims = 3
        self.area = 0.0
        self.num = 2
       
class DisjointTest(unittest.TestCase):
    def test_disjoint(self):
        g1 = createGeomFromWKB(LINESTRING)
        g2 = createGeomFromWKB(POINTB)
        self.assert_(g1.disjoint(g2))
        g3 = createGeomFromWKB(POINTA)
        self.assert_(not g1.disjoint(g3))

class ContainsTest(unittest.TestCase):
    def test_contains(self):
        g1 = createGeomFromWKB(POLYGON)
        g2 = createGeomFromWKB(POINTA)
        g3 = createGeomFromWKB(POINTB)
        self.assert_(not g1.contains(g2))
        self.assert_(not g2.within(g1))
        self.assert_(g1.contains(g3))
        self.assert_(g3.within(g1))

class BufferTest(unittest.TestCase):
    def test_buffer(self):
        g1 = createGeomFromWKB(POINTA)
        g2 = g1.buffer(1.0, 64)
        self.assertAlmostEqual(g2.area(), 3.141277, 6, g2.area())

class IntersectionTest(unittest.TestCase):
    def test_intersection(self):
        g1 = createGeomFromWKB(POINTA)
        g2 = createGeomFromWKB(POLYGON)
        g3 = g1.intersection(g2)
        self.assertEqual(g3.getTypeId(), 7, g3.getTypeId())
        self.assertEqual(g1.getTypeId(), 0, g1.getTypeId())
        self.assertEqual(g2.getTypeId(), 3, g2.getTypeId())
        g1 = createGeomFromWKB(POINTB)
        g3 = g1.intersection(g2)
        self.assertEqual(g3.getTypeId(), 0, g3.getTypeId())
        self.assertEqual(g1.getTypeId(), 0, g1.getTypeId())
        self.assertEqual(g2.getTypeId(), 3, g2.getTypeId())

class GeomAccessTest(unittest.TestCase):
    def test_geoms(self):
        g1 = createGeomFromWKB(MULTIPOLYGON)
        self.assertEqual(g1.numGeometries(), 2)
        self.assertRaises(IndexError, g1.getGeometryN, 3)
        g2 = g1.getGeometryN(1)
        self.assertEqual(g2.getTypeId(), GEOS_POLYGON)
        g3 = createGeomFromWKB(POINTA)

class RingAccessTest(unittest.TestCase):
    def test_rings(self):
        g1 = createGeomFromWKB(POLYGON)
        self.assertEqual(g1.numInteriorRings(), 1)
        self.assertRaises(Exception, g1.getInteriorRingN, 2)
        g2 = g1.getExteriorRing()
        self.assertEqual(g2.getTypeId(), GEOS_LINEARRING)
        g3 = createGeomFromWKB(POINTA)
        self.assertRaises(TypeError, g3.numInteriorRings)

class PointSeqAccessTest(unittest.TestCase):
    def test_seq(self):
        g1 = createGeomFromWKB(POINTA)
        s = g1.getCoordSeq()
        self.assertEqual(s.getSize(), 1)
        self.assertEqual(s.getDimensions(), 3, s.getDimensions())
        self.assertAlmostEqual(s.getX(0), 0.0, 6)
        self.assertAlmostEqual(s.getY(0), 0.0, 6)
        self.assertAlmostEqual(s.getZ(0), 0.0, 6)
        
    def test_seq2(self):
        g1 = createGeomFromWKT('POINT (0 0)')
        s = g1.getCoordSeq()
        self.assertEqual(s.getSize(), 1)
        self.assertEqual(s.getDimensions(), 3, s.getDimensions())
        self.assertAlmostEqual(s.getX(0), 0.0, 6)
        self.assertAlmostEqual(s.getY(0), 0.0, 6)
        self.assertAlmostEqual(s.getZ(0), 0.0, 6)

class SeqAccessTest(unittest.TestCase):
    def test_seqs(self):
        g1 = createGeomFromWKB(POLYGON)
        r1 = g1.getExteriorRing()
        s = r1.getCoordSeq()
        self.assertEqual(s.getSize(), 5)
        self.assertEqual(s.getDimensions(), 3, s.getDimensions())
        self.assert_(s.getX(0) == s.getX(4) == -1)
        self.assert_(s.getY(0) == s.getY(4) == -1)
        self.assert_(s.getZ(0) == s.getZ(4))
        
    def test_seqs2(self):
        g1 = createGeomFromWKT('POLYGON ((-1 -1,-1 1,1 1,1 -1,-1 -1))')
        r1 = g1.getExteriorRing()
        s = r1.getCoordSeq()
        self.assertEqual(s.getSize(), 5)
        self.assertEqual(s.getDimensions(), 3, s.getDimensions())
        self.assert_(s.getX(0) == s.getX(4) == -1)
        self.assert_(s.getY(0) == s.getY(4) == -1, s.getY(0))
        self.assert_(s.getZ(0) == s.getZ(4))
        self.assertAlmostEqual(s.getZ(0), 0.0, 6)

class DiffTest(unittest.TestCase):
    def test_d(self):
        g1 = createGeomFromWKT('POLYGON ((-10 -10, -10 10, 10 10, 10 -10, -10 -10))')
        g2 = createGeomFromWKB(POLYGON)
        g3 = g1.difference(g2)
        self.assertEqual(g3.getTypeId(), 6)

class SequenceTest(unittest.TestCase):
    def test_new(self):
        s = createCoordSeq(4)
        self.assertEqual(s.getSize(), 4)
        # Seems like Coordinate sequences are always 3D
        self.assertEqual(s.getDimensions(), 3)
        self.assertAlmostEqual(s.getZ(0), 0.0, 6)
    def test_new0(self):
        self.assertRaises(ValueError, createCoordSeq, 0)
    def test_ords(self):
        s = createCoordSeq(5)
        s.setX(0, -1.0)
        s.setY(0, -1.0)
        self.assertAlmostEqual(s.getX(0), -1.0, 6)
        self.assertAlmostEqual(s.getY(0), -1.0, 6)
        self.assertRaises(IndexError, s.setX, 5, 2.0)

class SeqConstructTest(unittest.TestCase):
    def test_double(self):
        s = createCoordSeq(1)
        s.setX(0, 1.0)
        s = createCoordSeq(2)
        self.assertAlmostEqual(s.getX(0), 0.0)
    def test_point(self):
        s = createCoordSeq(1)
        s.setX(0, 1.0)
        s.setY(0, 1.0)
        g = createGeomFromCoordSeq(s, GEOS_POINT)
        self.assertEqual(g.getTypeId(), GEOS_POINT)
        self.assertAlmostEqual(s.getX(0), 1.0, 6)
        self.assertAlmostEqual(s.getY(0), 1.0, 6)
        s = createCoordSeq(1)
        g = createGeomFromCoordSeq(s, GEOS_POINT)
        self.assertEqual(g.getTypeId(), GEOS_POINT)
        self.assertAlmostEqual(s.getX(0), 0.0, 6)
        self.assertAlmostEqual(s.getY(0), 0.0, 6)
    def test_linestring(self):
        s = createCoordSeq(2)
        s.setX(0, 0.0)
        s.setY(0, 0.0)
        s.setX(1, 1.0)
        s.setY(1, 1.0)
        g = createGeomFromCoordSeq(s, GEOS_LINESTRING)
        self.assertEqual(g.getTypeId(), GEOS_LINESTRING)
    def test_polygon(self):
        s = createCoordSeq(5)
        s.setX(0, -2.0)
        s.setY(0, -2.0)
        s.setX(1, -2.0)
        s.setY(1, 2.0)
        s.setX(2, 2.0)
        s.setY(2, 2.0)
        s.setX(3, 2.0)
        s.setY(3, -2.0)
        s.setX(4, -2.0)
        s.setY(4, -2.0)
        # Geometry takes ownership of memory allocated to s
        shell = createGeomFromCoordSeq(s, GEOS_LINEARRING)
        # make a hole
        s = createCoordSeq(5)
        s.setX(0, -1.0)
        s.setY(0, -1.0)
        s.setX(1, -1.0)
        s.setY(1, 1.0)
        s.setX(2, 1.0)
        s.setY(2, 1.0)
        s.setX(3, 1.0)
        s.setY(3, -1.0)
        s.setX(4, -1.0)
        s.setY(4, -1.0)
        hole = createGeomFromCoordSeq(s, GEOS_LINEARRING)
        g = createPolygon(shell, [hole])
        self.assertEqual(g.getTypeId(), GEOS_POLYGON)
        self.assertEqual(g.numInteriorRings(), 1)
        self.assertAlmostEqual(g.area(), 12.0, 6)
    def test_nonseq(self):
        self.assertRaises(TypeError, createGeomFromCoordSeq, [0,0], GEOS_POINT)
    def test_nonwkb(self):
        self.assertRaises(TypeError, createGeomFromWKB, 1)
    def test_wronglen(self):
        s = createCoordSeq(2)
        self.assertRaises(ValueError, createGeomFromCoordSeq, s, GEOS_POINT)
       
# XXX: Danger, GEOS can segfault on bad WKT. Always check first.
class AnomalousWKTTest(unittest.TestCase):
    def test_badwkt(self):
        self.assertRaises(GEOSError, createGeomFromWKT, 'POINT ()')

class AnomalousWKBTest(unittest.TestCase):
    def test_badwkb(self):
        p = "110400000000000000000000000000000000000000".decode("hex")
        self.assertRaises(GEOSError, createGeomFromWKB, p)

# ============================================================================
if __name__ == '__main__':
    unittest.main()

