# $Id: testmapping.py,v 1.8 2005/04/22 03:24:15 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import os
import unittest

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
except:
    pass

from cartography.styles import *
from cartography.proj import SpatialReference
from cartography.geometry import Point
from cartography.data import ows #, disk

from environment import *

# ============================================================================
# Tests begin now

#class LayerTestCase(unittest.TestCase):
#    """testing cartography.mapping.Layer"""
#    def setUp(self):
#        self.path = os.path.join(TESTDATA_PATH, 'diskstore.ovf')
#        store = disk.DiskFeatureStore(self.path)
#        self.layer = Layer(store, 'point', maxfeatures=20)
#    def testInstance(self):
#        """Return uninitialized instance"""
#        assert isinstance(self.layer, Layer)
#    def testDatastore(self):
#        """Get back the initialize path"""
#        assert self.layer.datastore.path == self.path
#    def test_typename(self):
#        assert self.layer.typename == 'point', self.layer.typename
#        self.assertEqual(self.layer.maxfeatures, 20)

class WMSLayerTestCase(unittest.TestCase):
    """testing cartography.mapping.WMSLayer"""
    def setUp(self):
        self.url = 'http://wms.jpl.nasa.gov/wms.cgi?'
        self.store = ows.WMSRasterStore(self.url, version='1.1.1') 
        self.layer = Layer(self.store, 'layer/style')
        self.layer.ows_format = 'image/jpeg'
        self.layer.ows_srs = 'EPSG:4326'
    def testInstance(self):
        """Return uninitialized instance"""
        assert isinstance(self.layer, Layer)
    def testDatastore(self):
        """Get back the initialize path"""
        assert self.layer.datastore.url == self.url
    def test_adaption(self):
        assert self.layer.ows_format == 'image/jpeg'
        assert self.layer.ows_srs == 'EPSG:4326'

class MarkTest(unittest.TestCase):
    def test_default(self):
        m = Mark('x')
        self.assert_(m.stroke.color == '#000000')
        self.assert_(m.fill.color == '#808080')
    def test_none(self):
        m = Mark('x', stroke=None, fill=None)
        self.assert_(m.stroke == None)
        self.assert_(m.fill == None)
    
class StrokeTest(unittest.TestCase):
    def test_default(self):
        s = Stroke()
        self.assert_(s.color == '#000000')
        
class LineSymbolizerTestCase(unittest.TestCase):
    """testing mapping.LineSymbolizer"""
    def testStrokeFactory(self):
        """get a Stroke dictionary"""
        s = Stroke().todict()
        assert s['stroke'] == '#000000'
        assert s['stroke-opacity'] == 1.0
        assert s['stroke-width'] == 1.0
        assert s['stroke-dasharray'] == []
        assert s['graphicfill'] == None
        assert s['graphicstroke'] == None
    def testInitialziedStrokeFactory(self):
        """get an initialized Stroke dictionary"""
        s = Stroke('#111111', width=3).todict()
        assert s['stroke'] == '#111111'
        assert s['stroke-opacity'] == 1.0
        assert s['stroke-width'] == 3.0
        assert s['stroke-dasharray'] == []
        assert s['graphicfill'] == None
        assert s['graphicstroke'] == None
    def testDefaultSymbolizer(self):
        """create a new default line symbolizer"""
        sym = LineSymbolizer()
        assert sym.stroke.todict()['stroke'] == '#000000'
    def testNonDefaultSymbolizer(self):
        """create a new line symbolizer"""
        s = Stroke('#111111')
        sym = LineSymbolizer(s)
        assert sym.stroke.todict()['stroke'] == '#111111'
        
class PolygonSymbolizerTestCase(unittest.TestCase):
    """testing mapping.PolygonSymbolizer"""
    def testFillFactory(self):
        """get a Fill dictionary"""
        f = Fill().todict()
        assert f['fill'] == '#808080'
        assert f['fill-opacity'] == 0.5
        assert f['graphicfill'] == None
    def testFillFactoryValues(self):
        """get an initialized Fill dictionary"""
        f = Fill('#111111', 0.8).todict()
        assert f['fill'] == '#111111'
        assert f['fill-opacity'] == 0.8
        assert f['graphicfill'] == None
    def testDefaultSymbolizer(self):
        """create a new default polygon symbolizer"""
        sym = PolygonSymbolizer()
        assert sym.stroke.todict()['stroke'] == '#000000'
        assert sym.fill.todict()['fill'] == '#808080'
    def testNonDefaultSymbolizer(self):
        """create a new polygon symbolizer"""
        s = Stroke('#111111')
        f = Fill('#111111')
        sym = PolygonSymbolizer(s, f)
        assert sym.stroke.todict()['stroke'] == '#111111'
        assert sym.fill.todict()['fill'] == '#111111'
   
class PointSymbolizerTestCase(unittest.TestCase):
    """testing mapping.PointSymbolizer"""
    def testDefaultSymbolizer(self):
        """create a new default point symbolizer"""
        sym = PointSymbolizer()
        self.assert_(sym.graphic.mark.name == 'circle')
    def testNewDefaultMarkSymbolizer(self):
        """create a new mark symbolizer"""
        circle = Mark('circle')
        sym = PointSymbolizer(graphic=Graphic(mark=circle, size=8.0))
        assert sym.graphic.size == 8
        assert sym.graphic.mark.name == 'circle'
        assert sym.graphic.mark.fill.todict()['fill'] == '#808080'
        assert sym.graphic.mark.stroke.todict()['stroke'] == '#000000'
    def testNewDefaultGraphicSymbolizer(self):
        """create a new graphic symbolizer"""
        import StringIO
        house = Image()
        f = open(os.path.join(TESTIMAGE_PATH, 'home.png'), 'r')
        house.data.write(f.read())
        f.close()
        sym = PointSymbolizer(graphic=Graphic(house, size=16.0))
        self.assertEqual(sym.graphic.size, 16.0)
        f = open('house.png', 'wb')
        s = StringIO.StringIO(sym.graphic.externalgraphic.getdata())
        f.write(s.read())
        f.close()

class PlacementTest(unittest.TestCase):
    def test_tuple(self):
        p = Placement(anchorpoint=(0,0), displacement=(0,0))
        self.assertEqual(p.anchorpoint.x, 0)
        self.assertEqual(p.anchorpoint.y, 0)
        self.assertEqual(p.displacement.x, 0)
        self.assertEqual(p.displacement.y, 0)
    def test_point(self):
        class P:
            def __init__(self, x, y):
                self.x = x
                self.y = y
        p = Placement(anchorpoint=P(0,0), displacement=P(0,0))
        self.assertEqual(p.anchorpoint.x, 0)
        self.assertEqual(p.anchorpoint.y, 0)
        self.assertEqual(p.displacement.x, 0)
        self.assertEqual(p.displacement.y, 0)
        
class TextSymbolizerTestCase(unittest.TestCase):
    """testing mapping.TextSymbolizer"""
    def testDefaultSymbolizer(self):
        """create a new default text symbolizer"""
        text = TextSymbolizer()
        assert text.label == ''
        self.assertEqual(text.font.todict(),
            {'font-weight': '', 'font-size': 10.0,
             'font-family': '', 'font-style': ''} )
        assert text.placement.todict()['anchorpoint'] == None
        assert text.halo.radius == 0
        self.assertEqual(text.halo.fill.todict()['fill'].upper(), '#FFFFFF')

class RuleTestCase(unittest.TestCase):
    """testing mapping.Rule"""
    def testMakeRule(self):
        """create a rule"""
        rule = Rule('r0', None)
        assert rule.name == 'r0'
    def testDefaultScales(self):
        rule = Rule('r0', None)
        assert rule.minscaledenominator == 0
        assert rule.maxscaledenominator == 1e+20
    def testRuleScales(self):
        text = TextSymbolizer()
        rule = Rule('r0', None, None, 1000.0, 2000.0)
        assert rule.minscaledenominator == 1000
        assert rule.maxscaledenominator == 2000
        
class StyleTestCase(unittest.TestCase):
    """testing mapping.Style"""
    def testDefaultStyle(self):
        """create a new default style"""
        style = Style('s0')
        assert style.name == 's0'


# ===========================================================================
if __name__ == '__main__':
    unittest.main()

