# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import sys
import unittest
from zope.interface.verify import verifyClass

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
except:
    pass

from cartography.proj import SpatialReference
from cartography.geometry import Point, BoundingBox
from cartography.data.feature import Feature, PropertyTypeCollection
from cartography.data.feature import StringProperty, IntegerProperty, \
    GeometryProperty
from cartography.data.memory import MemoryFeatureStore, MemoryFeatureSource
from cartography.data.interfaces import IFeatureStore, IFeatureSource

class F(Feature):
    __schema__ = PropertyTypeCollection([
                    StringProperty('name'),
                    IntegerProperty('category'),
                    GeometryProperty('the_geom')
                    ])
    __typename__ = 'F'
    __srs__ = SpatialReference(epsg=4326)

class StoreTest(unittest.TestCase):
    def test_make(self):
        store = MemoryFeatureStore(F)

    def test_putget(self):
        store = MemoryFeatureStore(F)
        sink = store.source('F')
        sink.addfeature(F('1', name='foo', category=1, the_geom=Point(1,1)))
        sink.addfeature(F('2', name='bar', category=2, the_geom=Point(2,2)))

    def test_bounds(self):
        store = MemoryFeatureStore(F)
        sink = store.source('F')

        sink.addfeature(F('1', name='foo', category=1, the_geom=Point(1,1)))
        self.assertEquals(BoundingBox(1,1,1,1).totuple(),
                          store.bounds('F').totuple())

        sink.addfeature(F('2', name='bar', category=2, the_geom=Point(2,2)))
        self.assertEquals(BoundingBox(1,1,2,2).totuple(),
                          store.bounds('F').totuple())
        
        sink.addfeature(F('3', name='foobar', category=3, the_geom=Point(-50, 50)))
        self.assertEquals(BoundingBox(-50,1,2,50).totuple(),
                          store.bounds('F').totuple())

    def test_count(self):
        store = MemoryFeatureStore(F)
        sink = store.source('F')

        self.assertEquals(0, store.count('F'))

        for i in xrange(1, 100):
            sink.addfeature(F(str(i), name='foo%d' % i, category=i, the_geom=Point(i,-i)))
            self.assertEquals(i, store.count('F'))
    
    def test__expand_bounds(self):
        # Test for http://trac.gispython.org/projects/PCL/ticket/76.
        store = MemoryFeatureStore(F)
        sink = store.source('F')

        sink.addfeature(F('1', name='foo', category=1, the_geom=Point(1,1)))
        self.assertEquals(BoundingBox(1,1,1,1).totuple(),
                          store.bounds('F').totuple())

        # The bug in _expand_bounds() resulted in 'GEOMETRYCOLLECTION EMPTY'
        # being set as the bounds when a feature with the same geometry as the
        # bounds was inserted into the sink.
        sink.addfeature(F('2', name='bar', category=2, the_geom=Point(1,1)))
        self.assertEquals(BoundingBox(1,1,1,1).totuple(),
                          store.bounds('F').totuple())

    def test_typenames(self):
        store = MemoryFeatureStore(F)
        self.assertEquals(('F',), store.typenames())

    def test_info(self):
        store = MemoryFeatureStore(F)
        sink = store.source('F')
        sink.addfeature(F('1', name='foo', category=1, the_geom=Point(1,1)))
        sink.addfeature(F('2', name='bar', category=2, the_geom=Point(2,2)))
        self.assertEquals({'typename': 'F',
                           'srs': '+init=epsg:4326',
                           'bounds': (1.0, 1.0, 2.0, 2.0),
                           'schema': {'category': 'int', 'the_geom': 'geometry', 'name': 'string'}},
                          store.info('F'))


    def test_interface(self):
        self.failUnless(verifyClass(IFeatureStore, MemoryFeatureStore))
        self.failUnless(verifyClass(IFeatureSource, MemoryFeatureSource))
        
       

# =============================================================================
if __name__ == '__main__':
    unittest.main()

