# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import os
import sys
import unittest

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
except ImportError:
    pass
    
from cartography.data import postgis
from cartography.data.feature import *
from cartography.geometry import Polygon, Point, MultiPolygon, BoundingBox

# ============================================================================
# Test Store, Source, and Feature Type
# ============================================================================

# TODO: replace this dependency on the world borders dataset with a better
# fixture. For now see PCL-Data/fixtures/makefile_postgis.

from _testing import *

TYPENAMES = ['postgis.world_borders', 'postgis.us_borders', 'postgis.uk_borders']
TYPENAME0 = 'postgis.world_borders'
SCHEMA0 = [ RealProperty('area'),
            IntegerProperty('cat'),
            StringProperty('cntry_name'),
            IntegerProperty('gid'),
            RealProperty('pop_cntry'),
            GeometryProperty('the_geom', 'MultiPolygon'),
            StringProperty('fips_cntry') ]
INFO0 = { 'area': 'float', 'cat': 'int', 'cntry_name': 'string', 
          'gid': 'int', 'pop_cntry': 'float', 'the_geom': 'geometry',
          'fips_cntry': 'string'}
TYPENAME1 = 'postgis.uk_borders'
SCHEMA1 = SCHEMA0
INFO1 = INFO0
PROP0_NAME = 'gid'
PROP1_NAME = 'area'

class NewStoreTest(unittest.TestCase):
    def test_typenames(self):
        store = postgis.PGFeatureStore(DSN)
        typenames = store.typenames()
        self.assertEqual(typenames, TYPENAMES)
    def test_typenames_schema(self):
        store = postgis.PGFeatureStore(DSN, ['postgis'])
        typenames = store.typenames()
        self.assertEqual(typenames, 
            ['postgis.world_borders', 'postgis.us_borders', 'postgis.uk_borders'])

class TypeActivationTest(unittest.TestCase):
    def test_activation(self):
        store = postgis.PGFeatureStore(DSN, empty=True)
        typenames = store.typenames()
        self.assert_(typenames == [], typenames)
        store.activate('postgis.world_borders')
        self.assert_(store.typenames() == ['postgis.world_borders'])
        self.assertRaises(KeyError, store.activate, 'postgis.world_borders')
        self.assertRaises(KeyError, store.deactivate, 'postgis.bogus')
        store.deactivate('postgis.world_borders')
        self.assert_(store.typenames() == [])
        

class UserTypeTest(unittest.TestCase):
    def test_usertype(self):
        store = postgis.PGFeatureStore(DSN, empty=True)
        class TestFeature(postgis.PGSelectionFeature):
            __typename__ = TYPENAME_CUSTOM
            __schema__ = PropertyTypeCollection(SCHEMA_CUSTOM)
            __sql_from__ = "postgis.world_borders"
            __sql_select__ = "gid, the_geom, fips_cntry"
            __sql_where__ = "fips_cntry = 'UK'"
            __gid__ = "gid"
            __srid__ = 4269
        self.assert_(TestFeature.typename() == TYPENAME_CUSTOM)
        self.assert_(TestFeature.schema().names() == INFO_CUSTOM.keys())
        self.assertEqual(TestFeature.defaultgeometry().name, 'the_geom')
        self.assertEqual(TestFeature.defaultsrs().tostring(), '+init=epsg:4269')
        store.addfeaturetype(TestFeature)
        typenames = store.typenames()
        self.assert_(typenames == ['custom'], typenames)
        store.deactivate('custom')
        self.assert_(store.typenames() == [])
    def test_usertype_multi(self):
        store = postgis.PGFeatureStore(DSN, empty=True)
        class TestFeature(postgis.PGSelectionFeature):
            __typename__ = TYPENAME_CUSTOM
            __schema__ = PropertyTypeCollection(SCHEMA_CUSTOM + [GeometryProperty('centroid')])
            __sql_from__ = "postgis.world_borders"
            __sql_select__ = "gid, the_geom, fips_cntry, centroid(the_geom)"
            __sql_where__ = "fips_cntry = 'UK'"
            __gid__ = "gid"
            __g__ = "centroid"
            __srid__ = 4269
        self.assertEqual(TestFeature.typename(), TYPENAME_CUSTOM)
        self.assertEqual(TestFeature.schema().names(), ['gid', 'the_geom', 'fips_cntry', 'centroid'])
        self.assertEqual(TestFeature.defaultgeometry().name, 'centroid')
        self.assertEqual(TestFeature.defaultsrs().tostring(), '+init=epsg:4269')
        store.addfeaturetype(TestFeature)
        typenames = store.typenames()
        self.assert_(typenames == ['custom'], typenames)
        store.deactivate('custom')
        self.assert_(store.typenames() == [])


# ============================================================================
# Store tests
# ============================================================================

class StoreTest(unittest.TestCase):
    def setUp(self):
        self.store = postgis.PGFeatureStore(DSN)
    # Helper
    def assertBoundsAlmostEqual(self, b1, b2, arg):
        try:
            for c1, c2 in zip(b1, b2):
                self.assertAlmostEqual(c1, c2, 6)
        except:
            raise AssertionError, arg

# Test that a PostgreSQL table can be turned into a feature type
class TableTest(StoreTest):
    def test_featuretype(self):
        ftype = self.store.featuretype(TYPENAME0)
        self.assert_(ftype.typename() == TYPENAME0)
        self.assertEqual(ftype.schema().names(),
            ['area', 'cat', 'cntry_name', 'fips_cntry', 'gid',
                'pop_cntry', 'the_geom'])
        self.assertEqual(ftype.defaultgeometry().name, 'the_geom')
        self.assertEqual(ftype.defaultsrs().tostring(), '+init=epsg:4269')
    def test_source(self):
        source = self.store.source(TYPENAME0)
        assert isinstance(source, postgis.PGFeatureSource)
    def test_source_query(self):
        f = "fips_cntry = 'FOO'"
        source = self.store.source(TYPENAME0, properties=f)
        assert isinstance(source, postgis.PGFeatureSource)
    def test_bounds(self):
        bounds = self.store.bounds(TYPENAME0)
        self.assert_(isinstance(bounds, BoundingBox))
        bounds_tuple = bounds.totuple()
        self.assert_(len(bounds_tuple) == 4, bounds_tuple)
        self.assertBoundsAlmostEqual(bounds_tuple,
                (-180.0, -90.0, 180.0, 83.623596191406193), bounds_tuple)
    def test_bounds_fast(self):
        bounds = self.store.bounds(TYPENAME0, fastest=True)
        self.assert_(isinstance(bounds, BoundingBox))
        bounds_tuple = bounds.totuple()
        self.assert_(len(bounds_tuple) == 4, bounds_tuple)
        self.assertBoundsAlmostEqual(bounds_tuple,
         (-180.0, -90.0, 180.0, 83.623596191406193), bounds_tuple)
    def test_count(self):
        count = self.store.count(TYPENAME0)
        self.assert_(count == 3784, count)
    def test_info(self):
        info = self.store.info(TYPENAME0)
        self.assert_(info['typename'] == TYPENAME0, info)
        self.assertEqual(info['schema'], INFO0)
   
# Test that a view can be turned into a feature type
class ViewTest(StoreTest):
    def test_featuretype(self):
        ftype = self.store.featuretype(TYPENAME1)
        self.assert_(ftype.typename() == TYPENAME1)
        self.assertEqual(ftype.schema().names(),
            ['area', 'cat', 'cntry_name', 'fips_cntry',
             'gid', 'pop_cntry', 'the_geom'])
        self.assertEqual(ftype.defaultgeometry().name, 'the_geom')
        self.assertEqual(ftype.defaultsrs().tostring(), '+init=epsg:4269')
    def test_source(self):
        source = self.store.source(TYPENAME1)
        assert isinstance(source, postgis.PGFeatureSource)
    def test_source_query(self):
        f = "fips_cntry = 'FOO'"
        source = self.store.source(TYPENAME1, properties=f)
        assert isinstance(source, postgis.PGFeatureSource)
    def test_bounds(self):
        bounds = self.store.bounds(TYPENAME1)
        self.assert_(isinstance(bounds, BoundingBox))
        bounds_tuple = bounds.totuple()
        self.assert_(len(bounds_tuple) == 4, bounds_tuple)
        self.assertBoundsAlmostEqual(bounds_tuple,
                (-8.6213893890380895, 49.911655426025398, 1.74944400787354, 60.844444274902301), bounds_tuple)
    def test_bounds_fast(self):
        bounds = self.store.bounds(TYPENAME1, fastest=True)
        self.assert_(isinstance(bounds, BoundingBox))
        bounds_tuple = bounds.totuple()
        self.assert_(len(bounds_tuple) == 4, bounds_tuple)
        self.assertBoundsAlmostEqual(bounds_tuple,
                (-8.6213893890380895, 49.911655426025398, 1.74944400787354, 60.844444274902301), bounds_tuple)
    def test_count(self):
        count = self.store.count(TYPENAME1)
        self.assert_(count == 48, count)
    def test_info(self):
        info = self.store.info(TYPENAME1)
        self.assert_(info['typename'] == TYPENAME1, info)
        self.assertEqual(info['schema'], INFO1)

# ============================================================================
# Selection Feature Type
# ============================================================================

TYPENAME_CUSTOM = 'custom'
SCHEMA_CUSTOM = [ IntegerProperty('gid'),
                  GeometryProperty('the_geom'),
                  StringProperty('fips_cntry') ]
INFO_CUSTOM = {'gid': 'int', 'the_geom': 'geometry', 'fips_cntry': 'string'}

class SelectionTest(StoreTest):
    def setUp(self):
        self.store = postgis.PGFeatureStore(DSN, empty=True)
        class TestFeature(postgis.PGSelectionFeature):
            __typename__ = TYPENAME_CUSTOM
            __schema__ = PropertyTypeCollection(SCHEMA_CUSTOM)
            __sql_from__ = "postgis.world_borders"
            __sql_select__ = "gid, the_geom, fips_cntry"
            __sql_where__ = "fips_cntry = 'UK'"
            __gid__ = "gid"
            __srid__ = 4269
        self.store.addfeaturetype(TestFeature)
    def test_featuretype(self):
        ftype = self.store.featuretype(TYPENAME_CUSTOM)
        self.assertEqual(ftype.typename(), TYPENAME_CUSTOM)
        self.assertEqual(ftype.schema().names(), INFO_CUSTOM.keys())
        self.assertEqual(ftype.defaultgeometry().name, 'the_geom')
        self.assertEqual(ftype.defaultsrs().tostring(), '+init=epsg:4269')
    def test_source(self):
        source = self.store.source(TYPENAME_CUSTOM)
        assert isinstance(source, postgis.PGFeatureSource)
    def test_source_query(self):
        f = "fips_cntry = 'FOO'"
        source = self.store.source(TYPENAME_CUSTOM, properties=f)
        assert isinstance(source, postgis.PGFeatureSource)
    def test_bounds(self):
        bounds = self.store.bounds(TYPENAME_CUSTOM)
        self.assert_(isinstance(bounds, BoundingBox))
        bounds_tuple = bounds.totuple()
        self.assert_(len(bounds_tuple) == 4, bounds_tuple)
        self.assertBoundsAlmostEqual(bounds_tuple,
                (-8.6213893890380895, 49.911655426025398, 1.74944400787354, 60.844444274902301), bounds_tuple)
    def test_bounds_fast(self):
        bounds = self.store.bounds(TYPENAME_CUSTOM, fastest=True)
        self.assert_(isinstance(bounds, BoundingBox))
        bounds_tuple = bounds.totuple()
        self.assert_(len(bounds_tuple) == 4, bounds_tuple)
        self.assertBoundsAlmostEqual(bounds_tuple,
                (-8.6213893890380895, 49.911655426025398, 1.74944400787354, 60.844444274902301), bounds_tuple)
    def test_count(self):
        count = self.store.count(TYPENAME_CUSTOM)
        self.assert_(count == 48, count)
    def test_info(self):
        info = self.store.info(TYPENAME_CUSTOM)
        self.assert_(info['typename'] == TYPENAME_CUSTOM, info)
        self.assertEqual(info['schema'], INFO_CUSTOM)

# ============================================================================
# Source and Access Tests
# ============================================================================

class TableSourceTest(unittest.TestCase):
    def setUp(self):
        self.store = postgis.PGFeatureStore(DSN)
        self.source = postgis.PGFeatureSource(self.store, TYPENAME0)
    def test_featuretype(self):
        ftype = self.source.featuretype
        self.assert_(ftype.typename() == TYPENAME0)
        self.assertEqual(ftype.schema().names(),
            ['area', 'cat', 'cntry_name', 'fips_cntry', 'gid', 
                'pop_cntry', 'the_geom'])

class TableAccessTest(unittest.TestCase):
    def setUp(self):
        self.store = postgis.PGFeatureStore(DSN)
        self.source = self.store.source(TYPENAME0)
    def test_unfiltered(self):
        source = self.store.source(TYPENAME0)
        ftype = source.featuretype
        fiter = source()
        feature = fiter.next()
        self.assert_(isinstance(feature, ftype))
        self.assert_(feature.id == '1', feature.id)
        self.assert_(feature.properties['cntry_name'] == 'Aruba')
    # back-end filtering
    def test_false(self):
        # expect 0 features when filter is globally False
        f = "fips_cntry = 'FOO'"
        source = self.store.source(TYPENAME0, properties=f)
        fiter = source()
        self.assertRaises(StopIteration, fiter.next)
    def test_true(self):
        f = "fips_cntry = 'UK'"
        source = self.store.source(TYPENAME0, properties=f)
        fiter = source()
        feature = fiter.next()
        self.assert_(feature.properties['fips_cntry'] == 'UK')
        self.assert_(isinstance(feature.properties.the_geom, MultiPolygon), feature.properties.the_geom)
        # Check the total number of results 
        i = 1
        for feature in fiter:
            i += 1
        self.assert_(i == 48, i)
    def test_bbox(self):
        b = BoundingBox(-10, 45, 10, 60)
        source = self.store.source(TYPENAME0, bbox=b)
        fiter = source()
        feature = fiter.next()
        self.assert_(feature.properties['fips_cntry'] == 'BE')
        self.assert_(isinstance(feature.properties.the_geom, MultiPolygon), feature.properties.the_geom)
        i = 1
        for feature in fiter:
            i += 1
        self.assert_(i == 107, i)
    def test_targeted_bbox(self):
        b = BoundingBox(-10, 45, 10, 60)
        source = self.store.source(TYPENAME0, bbox=b, geometry='the_geom')
        fiter = source()
        feature = fiter.next()
        self.assert_(feature.properties['fips_cntry'] == 'BE')
        self.assert_(isinstance(feature.properties.the_geom, MultiPolygon), feature.properties.the_geom)
        i = 1
        for feature in fiter:
            i += 1
        self.assert_(i == 107, i)
    def test_front_atts(self):
        b = BoundingBox(-10, 45, 10, 60)
        source = self.store.source(TYPENAME0, bbox=b)
        def func(s):
            return s.lower()
        ff = "func(f.fips_cntry) == 'uk'"
        fiter = source(filter=ff, func=func)
        i = 0
        for feature in fiter:
            i += 1
        self.assert_(i == 46, i)
    def test_front_geom(self):
        b = BoundingBox(-10, 45, 10, 60)
        source = self.store.source(TYPENAME0, bbox=b)
        def func(s):
            return s.lower()
        p = Point(0.0, 51.2)
        ff = "g.contains(p)"
        fiter = source(filter=ff, func=func, p=p)
        i = 0
        for feature in fiter:
            i += 1
        self.assert_(i == 1, i)

class ViewAccessTest(unittest.TestCase):
    def setUp(self):
        self.store = postgis.PGFeatureStore(DSN)
        self.source = self.store.source(TYPENAME1)
    def test_unfiltered(self):
        source = self.store.source(TYPENAME1)
        ftype = source.featuretype
        fiter = source()
        feature = fiter.next()
        self.assert_(isinstance(feature, ftype))
        self.assert_(feature.id == '3393', feature.id)
        self.assert_(feature.properties['cntry_name'] == 'United Kingdom', feature.properties.cntry_name)

class SelectionAccessTest(unittest.TestCase):
    def setUp(self):
        self.store = postgis.PGFeatureStore(DSN, empty=True)
        class TestFeature(postgis.PGSelectionFeature):
            __typename__ = TYPENAME_CUSTOM
            __schema__ = PropertyTypeCollection(SCHEMA_CUSTOM + [GeometryProperty('centroid')])
            __sql_from__ = "postgis.world_borders"
            __sql_select__ = "gid, the_geom, fips_cntry, centroid(the_geom)"
            __sql_where__ = "fips_cntry = 'UK'"
            __gid__ = "gid"
            __g__ = 'the_geom'
            __srid__ = 4269
        self.store.addfeaturetype(TestFeature)
        self.source = self.store.source(TYPENAME_CUSTOM)
    def test_unfiltered(self):
        source = self.store.source(TYPENAME_CUSTOM)
        ftype = source.featuretype
        fiter = source()
        feature = fiter.next()
        self.assert_(isinstance(feature, ftype))
        self.assert_(feature.id == '3393', feature.id)
        self.assert_(feature.properties['fips_cntry'] == 'UK', feature.properties.fips_cntry)
    def test_filtered(self):
        source = self.store.source(TYPENAME_CUSTOM)
        ftype = source.featuretype
        p = Point(0.0, 51.2)
        fiter = source(filter='g.contains(p)', p=p)
        feature = fiter.next()
        self.assert_(isinstance(feature, ftype))
        self.assert_(feature.id == '3438', feature.id)
        self.assert_(feature.properties['fips_cntry'] == 'UK', feature.properties.fips_cntry)
    def test_targeted_bbox(self):
        b = BoundingBox(-10, 45, 10, 60)
        source = self.store.source(TYPENAME_CUSTOM, bbox=b, geometry='centroid')
        fiter = source()
        feature = fiter.next()
        self.assert_(feature.properties['fips_cntry'] == 'UK')
        self.assert_(isinstance(feature.properties.the_geom, MultiPolygon), feature.properties.the_geom)
        i = 1
        for feature in fiter:
            i += 1
        self.assert_(i == 45, i)

        
# ============================================================================
# Feature Type SQL Tests
# ============================================================================

try:
    import psycopg2 as psycopg
except ImportError:
    import psycopg

class TableSQLTest(unittest.TestCase):
    def setUp(self):
        class TestFeature(postgis.PGFeature):
            __typename__ = TYPENAME0
            __schema__ = SCHEMA0
        self.ftype = TestFeature
        self.conn = psycopg.connect(DSN)
        self.cursor = self.conn.cursor()
    def tearDown(self):
        self.conn.close()
        self.conn = self.cursor = None
    def test_extent_sql(self):
        sql = self.ftype.extent_sql()
        self.assert_(sql == "SELECT extent(the_geom), area(extent(the_geom)) FROM postgis.world_borders ORDER BY area DESC LIMIT 1", sql)
        self.cursor.execute(sql)
        row = self.cursor.fetchone()
        self.assert_(row == ('BOX(-180 -90,180 83.6235961914062)', 62504.494628906199), row)
    def test_estimated_extent_sql(self):
        sql = self.ftype.estimated_extent_sql()
        self.assert_(sql == "SELECT estimated_extent('postgis', 'world_borders', 'the_geom'), area(estimated_extent('postgis', 'world_borders', 'the_geom')) ORDER BY area DESC LIMIT 1", sql)
        self.cursor.execute(sql)
        row = self.cursor.fetchone()
        self.assert_(row == ('BOX(-180 -90,180 83.6235961914062)', 62504.494628906199), row)

class ViewSQLTest(unittest.TestCase):
    def setUp(self):
        class TestFeature(postgis.PGFeature):
            __typename__ = TYPENAME1
            __schema__ = SCHEMA1
        self.ftype = TestFeature
        self.conn = psycopg.connect(DSN)
        self.cursor = self.conn.cursor()
    def tearDown(self):
        self.conn.close()
        self.conn = self.cursor = None
    def test_extent_sql(self):
        sql = self.ftype.extent_sql()
        self.assert_(sql == "SELECT extent(the_geom), area(extent(the_geom)) FROM postgis.uk_borders ORDER BY area DESC LIMIT 1", sql)
        self.cursor.execute(sql)
        row = self.cursor.fetchone()
        self.assert_(row == ('BOX(-8.62138938903809 49.9116554260254,1.74944400787354 60.8444442749023)', 113.38213171531601), row)
    def test_estimated_extent_sql(self):
        sql = self.ftype.estimated_extent_sql()
        self.assert_(sql.strip() == "SELECT estimated_extent('postgis', 'uk_borders', 'the_geom'), area(estimated_extent('postgis', 'uk_borders', 'the_geom')) ORDER BY area DESC LIMIT 1", sql)

class SelectionSQLTest(unittest.TestCase):
    def setUp(self):
        class TestFeature(postgis.PGSelectionFeature):
            __typename__ = TYPENAME_CUSTOM
            __schema__ = SCHEMA_CUSTOM
            __sql_from__ = "postgis.world_borders"
            __sql_select__ = "gid, the_geom, fips_cntry"
            __sql_where__ = "fips_cntry = 'UK'"
            __gid__ = "gid"
            __srid__ = 4269
        self.ftype = TestFeature
        self.conn = psycopg.connect(DSN)
        self.cursor = self.conn.cursor()
    def tearDown(self):
        self.conn.close()
        self.conn = self.cursor = None
    def test_extent_sql(self):
        sql = self.ftype.extent_sql()
        self.assert_(sql == "SELECT extent(the_geom), area(extent(the_geom)) FROM postgis.world_borders WHERE fips_cntry = 'UK' ORDER BY area DESC LIMIT 1", sql)
        self.cursor.execute(sql)
        row = self.cursor.fetchone()
        self.assert_(row == ('BOX(-8.62138938903809 49.9116554260254,1.74944400787354 60.8444442749023)', 113.38213171531601), row)
    def test_estimated_extent_sql(self):
        sql = self.ftype.estimated_extent_sql()
        self.assert_(sql == "SELECT extent(the_geom), area(extent(the_geom)) FROM postgis.world_borders WHERE fips_cntry = 'UK' ORDER BY area DESC LIMIT 1", sql)
        self.cursor.execute(sql)
        row = self.cursor.fetchone()
        self.assert_(row == ('BOX(-8.62138938903809 49.9116554260254,1.74944400787354 60.8444442749023)', 113.38213171531601), row)
    def test_feature_sql_noquery(self):
        sql = self.ftype.feature_sql('gid')
        self.assertEqual(sql.strip(), "SELECT gid, the_geom, fips_cntry FROM postgis.world_borders WHERE fips_cntry = 'UK'", sql)

#class ComplexSelectionSQLTest(unittest.TestCase):
#    def setUp(self):
#        class TestFeature(postgis.PGSelectionFeature):
#            __typename__ = TYPENAME_CUSTOM
#            __schema__ = {'gid': PT_INT, 'the_geom': PT_GEOMETRY, 'fips_cntry': PT_STRING}
#            __sql_from__ = "world_borders"
#            __sql_select__ = "gid, simplify(the_geom, 2) as the_geom, fips_cntry"
#            __sql_where__ = "fips_cntry = 'UK'"
#            __gid__ = "gid"
#            __srid__ = 4269
#        self.ftype = TestFeature
#        self.cursor = psycopg.connect(DSN).cursor()
#    def test_extent_sql(self):
#        sql = self.ftype.extent_sql()
#        self.assert_(sql == "SELECT extent(simplify(the_geom, 2)), area(extent(simplify(the_geom, 2))) FROM world_borders WHERE fips_cntry = 'UK' ORDER BY area DESC LIMIT 1", sql)
#        self.cursor.execute(sql)
#        row = self.cursor.fetchone()
#        self.assert_(row == ('BOX(-8.62138938903809 49.9116554260254,1.74944400787354 60.8444442749023)', 113.38213171531601), row)
#    def test_feature_sql_noquery(self):
#        sql = self.ftype.feature_sql('gid', Query())
#        self.assertEqual(sql, "SELECT gid, simplify(the_geom, 2) as the_geom, fips_cntry FROM world_borders WHERE fips_cntry = 'UK' AND 1=1", sql)
#       
#class Complex2SelectionSQLTest(unittest.TestCase):
#    def setUp(self):
#        class TestFeature(postgis.PGSelectionFeature):
#            __typename__ = TYPENAME_CUSTOM
#            __schema__ = {'gid': PT_INT, 'simplify': PT_GEOMETRY, 'fips_cntry': PT_STRING}
#            __sql_from__ = "world_borders"
#            __sql_select__ = "gid, simplify(the_geom, 2), fips_cntry"
#            __sql_where__ = "fips_cntry = 'UK'"
#            __gid__ = "gid"
#            __srid__ = 4269
#        self.ftype = TestFeature
#        self.cursor = psycopg.connect(DSN).cursor()
#    # XXX: extent of the simplified geometry is probably overkill
#    #def test_extent_sql(self):
#    #    sql = self.ftype.extent_sql()
#    #    self.assert_(sql == "SELECT extent(simplify(the_geom, 2)), area(extent(simplify(the_geom, 2))) FROM world_borders WHERE fips_cntry = 'UK' ORDER BY area DESC LIMIT 1", sql)
#    #    self.cursor.execute(sql)
#    #    row = self.cursor.fetchone()
#    #    self.assert_(row == ('BOX(-8.62138938903809 49.9116554260254,1.74944400787354 60.8444442749023)', 113.38213171531601), row)
#    def test_feature_sql_noquery(self):
#        sql = self.ftype.feature_sql('gid', Query())
# #       self.assertEqual(sql, "SELECT gid, simplify(the_geom, 2), fips_cntry FROM world_borders WHERE fips_cntry = 'UK' AND 1=1", sql)
        


# =============================================================================
if __name__ == '__main__':
    unittest.main()

