# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import unittest

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
except:
    pass
    
from cartography.geometry import *
from cartography.proj import SpatialReference, ProjTransform

from _testing import *


class GeometryTest(unittest.TestCase):
    def test_empty(self):
        g = Geometry()
        self.assertEqual(g.toWKT(), 'GEOMETRYCOLLECTION EMPTY')
        self.assertEqual(g.srs, None)
        g2 = g.copy()
        self.assertEqual(g2.toWKT(), 'GEOMETRYCOLLECTION EMPTY')
    
class PointTest(unittest.TestCase):
    # TODO: test len
    def test_empty(self):
        g = Point()
        self.assertEqual(g.isEmpty(), 1)
        self.assert_(isinstance(g, Point))
        self.assertEqual(g.geometryType(), 'Point')
        self.assertEqual(g.srs, None)
    def test_srs(self):
        g = Point(srs=SpatialReference(epsg=4326))
        self.assertEqual(g.isEmpty(), 1)
        self.assert_(isinstance(g, Point))
        self.assertEqual(g.geometryType(), 'Point')
        self.assertEqual(g.srs.tostring(), '+init=epsg:4326')
    def test_xy(self):
        g = Point(1.0, -1.0)
        self.assert_(isinstance(g, Point))
        self.assertAlmostEqual(g.x, 1.0, 7)
        self.assertAlmostEqual(g.y, -1.0, 7)
        self.assert_(isinstance(g.envelope(), BoundingBox))
    def test_xyz(self):
        g = Point(1.0, -1.0, 3.0)
        self.assert_(isinstance(g, Point))
        self.assertAlmostEqual(g.x, 1.0, 7)
        self.assertAlmostEqual(g.y, -1.0, 7)
        self.assertAlmostEqual(g.z, 3.0, 7)
    def test_wkb(self):
        g = Geometry.fromWKB(POINTA)
        self.assert_(isinstance(g, Point))
        self.assertAlmostEqual(g.x, 0.0, 7)
        self.assertAlmostEqual(g.y, 0.0, 7)
    def test_set(self):
        g = Geometry.fromWKB(POINTA)
        g.x = 2.0
        g.y = 2.0
        g.z = 100.0
        self.assertAlmostEqual(g.x, 2.0, 7)
        self.assertAlmostEqual(g.y, 2.0, 7)
        self.assertAlmostEqual(g.z, 100.0)

class LineStringTest(unittest.TestCase):
    def test_empty(self):
        g = LineString()
        self.assertEqual(g.isEmpty(), 1)
        self.assert_(isinstance(g, LineString))
        self.assertEqual(g.geometryType(), 'LineString')
        self.assertEqual(g.srs, None)
    def test_srs(self):
        g = LineString(srs=SpatialReference(epsg=4326))
        self.assertEqual(g.isEmpty(), 1)
        self.assert_(isinstance(g, LineString))
        self.assertEqual(g.geometryType(), 'LineString')
        self.assertEqual(g.srs.tostring(), '+init=epsg:4326')
    def test_points(self):
        g = Geometry.fromWKB(LINESTRING)
        self.assertEqual(g.numPoints(), 2)
        self.assertAlmostEqual(g.pointN(0).x, -2.0)
        self.assertAlmostEqual(g.pointN(1).x, 2.0)
        self.assertRaises(IndexError, g.pointN, 2)
        self.assertAlmostEqual(g.startPoint().x, -2.0)
        self.assertAlmostEqual(g.endPoint().x, 2.0)
        e = g.envelope()
        self.assert_(isinstance(e, BoundingBox))
    def test_length(self):
        g = Geometry.fromWKB(LINESTRING)
        self.assertEqual(g.length(), 4)
        
class LinearRingTest(unittest.TestCase):
    def test_empty(self):
        g = LinearRing()
        self.assertEqual(g.isEmpty(), 1)
        self.assert_(isinstance(g, LinearRing))
        self.assertEqual(g.geometryType(), 'LinearRing')
        self.assertEqual(g.srs, None)
    def test_srs(self):
        g = LinearRing(srs=SpatialReference(epsg=4326))
        self.assertEqual(g.isEmpty(), 1)
        self.assert_(isinstance(g, LinearRing))
        self.assertEqual(g.geometryType(), 'LinearRing')
        self.assertEqual(g.srs.tostring(), '+init=epsg:4326')
    def test_points(self):
        g = LinearRing([Point(-2,-2), Point(-2,2), Point(2,2), Point(2,-2), Point(-2,-2)])
        self.assertEqual(g.length(), 16.0)
    def test_bad(self):
        self.assertRaises(GEOSError, LinearRing, [Point(-2,-2), Point(-2,2), Point(2,2), Point(2,-2)])
                        
class PolygonTest(unittest.TestCase):
    def test_empty(self):
        g = Polygon()
        self.assert_(isinstance(g, Polygon))
        self.assertEqual(g.isEmpty(), 1)
        self.assertEqual(g.geometryType(), 'Polygon')
        self.assertEqual(g.srs, None)
    def test_srs(self):
        g = Polygon(srs=SpatialReference(epsg=4326))
        self.assert_(isinstance(g, Polygon))
        self.assertEqual(g.isEmpty(), 1)
        self.assertEqual(g.geometryType(), 'Polygon')
        self.assertEqual(g.srs.tostring(), '+init=epsg:4326')
    def test_shell(self):
        shell = LinearRing([Point(-2,-2), Point(-2,2), Point(2,2), Point(2,-2), Point(-2,-2)])
        g = Polygon(shell)
        self.assertEqual(g.numInteriorRings(), 0)
        self.assertAlmostEqual(g.area(), 16.0)
    def test_holes(self):
        shell = LinearRing([Point(-2,-2), Point(-2,2), Point(2,2), Point(2,-2), Point(-2,-2)])
        hole = LinearRing([Point(-1,-1), Point(-1,1), Point(1,1), Point(1,-1), Point(-1,-1)])
        g = Polygon(shell, [hole])
        self.assertEqual(g.numInteriorRings(), 1)
        self.assertAlmostEqual(g.area(), 12.0)
    def test_wkb(self):
        g = Geometry.fromWKB(POLYGON)
        self.assert_(isinstance(g, Polygon))
    def test_rings(self):
        g = Geometry.fromWKB(POLYGON)
        r = g.exteriorRing()
        self.assert_(isinstance(r, LinearRing))
        self.assertEqual(r.getTypeId(), 2)
        self.assertEqual(g.numInteriorRings(), 1)
        r = g.interiorRingN(0)
        self.assert_(isinstance(r, LinearRing))
    def test_centroid(self):
        g = Geometry.fromWKB(POLYGON)
        x = g.centroid()
        self.assert_(isinstance(x, Point))

class MultiPointTest(unittest.TestCase):
    def test_empty(self):
        g = MultiPoint()
        self.assert_(g.isEmpty())
        self.assertEqual(g.geometryType(), 'MultiPoint')
        self.assertEqual(g.srs, None)
    def test_srs(self):
        g = MultiPoint(srs=SpatialReference(epsg=4326))
        self.assert_(g.isEmpty())
        self.assertEqual(g.geometryType(), 'MultiPoint')
        self.assertEqual(g.srs.tostring(), '+init=epsg:4326')
    def test_points(self):
        g = MultiPoint([Point(-1,-1), Point(1,1)])
        self.assert_(isinstance(g, MultiPoint))
        self.assertEqual(g.numGeometries(), 2)
        self.assertAlmostEqual(g.envelope().area(), 4.0)

class MultiLineStringTest(unittest.TestCase):
    def test_empty(self):
        g = MultiLineString()
        self.assertEqual(g.isEmpty(), 1)
        self.assertEqual(g.geometryType(), 'MultiLineString')
        self.assertEqual(g.srs, None)
    def test_srs(self):
        g = MultiLineString(srs=SpatialReference(epsg=4326))
        self.assertEqual(g.isEmpty(), 1)
        self.assertEqual(g.geometryType(), 'MultiLineString')
        self.assertEqual(g.srs.tostring(), '+init=epsg:4326')
    def test_lines(self):
        g = MultiLineString([LineString([Point(1,1), Point(2,2)])])
        self.assert_(isinstance(g, MultiLineString))
        self.assertEqual(g.numGeometries(), 1)
        x = g.geometryN(0)
        self.assert_(isinstance(x, LineString))
        self.assertAlmostEqual(x.length(), 1.41421356)
    def test_wkb(self):
        g = Geometry.fromWKB(MULTILINESTRING)
        self.assert_(isinstance(g, MultiLineString))
        self.assertEqual(g.numGeometries(), 2)

class MultiPolygonTest(unittest.TestCase):
    def test_empty(self):
        g = MultiPolygon()
        self.assertEqual(g.isEmpty(), 1)
        self.assertEqual(g.geometryType(), 'MultiPolygon')
        self.assertEqual(g.srs, None)
    def test_srs(self):
        g = MultiPolygon(srs=SpatialReference(epsg=4326))
        self.assertEqual(g.isEmpty(), 1)
        self.assertEqual(g.geometryType(), 'MultiPolygon')
        self.assertEqual(g.srs.tostring(), '+init=epsg:4326')
    def test_polygons(self):
        p = Polygon(LinearRing([Point(-2,-2), Point(-2,2), Point(2,2), Point(2,-2), Point(-2,-2)]))
        g = MultiPolygon([p])
        self.assert_(isinstance(g, MultiPolygon))
        self.assertEqual(g.numGeometries(), 1)
        x = g.geometryN(0)
        self.assert_(isinstance(x, Polygon))
    def test_wkb(self):
        g = Geometry.fromWKB(MULTIPOLYGON)
        self.assert_(isinstance(g, MultiPolygon))
        self.assertEqual(g.numGeometries(), 2)

class BoundingBoxTest(unittest.TestCase):
    def test_empty(self):
        b = BoundingBox()
        self.assertEqual(b.isEmpty(), 1)
        self.assertEqual(b.srs, None)
    def test_validate(self):
        self.assertRaises(ValueError, BoundingBox, 1.0)
    def test_validate2(self):
        self.assertRaises(ValueError, BoundingBox, 2.0, -1.0, 1.0, 1.0)
    def test_new(self):
        b = BoundingBox(-1.0,-1.0,1.0,1.0)
        self.assertEqual(b.dimension, 2)
        self.assertEqual(b.getExteriorRing().isRing(), 1)
        self.assertEqual(b.minx, -1.0)
        self.assertEqual(b.maxx, 1.0)
        self.assertEqual(b.miny, -1.0)
        self.assertEqual(b.maxy, 1.0)
        self.assertEqual(b.contains(Point(0,0)), 1)
        self.assertEqual(b.contains(Point(-1.5,0)), 0)
    def test_modify(self):
        b = BoundingBox(-1.0,-1.0,1.0,1.0)
        b.minx = -2.0
        b.miny = -2.0
        b.maxx = 0.0
        b.maxy = 1.0
        # this point is right on the boundary
        self.assertEqual(b.intersects(Point(-1.5,0.99999)), 1)
        self.assertEqual(b.contains(Point(-1.5,0.99999)), 1)
        t = b.totuple()
        self.assertAlmostEqual(t[0], -2.0)
        b2 = b.copy()
        self.assertEqual(b2.toWKT(), 'POLYGON ((-2.0000000000000000 -2.0000000000000000, -2.0000000000000000 1.0000000000000000, 0.0000000000000000 1.0000000000000000, 0.0000000000000000 -2.0000000000000000, -2.0000000000000000 -2.0000000000000000))')

    def test_size(self):
        b = BoundingBox(-10, -10, 10, 10)
        self.assertEquals(20, b.width)
        self.assertEquals(20, b.height)
        
        b = BoundingBox(-20, -20, -10, -10)
        self.assertEquals(10, b.width)
        self.assertEquals(10, b.height)
        
        b = BoundingBox(10, 10, 20, 20)
        self.assertEquals(10, b.width)
        self.assertEquals(10, b.height)

    def test_transform(self):
        b = BoundingBox(3083821.75, 6639903.0, 3733174.25, 7779714.0)
        source = SpatialReference(epsg=2393)
        target = SpatialReference(epsg=4326)
        transformation = ProjTransform(target, source)
        b.transform(transformation)
        self.assertAlmostEqual(b.totuple()[0], 19.604105965419816)

class PointBoundingBoxTest(unittest.TestCase):
    def test_new(self):
        #b = Point(-1.0, -1.0).envelope()
        b = BoundingBox(-1.0,-1.0)
        self.assertEqual(b.dimension, 0)
        self.assertEqual(b.minx, -1.0)
        self.assertEqual(b.maxx, -1.0)
        self.assertEqual(b.miny, -1.0)
        self.assertEqual(b.maxy, -1.0)
        self.assertEqual(b.intersects(Point(-1.0, -1.0)), 1)
        self.assertEqual(b.contains(Point(-1.0, -1.0)), 1)
        self.assertEqual(b.contains(Point(-1.5, 0)), 0)

class OperationsTest(unittest.TestCase):
    def test_mismatched_srs(self):
        l = Geometry.fromWKB(LINESTRING, SpatialReference(epsg=4326))
        p = Geometry.fromWKB(POLYGON)
        self.assertRaises(SRSError, p.intersection, l)
    def test_intersection(self):
        l = Geometry.fromWKB(LINESTRING)
        p = Geometry.fromWKB(POLYGON)
        x = p.intersection(l)
        self.assertEqual(x.geometryType(), 'MultiLineString')
        self.assertEqual(x.envelope().totuple(), (-1.0, 0.0, 1.0, 0.0))
    def test_boundary(self):
        l = Geometry.fromWKB(LINESTRING)
        x = l.boundary()
        self.assertEqual(x.geometryType(), 'MultiPoint')
        self.assertEqual(x.envelope().totuple(), (-2.0, 0.0, 2.0, 0.0))
    def test_buffer(self):
        l = Geometry.fromWKB(LINESTRING)
        x = l.buffer(0.5, 24)
        self.assertEqual(x.geometryType(), 'Polygon')
        self.assertEqual(x.envelope().totuple(), (-2.5, -0.5, 2.5, 0.5))
    def test_hull(self):
        p = Geometry.fromWKB(POLYGON)
        x = p.convexHull()
        self.assertEqual(x.geometryType(), 'Polygon')
        self.assertEqual(x.envelope().totuple(), (-1.0, -1.0, 1.0, 1.0))
    def test_union(self):
        l = Geometry.fromWKB(LINESTRING)
        p = Geometry.fromWKB(POLYGON)
        x = p.union(l)
        self.assertEqual(x.geometryType(), 'GeometryCollection')
        self.assertEqual(x.envelope().totuple(), (-2.0, -1.0, 2.0, 1.0))
    def test_difference(self):
        l = Geometry.fromWKB(LINESTRING)
        p = Geometry.fromWKB(POLYGON)
        x = p.difference(l)
        self.assertEqual(x.geometryType(), 'Polygon')
        self.assertEqual(x.envelope().totuple(), (-1.0, -1.0, 1.0, 1.0))
    def test_symdiff(self):
        l = Geometry.fromWKB(LINESTRING)
        p = Geometry.fromWKB(POLYGON)
        x = p.symDifference(l)
        self.assertEqual(x.geometryType(), 'GeometryCollection')
        self.assertEqual(x.envelope().totuple(), (-2.0, -1.0, 2.0, 1.0))
    def test_boxes(self):
        b1 = BoundingBox(-2,-2,1,1)
        b2 = BoundingBox(-1,-1,2,2)
        x = b1.intersection(b2)
        self.assertEqual(x.geometryType(), 'Polygon')
        self.assertEqual(x.envelope().totuple(), (-1.0, -1.0, 1.0, 1.0))
    def test_boxes_intersect(self):
        b1 = BoundingBox(-2,-2,1,1)
        b2 = BoundingBox(-1,-1,0,0)
        self.assert_(b1.intersects(b2))
    def test_boxes_intersect_degenerate(self):
        b1 = BoundingBox(-2,-2,1,1)
        b2 = BoundingBox(-1,0)
        self.assert_(b1.intersects(b2))
    def test_box_point_intersect(self):
        b1 = BoundingBox(-2,-2,1,1)
        b2 = Point(-1,-1)
        self.assert_(b1.intersects(b2))

class PickleTest(unittest.TestCase):
    def test_pickle(self):
        import pickle
        o = Geometry.fromWKB(POLYGON)
        d = pickle.dumps(o)
        del o
        g = pickle.loads(d)
        self.assert_(isinstance(g, Polygon))
        r = g.exteriorRing()
        self.assert_(isinstance(r, LinearRing))
        self.assertEqual(r.getTypeId(), 2)
        self.assertEqual(g.numInteriorRings(), 1)
        r = g.interiorRingN(0)
        self.assert_(isinstance(r, LinearRing))
        

# ============================================================================
if __name__ == '__main__':
    unittest.main()

