# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
#
# $Id$
# =============================================================================

import os
import unittest

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
except:
    pass

from cartography.context import *
from cartography.styles import *
from cartography.proj import SpatialReference
from cartography.geometry import Point

from environment import *


class ViewTest(unittest.TestCase):
    def test_view(self):
        v = View(None, [-10,-10,10,10])
        self.assertEqual(v.getaspectratio(), 1.0)
        self.assertAlmostEqual(v.getscaledenominator([100, 100]), 714.285714, 5)
        p = Point(0, 0)
        self.assert_(v.contains(p))
        p = Point(-20, 0)
        self.assert_(not v.contains(p))
        v.units = 'DD'
        self.assertAlmostEqual(v.getscaledenominator([100, 100]), 79513921.995, 3)
        v.units = 'FEET'
        self.assertAlmostEqual(v.getscaledenominator([100, 100]), 217.770034, 5)
    
    def test_getviewforscale(self):
        v0 = View(None, [-10,-10,10,10], 'METERS')
        
        v = v0.getviewforscale(v0.getscaledenominator([100,100]), [100,100])
        self.failUnless(v0.ul.equals(v.ul))
        self.failUnless(v0.lr.equals(v.lr))
        
        v0.units = 'DD'
        v = v0.getviewforscale(v0.getscaledenominator([100,100]), [100,100])
        self.failUnless(v0.ul.equals(v.ul))
        self.failUnless(v0.lr.equals(v.lr))

        v0.units = 'FEET'
        v = v0.getviewforscale(v0.getscaledenominator([100,100]), [100,100])
        self.failUnless(v0.ul.equals(v.ul))
        self.failUnless(v0.lr.equals(v.lr))

class ContextLayerTest(unittest.TestCase):
    def test_layer(self):
        layer = Layer(None, None)
        style = Style()
        c = ContextLayer('foo', layer, style)
        self.assertEqual(c.name, 'foo')
        self.assertEqual(c.title, 'foo')
        self.assertEqual(c.visible, 1)
        self.assertEqual(c.layer, layer)
        self.assertEqual(c.style, style)
        c.visible = 0
        self.assertEqual(c.visible, False)
       
class MapContextTest(unittest.TestCase):
    def test_map(self):
        layer = Layer(None, None)
        style = Style()
        c = ContextLayer('foo', layer, style)
        m = MapContext([c])
        self.assertEqual(len(m), 1)
        self.assertEqual(len(m._layers), 1)
        self.assertEqual(len(m._names), 1)
        self.assertRaises(KeyError, m.append, c)
        self.assertRaises(KeyError, m.insert, 0, c)
        c = ContextLayer('bar', layer, style)
        m.append(c)
        self.assertEqual(len(m), 2)
        self.assertEqual(m.index('foo'), 0)
        self.assertEqual(m.index('bar'), 1)
        c = m[-1]
        self.assertEqual(c.name, 'bar')
        c = m[0]
        self.assertEqual(c.name, 'foo')
        self.assertRaises(KeyError, m.__setitem__,
                          0, ContextLayer('bar', layer, style))
        self.assertEqual([c.name for c in m], ['foo', 'bar'])
        m[0] = ContextLayer('you', layer, style)
        self.assertEqual(len(m), 2)
        self.assertEqual(m[0].name, 'you')
        del m[0]
        self.assertEqual(len(m), 1)
        self.assertEqual(m[0].name, 'bar')
        m.insert(0, ContextLayer('foo', layer, style))
        self.assertEqual(len(m), 2)
 

# ===========================================================================
if __name__ == '__main__':
    unittest.main()

