# $Id: testsrs.py,v 1.23 2005/05/19 17:15:06 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2005 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import unittest

import pkg_resources
pkg_resources.require('PCL-Core')

from cartography.proj import SpatialReference

class ProjTest(unittest.TestCase):
    def setUp(self):
        self.pitems = [ '+proj=aea',
                        '+ellps=clrk66',
                        '+lat_1=40.8333333',
                        '+lat_2=64.1666666',
                        '+lat_0=25.0',
                        '+lon_0=0.0',
                        '+x_0=0.0',
                        '+y_0=0.0',
                        '+units=m' ]
        self.pstring = ' '.join(self.pitems)     
    def testCreate(self):
        srs = SpatialReference(self.pstring)
        self.assert_(isinstance(srs, SpatialReference))
    def testToString(self):        
        srs = SpatialReference(self.pstring)
        self.assertEqual(srs.tostring(), self.pstring)
    def testToString2(self):        
        srs = SpatialReference(proj4=self.pstring)
        self.assertEqual(srs.tostring(), self.pstring)
    def testPlussedList(self):
        srs = SpatialReference(self.pitems)
        self.assertEqual(srs.tostring(), self.pstring)
    def testNonPlussedList(self):
        items = [item.replace('+', '') for item in self.pitems]
        srs = SpatialReference(items)
        self.assertEqual(srs.tostring(), self.pstring)
    def testNonPlussed(self):
        s = self.pstring.replace('+', '')
        srs = SpatialReference(s)
        self.assertEqual(srs.tostring(), self.pstring)
    def testStr(self):
        srs = SpatialReference(self.pstring)
        self.assertEqual(str(srs), self.pstring)
            
class EpsgTest(unittest.TestCase):
    def testCreate(self):
        srs = SpatialReference(epsg=4326)
        self.assert_(isinstance(srs, SpatialReference))
    def testToString(self):
        srs = SpatialReference(epsg=4326)
        self.assertEqual(srs.tostring(), '+init=epsg:4326')
    def testStr(self):
        srs = SpatialReference(epsg=4326)
        self.assertEqual(str(srs), '+init=epsg:4326')

class EqualsTest(unittest.TestCase):
    def testEPSGtoESPG(self):
        srs1 = SpatialReference(epsg=2393)
        srs2 = SpatialReference(epsg=4326)
        self.failIf(srs1.equals(srs2))
        self.failIf(srs2.equals(srs1))
        self.failUnless(srs1.equals(srs1))
        self.failUnless(srs2.equals(srs2))

    def testPROJ4toPROJ4(self):
        srs1 = SpatialReference(proj4='+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')
        srs2 = SpatialReference(proj4='+proj=tmerc +lat_0=0 +lon_0=27 +k=1.000000 +x_0=3500000 +y_0=0 +ellps=intl +towgs84=-90.7,-106.1,-119.2,4.09,0.218,-1.05,1.37 +units=m +no_defs')
        self.failIf(srs1.equals(srs2))
        self.failIf(srs2.equals(srs1))
        self.failUnless(srs1.equals(srs1))
        self.failUnless(srs2.equals(srs2))

    #def testPROJ4toEPSG(self):
    #    srs1_1 = SpatialReference(proj4='+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')
    #    srs1_2 = SpatialReference(epsg=4326)
    #    srs2_1 = SpatialReference(proj4='+proj=tmerc +lat_0=0 +lon_0=27 +k=1.000000 +x_0=3500000 +y_0=0 +ellps=intl +towgs84=-90.7,-106.1,-119.2,4.09,0.218,-1.05,1.37 +units=m +no_defs')
    #    srs2_2 = SpatialReference(epsg=2393)
    #
    #    self.failUnless(srs1_1.equals(srs1_2),
    #                    "%s || %s" % (str(srs1_1), str(srs1_2)))
    #    self.failUnless(srs2_1.equals(srs2_2),
    #                    "%s || %s" % (str(srs2_1), str(srs2_2)))
    #    self.failIf(srs1_1.equals(srs2_2))
    #    self.failIf(srs2_1.equals(srs1_2))

    #def testEPSGtoPROJ4(self):
    #    srs1_1 = SpatialReference(epsg=4326)
    #    srs1_2 = SpatialReference(proj4='+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')
    #    srs2_1 = SpatialReference(epsg=2393)
    #    srs2_2 = SpatialReference(proj4='+proj=tmerc +lat_0=0 +lon_0=27 +k=1.000000 +x_0=3500000 +y_0=0 +ellps=intl +towgs84=-90.7,-106.1,-119.2,4.09,0.218,-1.05,1.37 +units=m +no_defs')
    #
    #    self.failUnless(srs1_1.equals(srs1_2))
    #    self.failUnless(srs2_1.equals(srs2_2))
    #    self.failIf(srs1_1.equals(srs2_2))
    #    self.failIf(srs2_1.equals(srs1_2))

    def test_nonsemantic_differences(self):
        srs1 = SpatialReference(proj4='+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')
        srs2 = SpatialReference(proj4=' +proj=longlat   +datum=WGS84  +ellps=WGS84   +no_defs ')
        self.failUnless(srs1.equals(srs2))

    def test_magic_methods(self):
        srs1 = SpatialReference(epsg=4326)
        srs2 = SpatialReference(proj4='+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')
        srs3 = SpatialReference(epsg=2393)
        
        self.failUnless(srs1 == srs1)
        #self.failUnless(srs1 == srs2)
        #self.failIf(srs1 != srs2)
        self.failIf(srs1 == srs3)
        self.failUnless(srs1 != srs3)

    def testInvalidArgument(self):
        srs = SpatialReference(epsg=4326)
        self.assertRaises(ValueError, srs.equals, ("4326"))
        self.assertRaises(ValueError, srs.equals, ("4326"))

# =============================================================================
if __name__ == '__main__':
    unittest.main()

