# $Id: timewquery.py,v 1.1 2005/02/10 16:10:07 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

# =============================================================================
# This script is for racing the PCL filtered feature access vs the mapscript
# query API.
#
# PCL's API is much more powerful, but we don't want to fall very far behind
# mapscript in performance.
#
# Uses Erle Schuyler's world countries shapefile from mappinghacks.com
# =============================================================================

import os
import sys
import timeit
import unittest
import mapscript

sys.path.insert(0, '../../..')
import cartography

from cartography import data, spatial
from cartography.filter import PythonFilter

TESTPATH = '/home/sean/projects/ms_44/mapserver/tests/'
TESTMAPFILE = os.path.join(TESTPATH, 'test.map')

WORLD_COUNTRIES_SHP = '/var/wms_data/world/world_countries.shp'
mapobj = mapscript.mapObj(TESTMAPFILE)
mapobj.setExtent(-20.0,25.0,60.0,75.0)
l = mapobj.getLayer(1)
l.labelitem = 'FIPS_CNTRY'
l.classitem = 'FIPS_CNTRY'
l.data = WORLD_COUNTRIES_SHP

r = mapobj.extent
bbox = '%f,%f,%f,%f' % (r.minx, r.miny, r.maxx, r.maxy)
path = WORLD_COUNTRIES_SHP
store = data.DiskFeatureStore(path)

# =========================================================================
# Test 1A: layerObj.queryByRect()

s1a = """\
l.template = 'foo'
l.queryByRect(mapobj, r)
results = l.getResults()
l.open()
for i in range(results.numresults):
    result = results.getResult(i)
    shape = l.getFeature(result.shapeindex, result.tileindex)
    atts = [shape.getValue(j) for j in range(l.numitems)]
l.close()
print "1A: %d" % (results.numresults)
"""

def run1a():
    exec(s1a)

# =========================================================================
# Test 1B: data.DiskFeatureStore.iterfeatures()

s1b = """\
query = data.Query(filter=PythonFilter('g.BBOX(%s)' % (bbox)))
i = 0
for g, a in store.iterfeatures(query):
    i += 1
print "1B: %d" % (i)
"""

def run1b():
    exec(s1b)

# =========================================================================
# Test 2A: layerObj.queryByAttributes()

s2a = """\
l.template = 'foo'
l.queryByAttributes(mapobj, 'FIPS_CNTRY', 'UK', 1)
results = l.getResults()
l.open()
for i in range(results.numresults):
    result = results.getResult(i)
    shape = l.getFeature(result.shapeindex, result.tileindex)
    atts = [shape.getValue(j) for j in range(l.numitems)]
l.close()
print "2A: %d" % (results.numresults)
"""

def run2a():
    exec(s2a)

# =========================================================================
# Test 2B: data.DiskFeatureStore.iterfeatures()

s2b = """\
query = data.Query(filter=PythonFilter('g.BBOX(%s) and f.FIPS_CNTRY == "UK"' % (bbox)))
i = 0
for g, a in store.iterfeatures(query):
    i += 1
print "2B: %d" % (i)
"""

def run2b():
    exec(s2b)
    
# =========================================================================
# Test 3A: layerObj.queryByPoint()

p3a = mapscript.pointObj(0.0, 51.5)
s3a = """\
l.template = 'foo'
l.queryByPoint(mapobj, p3a, 1, -1.0)
results = l.getResults()
l.open()
for i in range(results.numresults):
    result = results.getResult(i)
    shape = l.getFeature(result.shapeindex, result.tileindex)
    atts = [shape.getValue(j) for j in range(l.numitems)]
l.close()
print "3A: %d" % (results.numresults)
"""

def run3a():
    exec(s3a)
    
# =========================================================================
# Test 3B: data.DiskFeatureStore.iterfeatures()

p3b = spatial.Point(0.0, 51.5)
s3b = """\
query = data.Query(filter=PythonFilter('g.BBOX(%s) and g.intersects(p)' % (bbox)), p=p3b)
i= 0
for g, a in store.iterfeatures(query):
    i += 1
print "3B: %d" % (i)
"""

def run3b():
    exec(s3b)

# ============================================================================
# Run timing tests

if __name__ == '__main__':

    print "Test 1A: mapscript.layerObj.queryByRect()"
    t = timeit.Timer(stmt=s1a, setup='from __main__ import mapobj, l, r')
    print "%.2f usec/pass" % (1000000 * t.timeit(number=10)/10)

    print "Test 1B: data.DiskFeatureStore.iterfeatures()"
    t = timeit.Timer(stmt=s1b, setup='from __main__ import data, PythonFilter, store, bbox')
    print "%.2f usec/pass" % (1000000 * t.timeit(number=10)/10)

    print "Test 2A: mapscript.layerObj.queryByAttributes()"
    t = timeit.Timer(stmt=s2a, setup='from __main__ import mapobj, l, r')
    print "%.2f usec/pass" % (1000000 * t.timeit(number=10)/10)

    print "Test 2B: data.DiskFeatureStore.iterfeatures()"
    t = timeit.Timer(stmt=s2b, setup='from __main__ import data, PythonFilter, store, bbox')
    print "%.2f usec/pass" % (1000000 * t.timeit(number=10)/10)

    print "Test 3A: mapscript.layerObj.queryByPoint()"
    t = timeit.Timer(stmt=s3a, setup='from __main__ import mapobj, l, p3a')
    print "%.2f usec/pass" % (1000000 * t.timeit(number=10)/10)

    print "Test 3B: data.DiskFeatureStore.iterfeatures()"
    t = timeit.Timer(stmt=s3b, setup='from __main__ import data, PythonFilter, store, p3b, bbox')
    print "%.2f usec/pass" % (1000000 * t.timeit(number=10)/10)


