# =============================================================================
# Python Cartographic Library. Copyright (C) 2004, 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import os

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
    pkg_resources.require('PCL-GDAL')
except:
    pass

from cartography.proj import SpatialReference
from cartography.geometry import Geometry, BoundingBox
from cartography.data.feature import Feature, PropertyTypeCollection
from cartography.data.feature import IntegerProperty, RealProperty, \
    StringProperty, GeometryProperty
from cartography.data.interfaces import IFeatureStore, IDataStore, IRasterStore

import gdal
import ogr
import osr

from zope.interface import implements
    
# ============================================================================
# Local Disk Data
# ============================================================================

GEOM_MAP = {
    1: 'Point', 
    2: 'LineString', 
    3: 'Polygon', 
    4: 'MultiPoint', 
    5: 'MultiLineString', 
    6: 'MultiPolygon', 
    7: 'GeometryCollection', 
    100: 'None', 
    101: 'LinearRing'
    }

class DiskDataStore:

    """Base class for data stores on disk"""
    
    
class DiskFeatureStore(DiskDataStore):
    
    """Represents a directory of files on disk
    
    Implements cartography.interfaces.IFeatureStore. Backed up by GDAL/OGR's
    ogr.py module.
    """
    
    implements(IFeatureStore)

    def __init__(self, path=''):
        """Initialize

        Parameters
        ----------
        path : string
            An absolute path to a directory on the file system
        """
        self.path = path
        self._featuretypes = {}
        self._initialize_featuretypes()

    def _initialize_featuretypes(self):
        # Use OGR to interrogate the data, then dynamically create feature
        # types, and populate the dict
        ptypes = [IntegerProperty, IntegerProperty, RealProperty, RealProperty,
                  StringProperty, StringProperty]
        if not os.access(self.path, os.F_OK):
            raise Exception, "can not access %s" % (self.path)
        ds = ogr.Open(self.path, update=0)
        for i in range(ds.GetLayerCount()):
            layer = ds.GetLayer(i)
            typename = layer.GetName()
            ldef = layer.GetLayerDefn()
            sref = layer.GetSpatialRef()
            if not sref:
                raise Exception, "Data source %s@%s has no spatial reference" \
                      % (typename, self.path)
            proj4 = sref.ExportToProj4().strip()
            srs = SpatialReference(proj4)
            schema = []
            for j in range(ldef.GetFieldCount()):
                fdef = ldef.GetFieldDefn(j)
                schema.append(ptypes[fdef.GetType()](fdef.GetName()))
            # OGR has a single geometry
            schema.append(GeometryProperty('the_geom', homosrs=srs,
                            homotype=GEOM_MAP[ldef.GetGeomType()]))
            ftype = type("%s_%sFeature" % (ds.GetName(), typename),
                (Feature,), dict(__schema__=PropertyTypeCollection(schema),
                                 __typename__=typename,
                                 __g__='the_geom', __srs__=srs))
            self._featuretypes[typename] = ftype
        ds.Destroy()

    def _get_ogr_datasource(self):
        """object return must be explicitly destroyed with Destroy()"""
        return ogr.Open(self.path, update=0)
  
    def typenames(self):
        """Return list of feature type names"""
        return self._featuretypes.keys()

    def featuretype(self, typename):
        """Return the feature type object by its name.
        
        Parameters
        ----------
        typename : string
            Name of a feature type
        """
        return self._featuretypes[typename]
        
    def source(self, typename, bbox=None, properties=None, maxfeatures=0):
        """Return a PGFeatureSource tapped into a feature type.
        
        Parameters
        ----------
        typename : string
            Name of a feature type
        bbox : iterable
            List or tuple like [minx, miny, maxx, maxy]. Only features that
            intersect with this bounding box will issue from the source.
        properties : string
            An SQL-like WHERE clause in terms of feature properties such as
            "fname = 'foo' AND ftype = 2".
        maxfeatures : int
            Maximum number of features to return.
        """
        return DiskFeatureSource(self, typename, bbox, properties, maxfeatures)
          
    def count(self, typename):
        """Return number of features of a type in the store.
        
        Parameters
        ----------
        typename : string
            Name of a feature type
        """
        ds = self._get_ogr_datasource()
        count = ds.GetLayerByName(typename).GetFeatureCount()
        ds.Destroy()
        return count

    def bounds(self, typename, fastest=False):
        """Return bounds of all features of a type.
        
        Parameters
        ----------
        typename : string
            Name of a feature type
        fastest : boolean
            If True, use the PostGIS estimated_extent function to get the
            fastest, but less accurate, estimate of the bounds.
        """
        ds = self._get_ogr_datasource()
        e = ds.GetLayerByName(typename).GetExtent()
        ds.Destroy()
        return BoundingBox(e[0], e[2], e[1], e[3])

    def info(self, typename):
        """Return an informational dict about a feature type.
                
        Parameters
        ----------
        typename : string
            Name of a feature type

        Returns a dict like 
        { 'typename': 'street', 
          'schema': {'name': 'string', 'type': 'int'},
          'srs': 'init=epsg:26913',
          'bounds': (400000, 4000000, 400400, 4000400) }
        """
        ft = self.featuretype(typename)
        p = ft.schema_info()
        n = self.count(typename)
        b = self.bounds(typename).totuple()
        srs = ft.defaultsrs()
        return {'typename': typename, 'schema': p, 
                'srs': srs.tostring(), 'bounds': b}


class DiskFeatureSource:

    def __init__(self, store, typename, bbox=None, properties=None,
                 maxfeatures=0):
        """Initialize

        Parameters
        ----------
        store : IDiskFeatureStore
            object that implements IDiskFeatureStore
        typename : string
            name of one of the types in the store
        bbox : iterable
            List or tuple like [minx, miny, maxx, maxy]. Only features that
            intersect with this bounding box will issue from the source.
        properties : string
            An SQL-like WHERE clause in terms of feature properties such as
            "fname = 'foo' AND ftype = 2".
        maxfeatures : int
            Maximum number of features to return.
        """
        self.store = store
        self.typename = typename
        self._bbox = bbox
        self._properties = properties
        self._maxfeatures = maxfeatures
        self._ds = None

    def __del__(self):
        """Clean up any possible open OGR connections"""
        if self._ds:
            self._ds.Destroy()
    
    def _get_ogr_datasource(self):
        # object returned must be explicitly destroyed with Destroy()
        return ogr.Open(self.store.path, update=0)
        
    def _get_featuretype(self):
        # Return the feature type object
        return self.store.featuretype(self.typename)
    
    featuretype = property(_get_featuretype)
    
    def iterfeatures(self, filter=None, **params):
        """Return iterator over features, applying optional front-end filter.
        
        Parameters
        ----------
        filter : string
            A Python expression
        params : mapping
            Environment in which to evaluate the filter expression

        Usage
        -----
        >>> for feature in a_source():
        >>>     # use feature     
        """
        return DiskFeatureIterator(self, self._bbox, self._properties, self._maxfeatures, filter, **params)

    __call__ = iterfeatures


class DiskFeatureIterator:
    
    """Iterates over stored features of a single type"""
    
    def __init__(self, source, bbox=None, properties=None, maxfeatures=0,
                 filter=None, **params):
        """Initialize

        The source parameter is a reference to the feature source.
        """
        self._ds = ogr.Open(source.store.path, update=0)
        self._layer = self._ds.GetLayerByName(source.typename)
        self._layer.ResetReading()
        self._typename = source.typename
        self._featuretype = source.featuretype
        self._schema_info = self._featuretype.schema_info()
        self._maxfeatures = maxfeatures
        self._findex = 0
        
        # --
        ldef = self._layer.GetLayerDefn()
        self._fields = []
        for j in range(ldef.GetFieldCount()):
            fdef = ldef.GetFieldDefn(j)
            self._fields.append(fdef.GetName())

        self.hasfrontfilter = 0
        if filter is not None:
            self._compiled_filter = compile(filter, 'string', 'eval')
            self._locals = {}
            self._locals.update(params)
            self.hasfrontfilter = 1

        # Possible bounding box filter
        self._f_geom = None
        if bbox:
            wkt = "POLYGON((%f %f, %f %f, %f %f, %f %f, %f %f))" \
                % (bbox.minx, bbox.miny, bbox.minx, bbox.maxy,
                   bbox.maxx, bbox.maxy, bbox.maxx, bbox.miny,
                   bbox.minx, bbox.miny)
            self._f_geom = ogr.CreateGeometryFromWkt(wkt)
            self._layer.SetSpatialFilter(self._f_geom)
        if properties:
            self._layer.SetAttributeFilter(properties)

    def __del__(self):
        """Delete instance"""
        if self._f_geom:
            self._f_geom.Destroy()
        self._ds.Destroy()

    def __iter__(self):
        """Return self, following the Python iterator protocol"""
        return self

    def next(self):
        """Return next feature"""
        # Stop if we've hit the limit
        if self._maxfeatures:
            if self._findex >= self._maxfeatures:
                raise StopIteration

        # -------------------------------------------------------------------
        # Step through features till one matches the provided filter
        # -------------------------------------------------------------------
        
        feature = None
        
        while 1:
            
            fi = self._layer.GetNextFeature()
            if not fi:
                raise StopIteration

            atts = []
            si = self._schema_info
            for i in range(len(self._fields)):
                field = self._fields[i]
                ft = self._schema_info[field]
                if ft == 'int':
                    val = fi.GetFieldAsInteger(i)
                elif ft == 'float':
                    val = fi.GetFieldAsDouble(i)
                elif ft == 'string':
                    val = fi.GetFieldAsString(i)
                else: # don't get here
                    raise Exception, "invalid field type"
                atts.append(val)

            gref = fi.GetGeometryRef()
            try:
                the_geom = Geometry.fromWKB(gref.ExportToWkb())
            except:
                the_geom = Geometry.fromWKT(gref.ExportToWkt())
            d = dict(zip(self._fields + ['the_geom'], atts + [the_geom]))
            feature = self._featuretype(fi.GetFID(), pdict=d)
            f = feature.properties
            g = feature.properties.the_geom
            
            matches = True
            if self.hasfrontfilter:
                self._locals.update({'g': g, 'f': f})
                matches = eval(self._compiled_filter, globals(), self._locals)
                
            if not matches:
                del self._locals['g']
                del self._locals['f']
                del gref
                fi.Destroy()
                continue
           
            del gref
            fi.Destroy()
            break
        
        self._findex += 1
        return feature

# ============================================================================
# Raster
# ============================================================================

class DiskRasterStore(DiskDataStore):
    
    """Represents a raster file on disk.
    
    Instances are handed directly to the underlying mapping engine and so yet
    have no methods.
    """

    implements(IRasterStore)

    def __init__(self, path=''):
        """Initialize.

        Parameters
        ----------
        path : string
            An absolute path to a raster file on the file system
        """
        if not os.access(path, os.F_OK):
            raise Exception, "Cannot access %s" % (path)
        self.path = path

    def typenames(self):
        """Return names of types.

        Presently, our raster store contains a single self-named type
        """
        return [os.path.basename(self.path)]

    def bounds(self, typename=None):
        """Return bounds of the data."""
        ds = gdal.Open(self.path)
        srs = osr.SpatialReference()
        srs.ImportFromWkt(ds.GetProjection())
        proj4 = srs.ExportToProj4()
        proj = proj4.strip()
        g = ds.GetGeoTransform()
        size = (ds.RasterXSize, ds.RasterYSize)
        del ds
        return BoundingBox(g[0], g[3] + g[5]*size[1], g[0] + g[1]*size[0], g[3],
                srs=SpatialReference(proj))
        
    def info(self, typename=None):
        """Return an informational dict, implementing IDataSource"""
        ds = gdal.Open(self.path)
        srs = osr.SpatialReference()
        srs.ImportFromWkt(ds.GetProjection())
        proj4 = srs.ExportToProj4()
        proj = proj4.strip()
        srs = SpatialReference(proj).tostring()
        g = ds.GetGeoTransform()
        size = (ds.RasterXSize, ds.RasterYSize)
        p = []
        for i in range(ds.RasterCount):
            band = ds.GetRasterBand(i+1)
            type = band.DataType
            j = BANDTYPEVALUES.index(type)
            name = BANDTYPENAMES[j]
            p.append(['Band %d' % (i+1), name])
        del ds
        b = (g[0], g[3] + g[5]*size[1], g[0] + g[1]*size[0], g[3])
        r = ((b[2] - b[0]) / size[0], (b[3] - b[1]) / size[1])
        return {'typename': typename or self.typenames()[0], 
                'srs': srs, 'bounds': b,
                'bands': p, 'size': size, 'resolution': r}

BANDTYPENAMES = [s[4:] for s in dir(gdal) if s.startswith('GDT')]
BANDTYPEVALUES = [eval('gdal.%s' % (s)) for s in dir(gdal) \
                  if s.startswith('GDT')]


