/*
# =============================================================================
# OSR Reference Utilities Module  Copyright(C) 2006 Howard C. Butler
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: hobu@hobu.net
# =============================================================================
*/ 

#include "Python.h"
#include <ogr_srs_api.h>
#include <cpl_port.h>
#include <cpl_error.h>

static PyObject *OSRError;

void ThrowIfCPLError(OGRErr err)
{
    if (err)
    {
        PyErr_SetString(OSRError, CPLGetLastErrorMsg());
        CPLErrorReset();    
    }
}

PyObject* to_wkt(OGRSpatialReferenceH *ref)
{
    char* wkt;
    OGRErr err;
    PyObject* out_string;
    err = OSRExportToWkt(ref, &wkt);
    if (!err)
    {
        out_string= PyString_FromString(wkt);
        free(wkt);
        return out_string;
    }
    else
    {
        ThrowIfCPLError(err);    
        return NULL;
    }
}

PyObject* to_xml(OGRSpatialReferenceH *ref)
{
    char* wkt;
    OGRErr err;
    PyObject* out_string;
    err = OSRExportToXML(ref, &wkt, NULL);
    if (!err)
    {
        out_string= PyString_FromString(wkt);
        free(wkt);
        return out_string;
    }
    else
    {
        ThrowIfCPLError(err);    
        return NULL;
    }
}

PyObject* to_proj4(OGRSpatialReferenceH *ref)
{
    char* wkt;
    OGRErr err;
    PyObject* out_string;
    err = OSRExportToProj4(ref, &wkt);
    if (!err)
    {
        out_string= PyString_FromString(wkt);
        free(wkt);
        return out_string;
    }
    else
    {
        ThrowIfCPLError(err);    
        return NULL;
    }
}
PyObject* to_esri(OGRSpatialReferenceH *ref)
{
    char* wkt;
    OGRErr err;
    PyObject* out_string;
    err = OSRMorphToESRI(ref);

    if (err)
    {
        ThrowIfCPLError(err);    
        return NULL;
    }

    err = OSRExportToWkt(ref, &wkt);
    if (!err)
    {
        out_string= PyString_FromString(wkt);
        free(wkt);
        return out_string;
    }
    else
    {
        ThrowIfCPLError(err);    
        return NULL;
    }
}


OGRSpatialReferenceH * from_esri(char* ref_defn)
{
    OGRErr err;
    OGRSpatialReferenceH ref=NULL;
    char *list[2] = {NULL,NULL};

    ref = OSRNewSpatialReference(NULL);

    list[0] = ref_defn;
    list[1] = NULL;
    err = OSRImportFromESRI(ref, list);
    if (err)
    {
        ThrowIfCPLError(err);
        return NULL;
    }
    return ref;   
}

OGRSpatialReferenceH * from_epsg(char* ref_defn)
{
    OGRErr err;
    OGRSpatialReferenceH ref=NULL;

    ref = OSRNewSpatialReference(NULL);
    err = OSRImportFromEPSG(ref, atoi(ref_defn));
    if (err)
    {
        ThrowIfCPLError(err);
        return NULL;
    }
    return ref;
}

OGRSpatialReferenceH * from_proj4(char* ref_defn)
{
    OGRErr err;
    OGRSpatialReferenceH ref=NULL;

    ref = OSRNewSpatialReference(NULL);
    err = OSRImportFromProj4(ref, ref_defn);
    if (err)
    {
        ThrowIfCPLError(err);
        return NULL;
    }
    return ref;
}

OGRSpatialReferenceH * from_xml(char* ref_defn)
{
    OGRErr err;
    OGRSpatialReferenceH ref=NULL;

    ref = OSRNewSpatialReference(NULL);
    err = OSRImportFromXML(ref, ref_defn);
    if (err)
    {
        ThrowIfCPLError(err);
        return NULL;
    }
    return ref;
}

OGRSpatialReferenceH * from_wkt(char* ref_defn)
{
    OGRErr err;
    OGRSpatialReferenceH ref=NULL;

    ref = OSRNewSpatialReference(NULL);
    err = OSRImportFromWkt(ref, &ref_defn);
    if (err)
    {
        ThrowIfCPLError(err);
        return NULL;
    }
    return ref;
}

int validate_input(char* input)
{
    if (!strcmp(input,"epsg"))
        return 1;
    if (!strcmp(input,"esri"))
        return 1;
    if (!strcmp(input,"proj4"))
        return 1;
    if (!strcmp(input,"wkt"))
        return 1;
    if (!strcmp(input,"xml"))
        return 1;
        
    return 0;
}

int validate_output(char* input)
{
    if (!strcmp(input,"esri"))
        return 1;
    if (!strcmp(input,"proj4"))
        return 1;
    if (!strcmp(input,"wkt"))
        return 1;
    if (!strcmp(input,"xml"))
        return 1;
        
    return 0;
}
static PyObject *
_translate(PyObject *self, PyObject *args)
{
    PyObject *output=NULL;
    char* in_ref_type;
    char* out_ref_type;
    char* ref_defn;

    OGRSpatialReferenceH ref=NULL;

    if (!PyArg_ParseTuple(args, "sss:_translate", &ref_defn, &in_ref_type, &out_ref_type))
        goto fail;

    if (!validate_input(in_ref_type))
    {
        PyErr_SetString(OSRError, "input SRS type not recognized ... use 'esri','proj4','wkt','xml', or 'epsg'");
        goto fail;
    }
    if (!validate_output(out_ref_type))
    {
        PyErr_SetString(OSRError, "output SRS type not recognized... use 'esri','proj4','wkt', or 'xml'");    
        goto fail;
    }

    if (!strcmp(in_ref_type,"epsg"))
    {
        ref = from_epsg(ref_defn);   
    }    

    if (!strcmp(in_ref_type,"esri"))
    {
        ref = from_esri(ref_defn);   
    }

    if (!strcmp(in_ref_type,"proj4"))
    {
        ref = from_proj4(ref_defn);   
    }

    if (!strcmp(in_ref_type,"xml"))
    {
        ref = from_xml(ref_defn);   
    } 

    if (!strcmp(in_ref_type,"wkt"))
    {
        ref = from_wkt(ref_defn);   
    } 
        
    if (ref) 
    {
        if (!strcmp(out_ref_type,"wkt"))
        {
            output = to_wkt(ref);
        }
        if (!strcmp(out_ref_type,"xml"))
        {
            output = to_xml(ref);
        }
        if (!strcmp(out_ref_type,"esri"))
        {
            output = to_esri(ref);
        }
        if (!strcmp(out_ref_type,"proj4"))
        {
            output = to_proj4(ref);
        }
    }
    if (output)
    {
        return output;
    }
    else
    {
        return NULL;
    }

    fail:
    return NULL;
}


/*
===============================================================================
Initialization
===============================================================================
*/
static PyMethodDef OSRMethods[] = {

    {"_translate", (PyCFunction)_translate, METH_VARARGS,
        "Translates an SRS definition in one format to another"},
    {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC
init_osrtranslate(void)
{
    PyObject* module;
    module = Py_InitModule3("_osrtranslate", OSRMethods,
                       "OSR reference system translation module");
    if (module == NULL)
      return;

    OSRError = PyErr_NewException("_osrtranslate.OSRError", NULL, NULL);
    Py_INCREF(OSRError);
    PyModule_AddObject(module, "OSRError", OSRError);
    
    /* Suppress GDAL from emitting errors to stdout by default */
    CPLPushErrorHandler(CPLQuietErrorHandler);
}

