# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import os
import sys
import unittest

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
    pkg_resources.require('PCL-GDAL')
except:
    pass
    
from cartography.data.disk import DiskFeatureStore, DiskFeatureSource
from cartography.data.disk import DiskRasterStore
from cartography.geometry import Polygon, Point, BoundingBox

from cartography.data.interfaces import *

from _testing import *

# ============================================================================
# Test Store, Source, and Feature Type
# ============================================================================

TYPENAMES = ['world_borders', 'line', 'polygon', 'point']
TYPENAME0 = 'world_borders'
#SCHEMA0 = { 'AREA': 'float', 'CAT': 'float',
#            'CNTRY_NAME': 'string', 'POP_CNTRY': 'float',
#            'the_geom': 'POLYGON', 'FIPS_CNTRY': 'string' }
PROP0_NAME = 'CAT'
INFO0 = { 'AREA': 'float', 'CAT': 'float', 'CNTRY_NAME': 'string', 
          'POP_CNTRY': 'float', 'the_geom': 'geometry',
          'FIPS_CNTRY': 'string'}


class NewStoreTest(unittest.TestCase):
    def test_typenames(self):
        store = DiskFeatureStore(DISKSTORE)
        self.assert_(IFeatureStore.providedBy(store))
        self.assert_(IDataStore.providedBy(store))
        typenames = store.typenames()
        self.assert_(typenames == TYPENAMES, typenames)

class StoreTest(unittest.TestCase):
    def setUp(self):
        self.store = DiskFeatureStore(DISKSTORE)
    # Helper
    def assertBoundsAlmostEqual(self, b1, b2, arg):
        try:
            for c1, c2 in zip(b1, b2):
                self.assertAlmostEqual(c1, c2, 6)
        except:
            raise AssertionError, arg
    def test_featuretype(self):
        ftype = self.store.featuretype(TYPENAME0)
        self.assert_(ftype.typename() == TYPENAME0)
        self.assertEqual(ftype.schema_info(), INFO0)
        s = ftype.schema()
        self.assertEqual(s['the_geom'].homotype, 'Polygon')
    def test_source(self):
        source = self.store.source(TYPENAME0)
        assert isinstance(source, DiskFeatureSource)
    def test_source_query(self):
        p = "FIPS_CNTRY = 'FOO'"
        source = self.store.source(TYPENAME0, properties=p)
        assert isinstance(source, DiskFeatureSource)
    def test_bounds(self):
        bounds = self.store.bounds(TYPENAME0)
        self.assert_(isinstance(bounds, BoundingBox))
        bounds_tuple = bounds.totuple()
        self.assert_(len(bounds_tuple) == 4, bounds_tuple)
        self.assertBoundsAlmostEqual(bounds_tuple,
                (-180.0, -90.0, 180.0, 83.623596191406193), bounds_tuple)
    def test_bounds_fast(self):
        bounds = self.store.bounds(TYPENAME0, fastest=True)
        self.assert_(isinstance(bounds, BoundingBox))
        bounds_tuple = bounds.totuple()
        self.assert_(len(bounds_tuple) == 4, bounds_tuple)
        self.assertBoundsAlmostEqual(bounds_tuple,
                (-180.0, -90.0, 180.0, 83.623596191406193), bounds_tuple)
    def test_count(self):
        count = self.store.count(TYPENAME0)
        self.assert_(count == 3784, count)
    def test_info(self):
        info = self.store.info(TYPENAME0)
        self.assertEqual(info['typename'], TYPENAME0)
        self.assertEqual(info['schema'], INFO0)
   
class SourceTest(unittest.TestCase):
    def setUp(self):
        self.store = DiskFeatureStore(DISKSTORE)
        self.source = DiskFeatureSource(self.store, TYPENAME0)
    def test_featuretype(self):
        ftype = self.source.featuretype
        self.assert_(ftype.typename() == TYPENAME0)

class AccessTest(unittest.TestCase):
    def setUp(self):
        self.store = DiskFeatureStore(DISKSTORE)
        self.source = self.store.source(TYPENAME0)
    def test_unfiltered(self):
        source = self.store.source(TYPENAME0)
        ftype = source.featuretype
        fiter = source()
        feature = fiter.next()
        self.assert_(isinstance(feature, ftype))
        self.assert_(feature.id == '0', feature.id)
        self.assert_(feature.properties['CNTRY_NAME'] == 'Aruba')
    # back-end filtering
    def test_false(self):
        # expect 0 features when filter is globally False
        p = "FIPS_CNTRY = 'FOO'"
        source = self.store.source(TYPENAME0, properties=p)
        fiter = source()
        self.assertRaises(StopIteration, fiter.next)
    def test_true(self):
        p = "FIPS_CNTRY = 'UK'"
        source = self.store.source(TYPENAME0, properties=p)
        fiter = source()
        feature = fiter.next()
        self.assert_(feature.properties['FIPS_CNTRY'] == 'UK')
        self.assert_(isinstance(feature.properties.the_geom, Polygon), feature.properties.the_geom)
        # Check the total number of results 
        i = 1
        for feature in fiter:
            i += 1
        self.assert_(i == 48, i)
    def test_bbox(self):
        b = BoundingBox(-10, 45, 10, 60)
        source = self.store.source(TYPENAME0, bbox=b)
        fiter = source()
        feature = fiter.next()
        self.assert_(feature.properties['FIPS_CNTRY'] == 'BE')
        self.assert_(isinstance(feature.properties.the_geom, Polygon), feature.properties.the_geom)
        i = 1
        for feature in fiter:
            i += 1
        self.assert_(i == 107, i)
    def test_front_atts(self):
        b = BoundingBox(-10, 45, 10, 60)
        source = self.store.source(TYPENAME0, bbox=b)
        def func(s):
            return s.lower()
        ff = "func(f.FIPS_CNTRY) == 'uk'"
        fiter = source(filter=ff, func=func)
        i = 0
        for feature in fiter:
            i += 1
        self.assert_(i == 46, i)
    def test_front_geom(self):
        b = BoundingBox(-10, 45, 10, 60)
        source = self.store.source(TYPENAME0, bbox=b)
        def func(s):
            return s.lower()
        p = Point(0.0, 51.2)
        ff = "g.contains(p)"
        fiter = source(filter=ff, func=func, p=p)
        i = 0
        for feature in fiter:
            i += 1
        self.assert_(i == 1, i)

# ============================================================================
# Rasters
# ============================================================================

class RasterTest(unittest.TestCase):
    def test_raster(self):
        store = DiskRasterStore(RASTERSTORE)
        self.assert_(IDataStore.providedBy(store))
        self.assertEqual(store.typenames(), ['raster.vrt'])
        self.assertEqual(store.info(),
            {'srs': '+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs',
             'bands': [['Band 1', 'Byte']],
             'bounds': (-8.0333333000000007, 49.233343899999994, 2.8333224999999995, 59.8333333), 
             'typename': 'raster.vrt', 
             'resolution': (0.066666600000000006, 0.066666600000000034), 
             'size': (163, 159)})
        self.assertEqual(store.bounds().totuple(),
            (-8.0333333000000007, 49.233343899999994, 
             2.8333224999999995, 59.8333333))

# ============================================================================
if __name__ == '__main__':
    unittest.main()

