/*
# $Id: mapserverbridge.c,v 1.1 2005/01/27 04:53:51 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================
*/

#include "mapserverbridge.h"

/*===========================================================================*/
/* Utility functions from MapServer are included in this file because it is  */
/* not feasible to link against any subset of libmap -- it's all or nothing  */
/* because MapServer is entangled.                                           */
/*===========================================================================*/

void msInitShape(shapeObj *shape)
{
  // spatial component
  shape->line = NULL;
  shape->numlines = 0;
  shape->type = MS_SHAPE_NULL;
  shape->bounds.minx = shape->bounds.miny = -1;
  shape->bounds.maxx = shape->bounds.maxy = -1;
  
  // attribute component
  shape->values = NULL;
  shape->numvalues = 0;

  // annotation component
  shape->text = NULL;

  // bookkeeping component
  shape->classindex = 0; // default class
  shape->tileindex = shape->index = -1;
}

void msFreeShape(shapeObj *shape)
{
  int c;

  for (c= 0; c < shape->numlines; c++)
    free(shape->line[c].point);
  free(shape->line);

  if(shape->values) msFreeCharArray(shape->values, shape->numvalues);
  if(shape->text) free(shape->text);
  
  msInitShape(shape); // now reset
}

void msComputeBounds(shapeObj *shape)
{
  int i, j;

  if(shape->numlines <= 0) return;
  if(shape->line[0].numpoints <= 0) return;

  shape->bounds.minx = shape->bounds.maxx = shape->line[0].point[0].x;
  shape->bounds.miny = shape->bounds.maxy = shape->line[0].point[0].y;
    
  for( i=0; i<shape->numlines; i++ ) {
    for( j=0; j<shape->line[i].numpoints; j++ ) {
      shape->bounds.minx = MS_MIN(shape->bounds.minx, shape->line[i].point[j].x);
      shape->bounds.maxx = MS_MAX(shape->bounds.maxx, shape->line[i].point[j].x);
      shape->bounds.miny = MS_MIN(shape->bounds.miny, shape->line[i].point[j].y);
      shape->bounds.maxy = MS_MAX(shape->bounds.maxy, shape->line[i].point[j].y);
    }
  }
}

void msFree(void *p) {
  if(p) free(p);
}

/*
** Free memory allocated for a character array
*/
void msFreeCharArray(char **array, int num_items)
{
  int i;

  if((num_items < 0) || !array) return;

  for(i=0;i<num_items;i++)
    msFree(array[i]);
  msFree(array);

  return;
}

int msAddLine(shapeObj *p, lineObj *new_line)
{
    lineObj lineCopy;

    lineCopy.numpoints = new_line->numpoints;
    lineCopy.point = (pointObj *) malloc(new_line->numpoints*sizeof(pointObj));
    if (lineCopy.point == NULL )
    {
        PyErr_SetString(PyExc_Exception, "Failed to allocate line copy");
        return MS_FAILURE;
    }
    
    memcpy(lineCopy.point, new_line->point, 
           sizeof(pointObj) * new_line->numpoints );

    return msAddLineDirectly(p, &lineCopy);
}

/*
** Same as msAddLine(), except that this version "seizes" the points
** array from the passed in line and uses it instead of copying it.  
*/
int msAddLineDirectly(shapeObj *p, lineObj *new_line)
{
  int c;

  if( p->numlines == 0 )
      p->line = (lineObj *) malloc(sizeof(lineObj));
  else
      p->line = (lineObj *) realloc(p->line, (p->numlines+1)*sizeof(lineObj));
  
  /* Copy the new line onto the end of the extended line array */
  c= p->numlines;
  p->line[c].numpoints = new_line->numpoints;  
  p->line[c].point = new_line->point;

  /* strip points reference off the passed in lineObj */
  new_line->point = NULL;
  new_line->numpoints = 0;

  /* Update the polygon information */
  p->numlines++;

  return MS_SUCCESS;
}

/* ==========================================================================
 Geometry Conversion
=========================================================================== */

int
mbShapeFromGEOSGeom_Point(shapeObj *shape, GEOSGeom g)
{
    GEOSCoordSeq seq;
    int num_points;
    int i;
    double x, y;
    lineObj line;
   
    seq = GEOSGeom_getCoordSeq(g);
    GEOSCoordSeq_getSize(seq, &num_points);
    line.point = (pointObj *) malloc(sizeof(pointObj)*num_points);
    if (line.point == NULL)
    {
	    PyErr_SetString(PyExc_Exception, "Unable to allocate lineObj");
	        return MS_FAILURE;
    }
    line.numpoints = num_points;

    /* Iterate over points, setting the shape vertices */
    for (i=0; i<num_points; i++)
    {
        GEOSCoordSeq_getX(seq, i, &x);
        GEOSCoordSeq_getY(seq, i, &y);
        line.point[i].x = x;
        line.point[i].y = y;
    }
    msAddLine(shape, &line);
    free(line.point);

    /* Assign shape type */
    shape->type = MS_SHAPE_POINT;

    /* Compute shape bounds */
    msComputeBounds(shape);
    
    return MS_SUCCESS;
}

int
mbShapeFromGEOSGeom_LineString(shapeObj *shape, GEOSGeom g)
{
    GEOSCoordSeq seq;
    int num_points;
    int i;
    double x, y;
    lineObj line;
   
    seq = GEOSGeom_getCoordSeq(g);
    GEOSCoordSeq_getSize(seq, &num_points);
    line.point = (pointObj *) malloc(sizeof(pointObj)*num_points);
    if (line.point == NULL)
    {
	    PyErr_SetString(PyExc_Exception, "Unable to allocate lineObj");
	        return MS_FAILURE;
    }
    line.numpoints = num_points;

    /* Iterate over points, setting the shape vertices */
    for (i=0; i<num_points; i++)
    {
        GEOSCoordSeq_getX(seq, i, &x);
        GEOSCoordSeq_getY(seq, i, &y);
        line.point[i].x = x;
        line.point[i].y = y;
    }
    msAddLine(shape, &line);
    free(line.point);

    /* Assign shape type */
    shape->type = MS_SHAPE_LINE;

    /* Compute shape bounds */
    msComputeBounds(shape);
    
    return MS_SUCCESS;
}

int
mbShapeFromGEOSGeom_Polygon(shapeObj *shape, GEOSGeom g)
{
    GEOSGeom cur;
    GEOSCoordSeq seq;
    int num_points, num_rings;
    lineObj line;
    int i, j;
    double x, y;
   
    /* Exterior ring */
    cur = GEOSGetExteriorRing(g);
    seq = GEOSGeom_getCoordSeq(cur);
    GEOSCoordSeq_getSize(seq, &num_points);
    line.point = (pointObj *) malloc(sizeof(pointObj)*num_points);
    if (line.point == NULL)
    {
	    PyErr_SetString(PyExc_Exception, "Unable to allocate lineObj");
        return MS_FAILURE;
    }
    line.numpoints = num_points;

    /* Iterate over points, setting the shape vertices */
    for (i=0; i<num_points; i++)
    {
        GEOSCoordSeq_getX(seq, i, &x);
        GEOSCoordSeq_getY(seq, i, &y);
        line.point[i].x = x;
        line.point[i].y = y;
    }
    msAddLine(shape, &line);
    free(line.point);
    
    /* Iterate over interior rings, allocating points */
    num_rings = GEOSGetNumInteriorRings(g);
    for (j=0; j<num_rings; j++)
    {
        cur = GEOSGetInteriorRingN(g, j);
        seq = GEOSGeom_getCoordSeq(cur);
        GEOSCoordSeq_getSize(seq, &num_points);
        line.point = (pointObj *) malloc(sizeof(pointObj)*num_points);
        if (line.point == NULL)
        {
	        PyErr_SetString(PyExc_Exception, "Unable to allocate lineObj");
            return MS_FAILURE;
        }
        line.numpoints = num_points;

        /* Iterate over points, setting the shape vertices */
        for (i=0; i<num_points; i++)
        {
            GEOSCoordSeq_getX(seq, i, &x);
            GEOSCoordSeq_getY(seq, i, &y);
            line.point[i].x = x;
            line.point[i].y = y;
        }
        msAddLine(shape, &line);
        free(line.point);
    }

    /* Assign shape type */
    shape->type = MS_SHAPE_POLYGON;

    /* Compute shape bounds */
    msComputeBounds(shape);
    
    return MS_SUCCESS;
}

int
mbShapeFromGEOSGeom_MultiPoint(shapeObj *shape, GEOSGeom g)
{
    GEOSGeom cur;
    GEOSCoordSeq seq;
    int num_points, num_geoms;
    int i, k;
    double x, y;
    lineObj line;
   
    num_geoms = GEOSGetNumGeometries(g);
    for (k=0; k<num_geoms; k++)
    {
        cur = GEOSGetGeometryN(g, k);
        seq = GEOSGeom_getCoordSeq(cur);
        GEOSCoordSeq_getSize(seq, &num_points);
        line.point = (pointObj *) malloc(sizeof(pointObj)*num_points);
        if (line.point == NULL)
        {
	        PyErr_SetString(PyExc_Exception, "Unable to allocate lineObj");
	        return MS_FAILURE;
        }
        line.numpoints = num_points;

        /* Iterate over points, setting the shape vertices */
        for (i=0; i<num_points; i++)
        {
            GEOSCoordSeq_getX(seq, i, &x);
            GEOSCoordSeq_getY(seq, i, &y);
            line.point[i].x = x;
            line.point[i].y = y;
        }
        msAddLine(shape, &line);
        free(line.point);
    }

    /* Assign shape type */
    shape->type = MS_SHAPE_POINT;

    /* Compute shape bounds */
    msComputeBounds(shape);
    
    return MS_SUCCESS;
}

int
mbShapeFromGEOSGeom_MultiLineString(shapeObj *shape, GEOSGeom g)
{
    GEOSGeom cur;
    GEOSCoordSeq seq;
    int num_points, num_geoms;
    int i, k;
    double x, y;
    lineObj line;
   
    num_geoms = GEOSGetNumGeometries(g);
    for (k=0; k<num_geoms; k++)
    {
        cur = GEOSGetGeometryN(g, k);
        seq = GEOSGeom_getCoordSeq(cur);
        GEOSCoordSeq_getSize(seq, &num_points);
        line.point = (pointObj *) malloc(sizeof(pointObj)*num_points);
        if (line.point == NULL)
        {
	        PyErr_SetString(PyExc_Exception, "Unable to allocate lineObj");
	        return MS_FAILURE;
        }
        line.numpoints = num_points;

        /* Iterate over points, setting the shape vertices */
        for (i=0; i<num_points; i++)
        {
            GEOSCoordSeq_getX(seq, i, &x);
            GEOSCoordSeq_getY(seq, i, &y);
            line.point[i].x = x;
            line.point[i].y = y;
        }
        msAddLine(shape, &line);
        free(line.point);
    }

    /* Assign shape type */
    shape->type = MS_SHAPE_LINE;

    /* Compute shape bounds */
    msComputeBounds(shape);
    
    return MS_SUCCESS;
}


int
mbShapeFromGEOSGeom_MultiPolygon(shapeObj *shape, GEOSGeom g)
{
    GEOSGeom p;
    GEOSGeom cur;
    GEOSCoordSeq seq;
    int num_points, num_rings, num_polygons;
    int i, j, k;
    double x, y;
    lineObj line;
   
    num_polygons = GEOSGetNumGeometries(g);
    for (k=0; k<num_polygons; k++)
    {
        p = GEOSGetGeometryN(g, k);
            
        /* Exterior ring */
        cur = GEOSGetExteriorRing(p);
        seq = GEOSGeom_getCoordSeq(cur);
        GEOSCoordSeq_getSize(seq, &num_points);
        line.point = (pointObj *) malloc(sizeof(pointObj)*num_points);
        if (line.point == NULL)
        {
	        PyErr_SetString(PyExc_Exception, "Unable to allocate lineObj");
	        return MS_FAILURE;
        }
        line.numpoints = num_points;

        /* Iterate over points, setting the shape vertices */
        for (i=0; i<num_points; i++)
        {
            GEOSCoordSeq_getX(seq, i, &x);
            GEOSCoordSeq_getY(seq, i, &y);
            line.point[i].x = x;
            line.point[i].y = y;
        }
        msAddLine(shape, &line);
        free(line.point);
    
        /* Iterate over interior rings, allocating points */
        num_rings = GEOSGetNumInteriorRings(p);
        for (j=0; j<num_rings; j++)
        {
            cur = GEOSGetInteriorRingN(p, j);
            seq = GEOSGeom_getCoordSeq(cur);
            GEOSCoordSeq_getSize(seq, &num_points);
            line.point = (pointObj *) malloc(sizeof(pointObj)*num_points);
            if (line.point == NULL)
            {
	            PyErr_SetString(PyExc_Exception, "Unable to allocate lineObj");
                return MS_FAILURE;
            }
            line.numpoints = num_points;

            /* Iterate over points, setting the shape vertices */
            for (i=0; i<num_points; i++)
            {
                GEOSCoordSeq_getX(seq, i, &x);
                GEOSCoordSeq_getY(seq, i, &y);
                line.point[i].x = x;
                line.point[i].y = y;
            }
            msAddLine(shape, &line);
            free(line.point);
        }
    }

    /* Assign shape type */
    shape->type = MS_SHAPE_POLYGON;

    /* Compute shape bounds */
    msComputeBounds(shape);
    
    return MS_SUCCESS;
}

int 
mbShapeFromGEOSGeom(shapeObj *s, GEOSGeom g)
{
    int result, t1, t2;
    GEOSGeom x;
    
    if (!g)
    {
        PyErr_SetString(PyExc_ValueError, "NULL GEOS Geometry");
        return MS_FAILURE;
    }

    t1 = GEOSGeomTypeId(g);
    switch(t1)
    {
        case GEOS_POINT:
            result = mbShapeFromGEOSGeom_Point(s, g);
            break;
        case GEOS_LINESTRING:
            result = mbShapeFromGEOSGeom_LineString(s, g);
            break;
        case GEOS_POLYGON:
            result = mbShapeFromGEOSGeom_Polygon(s, g);
            break;
        case GEOS_MULTIPOINT:
            result = mbShapeFromGEOSGeom_MultiPoint(s, g);
            break;
        case GEOS_MULTILINESTRING:
            result = mbShapeFromGEOSGeom_MultiLineString(s, g);
            break;
        case GEOS_MULTIPOLYGON:
            result = mbShapeFromGEOSGeom_MultiPolygon(s, g);
            break;
        case GEOS_GEOMETRYCOLLECTION:
            x = GEOSGetGeometryN(g, 0);
            t2 = GEOSGeomTypeId(x);
            switch(t2)
            {
                case GEOS_POINT:
                    result = mbShapeFromGEOSGeom_MultiPoint(s, g);
                    break;
                case GEOS_POLYGON:
                    result = mbShapeFromGEOSGeom_MultiPolygon(s, g);
                    break;
                case GEOS_LINESTRING:
                    result = mbShapeFromGEOSGeom_MultiLineString(s, g);
                    break;
                default:
                    PyErr_SetString(PyExc_Exception, "Unknown geometry type");
                    result = MS_FAILURE;
                    break;
            }
            break;
        default:
            PyErr_SetString(PyExc_Exception, "Unknown geometry type");
            result = MS_FAILURE;
            break;
    }

    return result;
}
