%module msbridge

%{
#include "map.h"
#include "mapserverbridge.h"
%}

%include typemaps.i
%include constraints.i

/*
===========================================================================
 Supporting 'None' as an argument to attribute accessor functions

 Typemaps to support NULL in attribute accessor functions
 provided to Steve Lime by David Beazley and tested for Python
 only by Sean Gillies.
  
 With the use of these typemaps we can execute statements like

   layer.group = None

=========================================================================
*/

#ifdef __cplusplus
%typemap(memberin) char * {
  if ($1) delete [] $1;
  if ($input) {
     $1 = ($1_type) (new char[strlen($input)+1]);
     strcpy((char *) $1,$input);
  } else {
     $1 = 0;
  }
}
%typemap(memberin,warning="451:Setting const char * member may leak
memory.") const char * {
  if ($input) {
     $1 = ($1_type) (new char[strlen($input)+1]);
     strcpy((char *) $1,$input);
  } else {
     $1 = 0;
  }
}
%typemap(globalin) char * {
  if ($1) delete [] $1;
  if ($input) {
     $1 = ($1_type) (new char[strlen($input)+1]);
     strcpy((char *) $1,$input);
  } else {
     $1 = 0;
  }
}
%typemap(globalin,warning="451:Setting const char * variable may leak
memory.") const char * {
  if ($input) {
     $1 = ($1_type) (new char[strlen($input)+1]);
     strcpy((char *) $1,$input);
  } else {
     $1 = 0;
  }
}
#else
%typemap(memberin) char * {
  if ($1) free((char*)$1);
  if ($input) {
     $1 = ($1_type) malloc(strlen($input)+1);
     strcpy((char*)$1,$input);
  } else {
     $1 = 0;
  }
}
%typemap(memberin,warning="451:Setting const char * member may leak
memory.") const char * {
  if ($input) {
     $1 = ($1_type) malloc(strlen($input)+1);
     strcpy((char*)$1,$input);
  } else {
     $1 = 0;
  }
}
%typemap(globalin) char * {
  if ($1) free((char*)$1);
  if ($input) {
     $1 = ($1_type) malloc(strlen($input)+1);
     strcpy((char*)$1,$input);
  } else {
     $1 = 0;
  }
}
%typemap(globalin,warning="451:Setting const char * variable may leak
memory.") const char * {
  if ($input) {
     $1 = ($1_type) malloc(strlen($input)+1);
     strcpy((char*)$1,$input);
  } else {
     $1 = 0;
  }
}
#endif // __cplusplus

// Translates Python None to C NULL for strings
%typemap(in,parse="z") char * "";

/* Structures from MapServer's mapprimitive.h */

typedef struct {
    double minx, miny, maxx, maxy;
} rectObj;

typedef struct {
    double x;
    double y;
} pointObj;

typedef struct {
    int numpoints;
    pointObj *point;
} lineObj;

typedef struct {
    int numlines;
    int numvalues;
    lineObj *line;
    char **values;
    rectObj bounds;
    int type;
    long index;
    int tileindex;
    int classindex;
    char *text;
#ifdef USE_GEOS
    void *geometry;
#endif
} shapeObj;


%extend shapeObj 
{
  
    shapeObj(int type, PyObject *geom)
    {
        int i;
        shapeObj *shape;
        Geometry *g;
        
        shape = (shapeObj *) malloc(sizeof(shapeObj));
        if (!shape)
        {
    	    PyErr_SetString(PyExc_Exception,
                            "Unable to allocate memory for shapeObj");
            return NULL;
        }

        msInitShape(shape);

        /* Allocate memory for 4 values */
        if ((shape->values = (char **) malloc(sizeof(char *)*4)) == NULL)
        {
    	    PyErr_SetString(PyExc_Exception,
                            "Failed to allocate memory for values");
            return NULL;
        }
        else
        {
            for (i=0; i<4; i++) shape->values[i] = strdup("");
        }
        shape->numvalues = 4;

        /* ---------------------------------------------------------------- */
        /* Set the shape from the input parameters                          */
        /* ---------------------------------------------------------------- */

        g = (Geometry *) geom;
        if (mbShapeFromGEOSGeom(shape, g->geom) != MS_SUCCESS)
        {
            PyErr_SetString(PyExc_Exception, "Failed to convert shape");
            return NULL;
        }
        
        if (!shape)
        {
    	    PyErr_SetString(PyExc_Exception,
                            "NULL shape!");
            return NULL;
        }
        
        return shape;
    }

    ~shapeObj() 
    {
        msFreeShape(self);
        free(self);		
    }

    char *getValue(int i) 
    { // returns an EXISTING value
        if (i >= 0 && i < self->numvalues && self->values)
            return (self->values[i]);
        else
            return NULL;
    }
   
    int setValue(int i, char *value)
    {
        if (!self->values || !value)
        {
    	    PyErr_SetString(PyExc_Exception, "Can't set value");
            return MS_FAILURE;
        }
        if (i >= 0 && i < 4)
        {
            msFree(self->values[i]);
            self->values[i] = strdup(value);
            if (!self->values[i])
            {
                return MS_FAILURE;
            }
            else 
                return MS_SUCCESS;
        }
        else
        {
    	    PyErr_SetString(PyExc_Exception,
                            "Invalid index, only 4 possible values");
            return MS_FAILURE;
        }
    }

}

