# $Id: rendering.py,v 1.7 2005/05/04 18:05:12 dokai Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import cStringIO
import re
import sys
from traceback import print_exc
import os
import glob

# Require PCL-Referencing
try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
    pkg_resources.require('PCL-MapServer')
except:
    pass

from cartography.engine.mapserver import msbridge
from cartography.geometry import BoundingBox, Point
from cartography.styles import PointSymbolizer, LineSymbolizer, \
    PolygonSymbolizer, TextSymbolizer
from cartography.context import Image

from cartography.geometry.interfaces import IPoint
from cartography.data.interfaces import IFeatureStore, IRasterStore
from cartography.styles.interfaces import IStyle
from cartography.styles.interfaces import ISLDPointSymbolizer, \
    ISLDLineSymbolizer, ISLDPolygonSymbolizer, ISLDTextSymbolizer, \
    ISLDRasterSymbolizer

import mapscript
from mapscript import MapServerError
import _mapscript
import ogr

# try to find elementtree or lxml
try:
    from cElementTree import Element, SubElement, tostring, fromstring
except:
    try:
        from elementtree.ElementTree import Element, SubElement
        from elementtree.ElementTree import tostring, fromstring
    except:
        from lxml.etree import Element, SubElement, tostring, fromstring

def isfrombackend(datastore):
    """Return True if the PCL datastore is from a disk, RDBMS, or WFS
    backend, otherwise return False"""
    if hasattr(datastore, 'path') \
    or hasattr(datastore, 'dsn') \
    or hasattr(datastore, 'url'):
        return True
    else:
        return False

class ImageRenderer:

    """MapServer mapping engine's map image renderer.
    """

    def __init__(self, **kw):
        """Initialize a MapServer ImageRenderer.

        Arguments
        ---------
        kw : mapping
            Keyword arguments, we specifcally want 'fontset' and/or
            'mapfile'.
        """
        self.fontset = kw.get('fontset', '')
        self.incoming = kw.get('incoming', '')
        self.mapfile = kw.get('mapfile', '')

    def render(self, map, view, format, size, bgcolor='#ffffff', 
        transparent=False, debug=False, scalebar=False):
        """Make a map and return an image.

        Parameters
        ----------
        map : MapContext
            A PCL map context
        view : View
            A mapping.View defining the spatial extents of map image
        format : string
            output image mimetype such as 'image/jpeg', 'image/tif',
            'image/gif', 'image/png'. The PNG and TIFF formats can also be
            extended with mode parameters like 'image/png; mode=NAME'. The
            available modes are:
                PNG:    'Palette', 'RGB', 'RGBA' (default is 'Palette')
                TIFF:   'Palette', 'RGB', 'RGBA', 'Int16', 'Float32'
        size : (int, int)
            (width, height) tuple of image dimensions (pixels)
        size : (int, int)
            (width, height) tuple of image dimensions (pixels)
        bgcolor : string
            map's background color as hex
        transparent : bool
            whether or not map's background should be transparent
        
        Returns
        -------
        An instance of mapping.Image.
        """

        # Catch exceptions, return an error image, but also re-raise the 
        # exception.
        try:
        
            # Initialize mapserver map
            mo = _init_map(self, size, view, format, bgcolor, transparent)
            if debug:
                mo.debug = mapscript.MS_ON
    
            # Prepare to render mapserver map 
            im = mo.prepareImage()
    
            # Initialization of layers
            # ===============================================================
            context = []
            
            for context_layer in map:
                # Munge PCL styled layer into a MapServer-like tree
                
                ltree = _munge_layers(context_layer.layer, context_layer.style)
                context.append(ltree)
                
                for ms_layering in ltree:
                    # Initialize a MapServer layerObj. Its type and data are
                    # set and classes/styles are initialized.
                    
                    lo = _initialize_layerObj(ms_layering, mo)
    
                    if context_layer.visible:
                        lo.status = mapscript.MS_ON
                    else:
                        lo.status = mapscript.MS_OFF
                    
                    li = mo.insertLayer(lo)
                    ms_layering[0].append(li)
                    
            # Rendering of layers
            # ===============================================================
            for ltree in context:
    
                for ms_layering in ltree:
                   
                    ms_ruling = ms_layering[1]
                    ldef = ms_layering[0]
                    layer, style, symbclass, geom, msidx = ldef
                    
                    # Get the MapServer layer from the mapObj 
                    lo = mo.getLayer(msidx)
                    
                    # Case 1: rasters
                    # -------------------------------------------------------
                    if ISLDRasterSymbolizer.implementedBy(symbclass):
                        if IFeatureStore.providedBy(layer.datastore):
                            # For the tile indexed case we create a temporary
                            # layer that points to the index layer
                            maxscaledenom = -1
                            minscaledenom = -1
                            for i in range(len(ms_ruling)):
                               r = ms_ruling[i][0]
                               if r.minscaledenominator != -1 :
                                  minscaledenom = r.minscaledenominator
                               if r.maxscaledenominator != -1 :
                                  maxscaledenom = r.maxscaledenominator
                            raster_lo = mapscript.layerObj(mo)
                            raster_lo.type = mapscript.MS_LAYER_RASTER
                            raster_lo.data = None
                            raster_lo.name = '_Temp_Raster'
                            raster_lo.status = mapscript.MS_DEFAULT
                            raster_lo.setProjection(lo.getProjection())
                            #raster_lo.debug = mapscript.MS_TRUE
                            raster_lo.tileindex = lo.name
                            raster_lo.minscale = minscaledenom
                            raster_lo.maxscale = maxscaledenom
                            raster_lo.transparency = lo.transparency
                            
                            # Switch the tileindex layer type to POLYGON
                            lo.type = mapscript.MS_LAYER_POLYGON
                            
                            # Draw raster
                            raster_lo.draw(mo, im)
                        else:
                            # The single raster file case
                            lo.draw(mo, im)
                        
                    # Cases 2*: internal classification of vectors
                    # -------------------------------------------------------
                    elif layer.classifier == 0:
                        classifications = {}
                        for i in range(len(ms_ruling)):
                            r = ms_ruling[i][0]
                            filter = r.filter
                            if not filter: filter = 'True'
                            classifications[str(r)] = \
                                [compile(filter, '<string>', 'eval'), i]
                        
                        _render_layer_extrinsic(im, mo, lo, layer, style,
                            ms_ruling, classifications, geom)
                
                    # Cases 2*: external classification of vectors
                    # -------------------------------------------------------
                    elif layer.classifier == 1:
                        for i in range(len(ms_ruling)):
                            r = ms_ruling[i][0]
                            filter = r.filter
                            if filter:
                                ms_expr = convertFilterToMSLogical(filter)
                                co = lo.getClass(i)
                                if len(ms_expr.strip()) > 2:
                                    #lo.classitem = "foo"
                                    co.setExpression(ms_expr)
                        lo.draw(mo, im)
    
            # End of layer rendering
            # ===============================================================
            # Complete the map

            mo.save('test.map')
            if scalebar:
                mo.scalebar.backgroundcolor.setHex('#FFFFFF')
                mo.scalebar.imagecolor.setHex(mo.imagecolor.toHex())
                mo.scalebar.color.setHex('#222222')
                mo.scalebar.outlinecolor.setHex('#000000')
                mo.scalebar.transparent = mapscript.MS_TRUE
                mo.scalebar.postlabelcache = mapscript.MS_TRUE
                mo.scalebar.height = 4
                mo.scalebar.width = 160
                mo.scalebar.intervals = 4
                mo.scalebar.label.color.setHex('#222222')
                mo.scalebar.label.size = mapscript.MS_SMALL
                mo.scalebar.position = mapscript.MS_LL
                mo.scalebar.status = mapscript.MS_EMBED
                if mo.scale < 40000:
                   mo.scalebar.units = mapscript.MS_METERS
                else:
                   mo.scalebar.units = mapscript.MS_KILOMETERS
                mo.embedScalebar(im)
            mo.drawLabelCache(im)
            return self._convert_image(im)
            
        except Exception, e:
            # print exception to stderr
            print_exc()

            # Make an error image
            mo = _init_map(self, size, view, format, '#FFCCCC',
                transparent=False)
            im = mo.prepareImage()
            lo = mapscript.layerObj(mo)
            lo.type = mapscript.MS_LAYER_ANNOTATION
            lo.transform = mapscript.MS_FALSE
            lo.status = mapscript.MS_ON
            co = mapscript.classObj(lo)
            co.label.type = mapscript.MS_BITMAP
            co.label.size = mapscript.MS_MEDIUM
            co.label.color.setHex('#FF0000')
            co.label.position = mapscript.MS_CC
            
            p = mapscript.pointObj(size[0]/2, size[1]/2)
            l = mapscript.lineObj()
            l.add(p)
            po = mapscript.shapeObj(mapscript.MS_SHAPE_POINT)
            po.add(l)
            po.text = "An error occurred. See log file."
            po.classindex = 0
            
            po.draw(mo, lo, im)
            mo.drawLabelCache(im)
            image = self._convert_image(im)
            
            # Stash the exception in the image and return
            image.error = e
            return image
        
    def _convert_image(self, im):
        image = Image()
        if hasattr(im, 'getBytes'):
            image.data.write(im.getBytes())
        else:
            im.write(image.data)
        image.size = [im.width, im.height]
        image.mimetype = im.format.mimetype
        return image
          
    def draw(self, symbolizer, format, size=[16, 16], bgcolor='#ffffff',
               transparent=False):
        """Make an icon-sized image.

        Arguments
        ---------
        symbolizer : mapping.Symbolizer
            symbolizer to be drawn.
            
        For other arguments, see the 'render' method.
        
        Returns
        -------
        An instance of mapping.Image.
        """
        image = Image()

        mo = mapscript.mapObj(self.mapfile)
        mo.setSize(size[0], size[1])
        mo.setExtent(0.0, 0.0, 1.0, 1.0)
        mo.imagecolor.setHex(bgcolor)
        if self.fontset:
            mo.setFontSet(self.fontset)
        
        _setOutputFormat(mo, format, transparent)
        
        _buildInMarks(mo)

        if transparent:
            mo.outputformat.transparent = mapscript.MS_ON

        im = mo.prepareImage()
      
        # Layer to hold one shape
        li = mo.insertLayer(mapscript.layerObj())
        lo = mo.getLayer(li)
        lo.name = 'draw'
        lo.status = mapscript.MS_ON
        lo.connectiontype = mapscript.MS_INLINE

        # Class and style
        ci = lo.insertClass(mapscript.classObj())
        co = lo.getClass(ci)
        co.status = mapscript.MS_ON

        try:
            symbolizers = [s for s in symbolizer]
        except:
            symbolizers = [symbolizer]
            
        for symb in symbolizers:
            if ISLDTextSymbolizer.providedBy(symb):
                _labelize(co, lo, symb)
            else:
                si = co.insertStyle(mapscript.styleObj())
                so = co.getStyle(si)
                _stylize(so, co, lo, mo, symb)

            # Form our shape
            if ISLDTextSymbolizer.providedBy(symb):
                shape = mapscript.shapeObj(mapscript.MS_SHAPE_POINT)
                line = mapscript.lineObj()
                point = mapscript.pointObj(0.5, 0.5)
                line.add(point)
                shape.add(line)
                shape.setBounds()
                shape.classindex = 0
                shape.text = 'Text'
                lo.type = mapscript.MS_LAYER_ANNOTATION
                shape.draw(mo, lo, im)
            elif ISLDPointSymbolizer.providedBy(symb):
                shape = mapscript.shapeObj(mapscript.MS_SHAPE_POINT)
                line = mapscript.lineObj()
                point = mapscript.pointObj(0.5, 0.5)
                line.add(point)
                shape.add(line)
                shape.setBounds()
                shape.classindex = 0
                lo.type = mapscript.MS_LAYER_POINT
                lo.addFeature(shape)
                lo.draw(mo, im)
            elif ISLDLineSymbolizer.providedBy(symb):
                shape = mapscript.shapeObj(mapscript.MS_SHAPE_LINE)
                line = mapscript.lineObj()
                line.add(mapscript.pointObj(0.2, 0.2))
                line.add(mapscript.pointObj(0.4, 0.8))
                line.add(mapscript.pointObj(0.6, 0.2))
                line.add(mapscript.pointObj(0.8, 0.8))
                assert line.numpoints == 4
                shape.add(line)
                assert shape.numlines == 1
                shape.setBounds()
                shape.classindex = 0
                lo.type = mapscript.MS_LAYER_LINE
                lo.addFeature(shape)
                lo.draw(mo, im)
                lo.type = mapscript.MS_LAYER_LINE
            elif ISLDPolygonSymbolizer.providedBy(symb):
                shape = mapscript.shapeObj(mapscript.MS_SHAPE_POLYGON)
                line = mapscript.lineObj()
                point = mapscript.pointObj(0.2, 0.2)
                line.add(point)
                point = mapscript.pointObj(0.2, 0.8)
                line.add(point)
                point = mapscript.pointObj(0.8, 0.8)
                line.add(point)
                point = mapscript.pointObj(0.8, 0.2)
                line.add(point)
                point = mapscript.pointObj(0.2, 0.2)
                line.add(point)
                shape.add(line)
                shape.setBounds()
                shape.classindex = 0
                lo.type = mapscript.MS_LAYER_POLYGON
                shape.draw(mo, lo, im)
        
        mo.drawLabelCache(im)

        # for the upcoming mapserver function
        if hasattr(im, 'getBytes'):
            image.data.write(im.getBytes())
        else:
            im.write(image.data)

        image.size = [im.width, im.height]
        image.mimetype = im.format.mimetype

        del im
        del mo
        
        return image
        

# ============================================================================
# MapServer/mapscript specific utility functions
# ============================================================================

_imagemode_mapping = {'Palette':  mapscript.MS_IMAGEMODE_PC256,
                      'RGB':      mapscript.MS_IMAGEMODE_RGB,
                      'RGBA':     mapscript.MS_IMAGEMODE_RGBA,
                      'Int16':    mapscript.MS_IMAGEMODE_INT16,
                      'Float32':  mapscript.MS_IMAGEMODE_FLOAT32 }
    
_format_driver_mapping = {'image/gif':  'GD/GIF',
                          'image/jpeg': 'GD/JPEG',
                          'image/png':  'GD/PNG',
                          'image/tif':  'GDAL/GTiff' }

_format_modes_mapping = {'image/gif':  ['Palette'],
                         'image/jpeg': ['RGB'],
                         'image/png':  ['Palette', 'RGB', 'RGBA'],
                         'image/tif':  ['Palette', 'RGB', 'RGBA',
                                        'Int16', 'Float32'] }

def _setOutputFormat(mo, format, transparent):
    """Take a PCL extended format string such as 

       image/png; mode=RGB

    and initialize a mapObj output format
    """

    format_fields = format.split(';')
    num_fields = len(format_fields)
   
    mimetype = format_fields[0]
    
    try:
        imagemode = _format_modes_mapping[mimetype][0]
    except KeyError:
        raise Exception, "no output driver available for unsupported format %s" % (format)

    options = {}
    
    if num_fields > 1:
        # iterate looking for a "mode=" parameter
        for i in range(1, num_fields):
            field = format_fields[i].strip()
            p, v = field.split('=')
            if p.lower() == 'mode':
                imagemode = v.strip()
                if imagemode not in _format_modes_mapping[mimetype]:
                    raise Exception, "Mode '%s' is invalid. Format '%s' supports only %s" % (imagemode, mimetype, str(_format_modes_mapping[mimetype]))
            if p.lower() == 'interlace':
                options[p.lower()]=v.strip()
                
    driver = _format_driver_mapping[mimetype]
    output_format = mapscript.outputFormatObj(driver, 'pcl_format')
    output_format.imagemode = _imagemode_mapping[imagemode]
    if transparent:
        output_format.transparent = mapscript.MS_ON
    
    for key in options.keys():
        output_format.setOption(key,options[key])
        
    
    mo.appendOutputFormat(output_format)
    mo.selectOutputFormat('pcl_format')

def _buildInMarks(mo):
    """Add well known marks to a mapObj"""
    circle = mapscript.symbolObj('circle')
    circle.type = mapscript.MS_SYMBOL_ELLIPSE
    circle.filled = mapscript.MS_TRUE
    line = mapscript.lineObj()
    line.add(mapscript.pointObj(1,1))
    circle.setPoints(line)
    circle.sizex = 1
    circle.sizey = 1
    mo.symbolset.appendSymbol(circle)

    square = mapscript.symbolObj('square')
    square.type = mapscript.MS_SYMBOL_VECTOR
    square.filled = mapscript.MS_TRUE
    line = mapscript.lineObj()
    line.add(mapscript.pointObj(0,0))
    line.add(mapscript.pointObj(0,2))
    line.add(mapscript.pointObj(2,2))
    line.add(mapscript.pointObj(2,0))
    line.add(mapscript.pointObj(0,0))
    square.setPoints(line)
    square.sizex = 2
    square.sizey = 2
    mo.symbolset.appendSymbol(square)

    triangle = mapscript.symbolObj('triangle')
    triangle.type = mapscript.MS_SYMBOL_VECTOR
    triangle.filled = mapscript.MS_TRUE
    line = mapscript.lineObj()
    line.add(mapscript.pointObj(0,4))
    line.add(mapscript.pointObj(2,0))
    line.add(mapscript.pointObj(4,4))
    line.add(mapscript.pointObj(0,4))
    triangle.setPoints(line)
    triangle.sizex = 4
    triangle.sizey = 4
    mo.symbolset.appendSymbol(triangle)

def _set_projection(mo, srs):
    """Set mapscript object's projection"""
    # EPSG code has priority
    if type(srs) == type('string'):
        mo.setProjection("init=%s" % (srs))
    else:
        mo.setProjection(srs.tostring())
        
def _labelize(classob, layerob, symbolizer, minscale=-1, maxscale=-1):
    """
    Applies text symbolization from 'symbolizer' to the specified mapscript
    classob, and layerob.
    """
    labelob = classob.label
    
    # labelitem
    if symbolizer.label:
        labelob.color.setHex(symbolizer.fill.color)
        layerob.labelitem = symbolizer.label

    # Font    
    if symbolizer.font.name:
        font = symbolizer.font
        labelob.type = mapscript.MS_TRUETYPE
        labelob.font = font.name
        labelob.size = int(font.size)
    else:
        labelob.type = mapscript.MS_BITMAP
        labelob.size = mapscript.MS_SMALL

    # halo/outline or billboard
    if symbolizer.halo.radius > 0:
        if symbolizer.halo.radius < 2:
            labelob.outlinecolor.setHex(symbolizer.halo.fill.color)
               
    # label placement
    pp = symbolizer.placement
    
    if not pp.anchorpoint:
        labelob.position = mapscript.MS_AUTO
    else:
        p = pp.anchorpoint
        vert = 'C'
        if p.y < 0.25: vert = 'L'
        elif p.y > 0.75: vert = 'U'
        horiz = 'C'
        if p.x < 0.25: horiz = 'L'
        elif p.x > 0.75: horiz = 'R'
        labelob.position = eval('mapscript.MS_%s%s' % (vert, horiz)) 
      
    if not pp.rotation:
        labelob.autoangle = mapscript.MS_TRUE
    else:
        labelob.angle = pp.rotation
        
    # offset
    if pp.displacement:
        d = pp.displacement
        labelob.offsetx = long(d.x)
        labelob.offsety = long(d.y)
            
    # scale thresholds
    if minscale > 0:
        classob.minscale = minscale
    if maxscale > 0:
        classob.maxscale = maxscale

    labelob.minfeaturesize = int(symbolizer.minfeaturesize)
    
    labelob.autominfeaturesize = eval('mapscript.MS_%s' \
                     % (str(bool(symbolizer.autominfeaturesize)).upper()))
    labelob.partials = eval('mapscript.MS_%s' \
                     % (str(bool(symbolizer.partials)).upper()))
    labelob.force = eval('mapscript.MS_%s' \
                     % (str(bool(symbolizer.force)).upper()))

    labelob.mindistance = int(pp.mindupedistance)
    labelob.buffer = int(pp.padding)
    
def _stylize(styleob, classob, layerob, mapob, symbolizer=None,
             minscale=-1, maxscale=-1):
    """
    Applies symbolization from 'symbolizer' to the mapscript styleob,
    layerob, and mapob.
    """
    if not symbolizer:
        _stylize_default(styleob, layerob)
    else:
        if ISLDLineSymbolizer.providedBy(symbolizer):
            s = symbolizer.stroke 
            # set the line brush
            if s.dasharray:
                dash = mapscript.symbolObj(str(symbolizer))
                dash.type = mapscript.MS_SYMBOL_ELLIPSE
                dash.filled = mapscript.MS_TRUE
                line = mapscript.lineObj()
                line.add(mapscript.pointObj(1,1))
                dash.setPoints(line)
                dash.sizex = 1
                dash.sizey = 1
                for i in range(len(s.dasharray)):
                    dash.setStyle(i, s.dasharray[i])
                dash.stylelength = len(s.dasharray)
                mapob.symbolset.appendSymbol(dash)
                styleob.symbol = mapob.symbolset.index(str(symbolizer))
            else:
                styleob.symbol = 1
            styleob.color.setHex(symbolizer.stroke.color)
            styleob.size = int(symbolizer.stroke.width)
            # mapserver's has per-layer transparency, wrong
            layerob.transparency = int(symbolizer.stroke.opacity*100)
        elif ISLDRasterSymbolizer.providedBy(symbolizer):
            layerob.transparency = int(symbolizer.opacity*100)
        elif ISLDPolygonSymbolizer.providedBy(symbolizer):
            f = symbolizer.fill
            s = symbolizer.stroke 
 
            # Luiz patch for ticket #31
            try:
                styleob.width = int(s.width)
            except:
                pass
                
            if f:
                styleob.color.setHex(f.color)
                layerob.transparency = int(f.opacity*100)
            try:
                styleob.outlinecolor.setHex(s.color)
                layerob.transparency = int(s.opacity*100)
            except:
                raise
                
        elif ISLDPointSymbolizer.providedBy(symbolizer):
            # TODO: remove these lines after updating tests
            #if not symbolizer.mark and not symbolizer.graphic:
            #    styleob.color.setRGB(123,123,123)
            #    styleob.outlinecolor.setRGB(0,0,0)
            #    layerob.transparency = int(symbolizer.fill.opacity*100)
            m = symbolizer.graphic.mark
            g = symbolizer.graphic.externalgraphic
            if m:
                f = m.fill
                s = m.stroke
                styleob.color.setHex(f.color)
                styleob.outlinecolor.setHex(s.color)
                styleob.symbol = mapob.symbolset.index(m.name)
                layerob.transparency = int(m.fill.opacity*100)
            elif g:
                gr_data = cStringIO.StringIO(g.getdata())
                sym = mapscript.symbolObj(str(symbolizer))
                sym.type = mapscript.MS_SYMBOL_PIXMAP
                sym_img = mapscript.imageObj(gr_data)
                sym.setImage(sym_img)
                sym_i = mapob.symbolset.appendSymbol(sym)
                styleob.symbol = sym_i
                layerob.transparency = mapscript.MS_GD_ALPHA

            styleob.size = int(symbolizer.graphic.size)
    
    # scale thresholds
    if minscale > 0:
        classob.minscale = minscale
    if maxscale > 0:
        classob.maxscale = maxscale

def _stylize_poly_stroke(styleob, classob, layerob, mapob, symbolizer=None,
                         minscale=-1, maxscale=-1):
    s = symbolizer.stroke 
    # set the line brush
    if s.dasharray:
        dash = mapscript.symbolObj(str(symbolizer))
        dash.type = mapscript.MS_SYMBOL_ELLIPSE
        dash.filled = mapscript.MS_TRUE
        line = mapscript.lineObj()
        line.add(mapscript.pointObj(1,1))
        dash.setPoints(line)
        dash.sizex = 1
        dash.sizey = 1
        for i in range(len(s.dasharray)):
            dash.setStyle(i, s.dasharray[i])
        dash.stylelength = len(s.dasharray)
        mapob.symbolset.appendSymbol(dash)
        styleob.symbol = mapob.symbolset.index(str(symbolizer))
    else:
        styleob.symbol = 1
    styleob.outlinecolor.setHex(s.color)
    styleob.size = int(s.width)
    # Luiz patch for ticket #31
    try:
        styleob.width = int(s.width)
    except:
        pass

    # mapserver's has per-layer transparency, wrong
    layerob.transparency = int(s.opacity*100)
    # scale thresholds
    if minscale > 0:
        classob.minscale = minscale
    if maxscale > 0:
        classob.maxscale = maxscale

def _stylize_poly_fill(styleob, classob, layerob, mapob, symbolizer=None,
                    minscale=-1, maxscale=-1):
    f = symbolizer.fill
    if f:
        styleob.color.setHex(f.color)
        layerob.transparency = int(f.opacity*100)
    # scale thresholds
    if minscale > 0:
        classob.minscale = minscale
    if maxscale > 0:
        classob.maxscale = maxscale

def _stylize_default(styleob, layerob):
    """Applies default symbolization to styleob and layerob in the absence
    of any symbolization.
    """
    if layerob.type == mapscript.MS_LAYER_LINE:
        symbolizer = LineSymbolizer()
        s = symbolizer.stroke
        styleob.symbol = 1
        styleob.color.setHex(s.color)
        styleob.size = int(s.width)
        # mapserver's has per-layer transparency, wrong
        layerob.transparency = int(s.opacity*100)
    elif layerob.type == mapscript.MS_LAYER_POLYGON:
        symbolizer = PolygonSymbolizer()
        f = symbolizer.fill
        s = symbolizer.stroke 
        styleob.color.setHex(f.color)
        styleob.outlinecolor.setHex(s.color)
        layerob.transparency = int(f.opacity*100)
    elif layerob.type == mapscript.MS_LAYER_POINT:
        symbolizer = PointSymbolizer()
        styleob.symbol = 1
        styleob.size = int(symbolizer.size)
        styleob.color.setHex(symbolizer.fill.color)
        styleob.outlinecolor.setHex(symbolizer.stroke.color)
      
_operator_map = [[' EQ ', ' == '], [' = ', ' == '], [' eq ', ' == '], # EQ
                 [' LT ', ' < '], [' lt ', ' < '],                    # LT
                 [' LE ', ' <= '], [' le ', ' <= '],                  # LE
                 [' GT ', ' > '], [' gt ', ' > '],                    # GT
                 [' GE ', ' >= '], [' ge ', ' >= '],                  # GE
                 [' AND ', ' and '], [' OR ', ' or ']]

def evaluateMSLogical(filter, fields, values):
    """Evaluate a MapServer logical expression, returns True or False"""
    expression = filter
    # replace attribute names with values
    for f, v in zip(fields, values):
        if f[1] == 'string':
            from_string = '"[%s]"' % (f[0])
            to_string = '"%s"' % (v)
        elif f[1] == 'int':
            from_string = '[%s]' % (f[0])
            to_string = '%d' % (v)
        elif f[1] == 'float':
            from_string = '[%s]' % (f[0])
            to_string = '%f' % (v)
        else:
            raise Exception, "error in _evaluate_filter(): %s" % (f[1])
        expression = expression.replace(from_string, to_string)
    # replace operators
    for target, replacement in _operator_map:
        expression = expression.replace(target, replacement)
        
    return eval(expression)

_reverse_operator_map = [[' EQ ', ' == '], # EQ
                 [' LT ', ' < '],                    # LT
                 [' LE ', ' <= '],                  # LE
                 [' GT ', ' > '],                    # GT
                 [' GE ', ' >= '],                  # GE
                 [' AND ', ' and '], [' OR ', ' or ']]

PATTERN_ATTREC_ATT = re.compile(r'f\.(\w+)')
PATTERN_ATTREC_KEY = re.compile(r'f\[\s*["\'](\w+)["\']\s*\]')
PATTERN_ATTREC_STR = re.compile(r'str\((.+?)\)')

def convertFilterToMSLogical(filter, fields=None):
    """Convert the python expression to a MapServer logical expression"""
    # convert every occurrance of f.FOO|f['FOO']|f["FOO"] to [FOO]
    # and str(f.FOO|f['FOO']|f["FOO"]) to "[FOO]"
    expression = filter
    # first pass, do the attributes
    expression, n = re.subn(PATTERN_ATTREC_ATT, r'[\1]', expression)
    # next the keys
    expression, n = re.subn(PATTERN_ATTREC_KEY, r'["\1"]', expression)
    # then str operations
    expression, n = re.subn(PATTERN_ATTREC_STR, r'"\1"', expression)
    # replace operators
    for replacement, target in _reverse_operator_map:
        expression = expression.replace(target, replacement)
    return '(%s)' % (expression)
   
PATTERN_LABELSCALE = re.compile(r'(and\s+)*g\.SIZE\s*(\S+)\s*(\S+)')
PATTERN_WKTRINGSEP = re.compile('\)\s*,\s*\(')

def _insert_feature(shapeob, layerob, classindex, text):
    shapeob.classindex = classindex
    shapeob.text = text
    layerob.addFeature(shapeob)

def _init_map(renderer, size, view, format, bgcolor, transparent):
    mo = mapscript.mapObj(renderer.mapfile)
    mo.setSize(size[0], size[1])
    try:
        mo.setExtent(view.ul.x, view.lr.y, view.lr.x, view.ul.y)
    except:
        raise Exception, "invalid map extents: " + str((view.ul.x, view.lr.y, view.lr.x, view.ul.y))
        
    mo.imagecolor.setHex(bgcolor)

    # set projection from View's srs
    _set_projection(mo, view.srs)
    mo.units = eval('mapscript.MS_%s' % (view.units.upper()))

    if renderer.fontset:
        mo.setFontSet(renderer.fontset)
    if renderer.incoming:
        mo.web.imagepath = renderer.incoming

    # Build-in circle and square symbols
    _buildInMarks(mo)
    
    _setOutputFormat(mo, format, transparent)

    return mo

def _get_backend_mapserver_layerdef(path, typename):
    """Return a 3-tuple of (connectiontype, connection, data)"""
    if not path:
        raise Exception, "A filesystem path was not specified"
    if not typename:
        raise Exception, "A feature typename was not specified"
    # if our path is a directory
    if os.path.isdir(path):
        dirname = path
        pattern = '%s.shp' % (os.path.join(dirname, typename))
        results = glob.glob(pattern)
        if results:
            return (
                mapscript.MS_SHAPEFILE,
                None, 
                os.path.abspath(results[0])
                )
        else:
            raise Exception, \
            "Misconfigured data: unable to find typename %s in directory %s" \
            % (typename, path)
    # a file
    else:
        ds = ogr.Open(path)
        driver_name = ds.GetDriver().GetName()
        ds.Destroy()
        if driver_name == 'VRT':
            f = open(path, 'r')
            e_vrtstore = fromstring(f.read())
            f.close()
            iter = e_vrtstore.getiterator('OGRVRTLayer')
            for e_vrtsource in iter:
                if e_vrtsource.attrib['name'] == typename:
                    e_src = e_vrtsource.find('SrcDataSource')
                    path2 = e_src.text
                    
                    # Look for the OGR-ODBC case
                    if path2.find('ODBC') == 0:
                        return (
                            mapscript.MS_OGR,
                            path,
                            typename
                            )
                    
                    # the shapefile case
                    elif e_src.attrib['relativeToVRT'] == '1':
                        base = os.path.dirname(path)
                        path2 = os.path.join(base, path2)
                    typename2 = e_vrtsource.find('SrcLayer').text
                    if os.path.isdir(path2):
                        dirname = path2
                        pattern = '%s.[shp|SHP]' \
                                % (os.path.join(dirname, typename2))
                        results = glob.glob(pattern)
                        if results:
                            return (
                                mapscript.MS_SHAPEFILE,
                                None,
                                os.path.abspath(results[0])
                                )
                        else:
                            raise Exception, \
                            "Misconfigured data: unable to find typename %s in directory %s" \
                            % (typename2, path2)
                    else:
                        return (
                            mapscript.MS_OGR,
                            path,
                            typename
                            )
            # don't get here
            raise Exception, \
            "Misconfigured data: unable to find typename %s in VRT %s" \
            % (typename, path)
        if driver_name == 'ESRI Shapefile':
            return (
                mapscript.MS_SHAPEFILE,
                None,
                os.path.abspath(path)
                )
        else:
            raise Exception, \
            "Misconfigured data: %s is unsupported data type %s" \
            % (path, driver_name)

def _munge_layers(layer, style, format=None):
    # New style generation of one or more MapServer layers from a PCL layer
    # and style.
    tags = []
    layers = []
    for rule in style.rules:
        nsymbs = len(rule.symbolizer)
        for i in range(nsymbs):
            s = rule.symbolizer[i]
            # Text symbolizers are a special case. we merge them with the
            # previous tag. If there is no previous tag, it becomes an
            # annotation layer.
            if ISLDTextSymbolizer.providedBy(s):
                if i == 0:
                    # MapServer annotation layer
                    t = [layer, style, s.__class__, s.geometry]
                    c = [[rule, [], s]]
                    tags.append(t)
                    layers.append(c)
                else:
                    # Becomes the label of a layer
                    prev = rule.symbolizer[i-1]
                    tprev = [layer, style, prev.__class__, prev.geometry]
                    l = layers[tags.index(tprev)]
                    # maybe as a new style of an existing class
                    to_new = 1
                    for c in l:
                        if c[0].filter == rule.filter:
                            c[2] = s
                            to_new = 0
                    # or a new class
                    if to_new:
                        l.append([rule, [], s])
            else:
                t = [layer, style, s.__class__, s.geometry]
                if t not in tags:
                    # New MapServer layer
                    tags.append(t)
                    l = [[rule, [s], None]]
                    layers.append(l) 
                else:
                    # Into an existing layer
                    l = layers[tags.index(t)]
                    # maybe as a new style of an existing class
                    to_new = 1
                    for c in l:
                        if c[0].filter == rule.filter:
                            c[1].append(s)
                            to_new = 0
                    # or a new class
                    if to_new:
                        l.append([rule, [s], None])
    
    # Reverse rulings since MapServer's classes have the opposite 
    # priority.
    for l in layers:
        l.reverse()

    return zip(tags, layers)
    
def _initialize_layerObj(ltree, mo):
    # Create and initialize a MapServer layerObj from the provided layer tree.
    # ltree is a 2-tuple: first item is a collection of PCL layer, symbolizer
    # type, and geometry, the second item is a list of layer classes. See
    # LayerInitTest in PCL-MapServer/tests/testrendering.py for details on the
    # structure of the layer tree. The simplest possible example is:
    #
    # ( [Layer (instance), Symbolizer (class), geometry name (string)]
    #   [
    #       [rule name (string), filter (string),
    #        [Symbolizer (instance)], TextSymbolizer (instance)],
    #   ],
    # )

    layer = ltree[0][0]
    datastore = layer.datastore
   
    # Create layerObj and set common parameters
    lo = mapscript.layerObj()
    lo.name = str(layer)
    lo.connectiontype = -1
    lo.maxfeatures = layer.maxfeatures or -1
    
    # Set layerObj type, data, connection parameters?
    if IFeatureStore.providedBy(datastore):
        symbclass = ltree[0][2]
        if ISLDPointSymbolizer.implementedBy(symbclass):
            lo.type = mapscript.MS_LAYER_POINT
        elif ISLDLineSymbolizer.implementedBy(symbclass):
            lo.type = mapscript.MS_LAYER_LINE
        elif ISLDPolygonSymbolizer.implementedBy(symbclass):
            lo.type = mapscript.MS_LAYER_POLYGON
        elif ISLDTextSymbolizer.implementedBy(symbclass):
            lo.type = mapscript.MS_LAYER_ANNOTATION

        #if layer.classifier = 1:
        # is a PostGIS layer?
        if hasattr(datastore, 'dsn'):
            ft = datastore.featuretype(layer.typename)
            geom_name = ltree[0][3] or ft.defaultgeometry().name
            _set_projection(lo, ft.defaultsrs())
            lo.connectiontype = mapscript.MS_POSTGIS
            lo.connection = datastore.dsn
            if layer.filter:
                lo.filter = layer.filter 
            if hasattr(ft, '__srid__'):
                lo.data = '%s FROM (%s) AS foo USING UNIQUE %s USING SRID=%d' \
                    % (geom_name, ft.feature_sql('gid'),
                       'gid', ft.__srid__)
            else:
                lo.data = '%s FROM %s USING UNIQUE %s' \
                    % (geom_name, layer.typename, 'gid')
                
        elif hasattr(datastore, 'path'):
            ft = datastore.featuretype(layer.typename)
            _set_projection(lo, ft.defaultsrs())
            ldef = _get_backend_mapserver_layerdef(datastore.path,
                                                   layer.typename)
            lo.connectiontype = ldef[0]
            lo.connection = ldef[1]
            lo.data = ldef[2]
            if layer.filter:
                lo.filter = layer.filter
 
        elif hasattr(datastore, 'url'):
            mo.web.imagepath = mo.web.imagepath or datastore.incoming
            # access datastore through IOWSDescriptor
            lo.connection = datastore._wfs.capabilities.getOperationByName('{http://www.opengis.net/wfs}GetFeature').methods['{http://www.opengis.net/wfs}Get']['url']
            lo.connectiontype = mapscript.MS_WFS
            lo.metadata.set('wfs_version', datastore.version)
            lo.metadata.set('wfs_typename', layer.typename)
            lo.metadata.set('wfs_service', 'WFS')
            lo.metadata.set('wfs_srs', layer.ows_srs)
            lo.metadata.set('wfs_connectiontimeout',
                    str(getattr(layer, 'timeout', 30)))
        else:
            # in-memory feature store
            ft = datastore.featuretype(layer.typename)
            srs = ft.defaultsrs()
            try:
                if srs == -1:
                    # Image coordinates
                    lo.transform = mapscript.MS_FALSE
            except:
                _set_projection(lo, srs)

    else:
        if hasattr(datastore, 'path'):
            lo.data = datastore.path
            lo.type = mapscript.MS_LAYER_RASTER
            bounds = datastore.bounds(layer.typename)
            _set_projection(lo, bounds.srs)

        elif hasattr(datastore, 'url'):
            mo.web.imagepath = mo.web.imagepath or datastore.incoming
            # access datastore through IOWSDescriptor
            lo.type = mapscript.MS_LAYER_RASTER
            lo.connection = datastore._wms.capabilities.getOperationByName('GetMap').methods['Get']['url']
            lo.connectiontype = mapscript.MS_WMS
            lname = layer.typename.split('/')[0]
            bounds = datastore.bounds(lname)
            _set_projection(lo, bounds.srs)
       
            lo.metadata.set('wms_server_version', datastore.version)
            lo.metadata.set('wms_srs', layer.ows_srs)
            # for format, try ows_format, fall back on the map format
            lo.metadata.set('wms_format', 
                        getattr(layer, 'ows_format', '') or format)

            # connection timeout
            lo.metadata.set('wms_connectiontimeout',
                    str(getattr(layer, 'timeout', 30)))
        
            # names and styles
            styled_layers = layer.typename.split(',')
            layer_names = []
            style_names = []
            for o in styled_layers:
                if o.find('/') < 0:
                    o = o + '/'
                layer_name, style_name = o.split('/')
                layer_names.append(layer_name)
                style_names.append(style_name)
            comma = ','
            layers = comma.join(layer_names)
            styles = comma.join(style_names)
            lo.metadata.set('wms_name', layers)
            lo.metadata.set('wms_style', styles)
    
            # Work around for mapserver 4.4 bug
            e = mo.extent
            lo.setExtent(e.minx, e.miny, e.maxx, e.maxy)
        else:
            raise Exception, "Invalid raster data store"

    # Create classes and styles
    # --------------------------
    for ruling in ltree[1]:
   
        rule = ruling[0]
        symbolizers = ruling[1]
        label = ruling[2]
        
        ci = lo.insertClass(mapscript.classObj())
        co = lo.getClass(ci)
        co.name = ruling[0].name

        if not symbolizers and label:
            # Annotation only
            _labelize(co, lo, label,
                rule.minscaledenominator, rule.maxscaledenominator)
        elif symbolizers:
            for symbolizer in symbolizers:
                if ISLDPolygonSymbolizer.providedBy(symbolizer):
                    si = co.insertStyle(mapscript.styleObj())
                    so = co.getStyle(si)
                    _stylize_poly_fill(so, co, lo, mo, symbolizer,
                        rule.minscaledenominator, rule.maxscaledenominator)
                    si = co.insertStyle(mapscript.styleObj())
                    so = co.getStyle(si)
                    _stylize_poly_stroke(so, co, lo, mo, symbolizer,
                        rule.minscaledenominator, rule.maxscaledenominator)
                else:
                    si = co.insertStyle(mapscript.styleObj())
                    so = co.getStyle(si)
                    _stylize(so, co, lo, mo, symbolizer,
                        rule.minscaledenominator, rule.maxscaledenominator)
            if label:
                _labelize(co, lo, label,
                    rule.minscaledenominator, rule.maxscaledenominator)
        else:
            si = co.insertStyle(mapscript.styleObj())
            so = co.getStyle(si)
            _stylize_default(so, lo)

    #lo.status = mapscript.MS_ON
    return lo

def _get_active_rule_ids(ruling, scale=1):
    # Compare cartographic scale to rule scale thresholds, returning a 
    # sequence of rules that are active
    rr = []
    for r in ruling:
        rule = r[0]
        if scale >= rule.minscaledenominator and \
            (scale <= rule.maxscaledenominator or rule.maxscaledenominator < 0):
            rr.append(str(rule))
    return rr #[str(rule) for rule in rules]

def _render_layer_extrinsic(im, mo, lo, layer, style, ruling, classifications,
    geom=None):
    """Render features from a store with extrinsic classification."""
    ft = layer.datastore.featuretype(layer.typename)
    gkey = ft.defaultgeometry().name
    fnames = ft.schema().names()

    # Back-end filtering - *always* do a bounding box filter
    # if the data store supports it.
    # ------------------------------------------------------
    # transform map extents to srs of the source
    e = mapscript.rectObj(mo.extent.minx, mo.extent.miny,
                          mo.extent.maxx, mo.extent.maxy)
    try:
        e.project(mapscript.projectionObj(mo.getProjection()),
                  mapscript.projectionObj(lo.getProjection()))
    except MapServerError, exc:
        if str(exc).find("msProcessProjection") >= 0: pass
        else: raise

    box = BoundingBox(e.minx, e.miny, e.maxx, e.maxy)
    
    # postgis
    if lo.connectiontype == mapscript.MS_POSTGIS:
        #fstr = "%s && 'BOX3D(%f %f, %f %f)'::box3d" \
        #     % (gkey, e.minx, e.miny, e.maxx, e.maxy)
        #query = Query(filter=SQLFilter(fstr))
        source = layer.datastore.source(layer.typename, bbox=box)
    # shapefile/ogr
    elif lo.connectiontype in [mapscript.MS_SHAPEFILE, mapscript.MS_OGR]:
        # TODO: spatial filter for disk feature source
        source = layer.datastore.source(layer.typename) #, bbox=box)
    # in-memory
    else:
        #import pdb; pdb.set_trace()
        source = layer.datastore.source(layer.typename,
            maxfeatures=layer.maxfeatures)

    # index of label property
    labelindex = -1
    if lo.labelitem:
        labelindex = lo.labelitem
        lo.labelitem = None
    
    # determine the rules that are active at the current map scale
    rr = _get_active_rule_ids(ruling, mo.scale)
    
    for feature in source():
        f = feature.properties
        if geom: g = f[geom]
        else: g = f[gkey]

        # Points in image coordinates
        if lo.transform == mapscript.MS_FALSE:
            if not IPoint.providedBy(g):
                raise TypeError, "Only points are supported in image coords"
            if g.srs == 'top-left':
                pass
            elif g.srs == 'bottom-left':
                raise NotImplemented, "bottom-left origin not implemented"
            elif g.srs == 'bottom-right':
                #import pdb; pdb.set_trace()
                #g.x = mo.width + g.x
                #g.y = mo.height + g.y
                g = Point(mo.width + g.x, mo.height + g.y)
        
        shape = msbridge.shapeObj(lo.type, g)
        #shape.transformed = 0
            
        # initialized number of rules
        num_rules = 0
           
        id = feature.id
        for rule_id in rr:
            filter, ci = classifications[rule_id]
            if eval(filter):
                if labelindex >= 0:
                    if isinstance(f[labelindex], unicode):
                        text = f[labelindex].encode('utf-8')
                    else:
                        text = str(f[labelindex])
                else:
                    text = None
                _insert_feature(shape, lo, ci, text)
                       
                if style.maxrules:
                    num_rules += 1
                    if num_rules >= style.maxrules:
                        break
            else:
                # insert no label or feature
                pass
            
    lo.connectiontype = mapscript.MS_INLINE
    lo.draw(mo, im)

