# $Id: testrendering.py,v 1.3 2005/04/20 19:49:22 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import os
import unittest
import mapscript

from framework import TESTPATH, FONTSET
from framework import cartography

from cartography import mapping, spatial, rendering, query
from cartography.data import disk, memory, ows

DISKSTORE = os.path.abspath(os.path.join(TESTPATH, 'diskstore.ovf'))

class MapServerRendererTestCase(unittest.TestCase):
    """testing cartography.mapping.MapServerRenderer"""
    def setUp(self):
        self.ul = spatial.Point(-10.0, 10.0)
        self.lr = spatial.Point(10.0, -10.0)
        self.srs = spatial.SpatialReference(epsg=4326)
        self.view = mapping.View(self.srs, self.ul, self.lr)
        
        # store
        self.path = DISKSTORE
        self.d_store = disk.DiskFeatureStore(self.path)
        # layers
        self.x_layer = mapping.Layer(self.d_store, 'point')
        self.l_layer = mapping.Layer(self.d_store, 'line')
        self.p_layer = mapping.Layer(self.d_store, 'polygon')
        
        # -------------------------------------------------------------------
        # Intrinsic Style
        # -------------------------------------------------------------------
        # a style for the polygon layer with two rules
        symb = mapping.PolygonSymbolizer(fill={'fill': '#FF8080'})
        text = mapping.TextSymbolizer('FNAME', {'font-family': 'VeraBd'})
        text2 = mapping.TextSymbolizer('FNAME', {'font-family': 'Vera', 
                                                 'font-size': 6.0}) 
        self.p_style = mapping.Style('p_style')
        # this rule is aimed at one particular feature
        f1 = query.PythonFilter('str(f.FNAME) == "A Polygon"')
        self.p_style.rules.append(mapping.Rule('named', symb, f1))
        # for same feature with a Text Symbolizer
        self.p_style.rules.append(mapping.Rule('named', text, f1))
        # this rule, with no filter is the catch-all 
        self.p_style.rules.append(mapping.Rule('catch-all', 
                                               mapping.PolygonSymbolizer()))
        self.p_style.rules.append(mapping.Rule('catch-all', text2))

        # ------------------------------------------------------------------
        # Extrinsic Style -- Rules are evaluated in Python, not by the
        # mapping engine
        self.pe_style = mapping.ExtrinsicStyle('pe_style')
        # this rule is aimed at one particular feature
        f2 = query.PythonFilter('f.FNAME.find("A P") == 0')
        f2b = query.PythonFilter('f.FNAME.find("A P") != 0')
        self.pe_style.rules.append(mapping.Rule('named', text, f2))
        self.pe_style.rules.append(mapping.Rule('named', symb, f2))
        # this rule, with no filter is the catch-all 
        self.pe_style.rules.append(mapping.Rule('catch-all', text2, f2b))
        self.pe_style.rules.append(mapping.Rule('catch-all', 
                                               mapping.PolygonSymbolizer(), f2b))
        self.pe_style.rules.append(mapping.Rule('false', 
                                               mapping.PolygonSymbolizer(),
                                               query.PythonFilter('False')))

        # --------------------------------------------------------------------
        # inline data layer
        geoms = [spatial.Polygon(wkt='POLYGON ((-1.0 50.5, -0.5 51.2, 0.3 50.9, -1 50.5))'),
                 spatial.Polygon(wkt='POLYGON((-0.7 50.3, 0.1 51.0, 0.6 50.1, -0.7 50.3))')]
        atts = [('A Polygon',), ('B Polygon',)]
        source = memory.MemoryFeatureSource('inline', [['FNAME', 'string']], self.srs)
        for g,a in zip(geoms, atts):
            source.add(g,a)
        self.w_store = memory.MemoryFeatureStore()
        self.w_store['inline'] = source
        self.w_layer = mapping.Layer(self.w_store, 'inline')

        # --------------------------------------------------------------------
        # raster layer

        self.r_path = os.path.join(TESTPATH, 'raster.vrt')
        self.r_store = disk.DiskRasterStore(self.r_path)
        self.r_layer = mapping.Layer(self.r_store, 'raster')

        # --------------------------------------------------------------------
        # WMS raster layer

        self.wms_url = "http://wms.jpl.nasa.gov/wms.cgi?"
        self.wms_store = ows.WMSRasterStore(self.wms_url, '1.1.1',
                                            incoming='/var/www-tmp/')
        self.wms_layer = mapping.Layer(self.wms_store,
                                       'global_mosaic_base/visual')
        # adapt
        self.wms_layer.ows_srs = 'EPSG:4326'
        self.wms_layer.ows_format = 'image/jpeg'

        # WFS
        u = 'http://zcologia.com:9001/mapserver/members/features.rpy'
        self.wfs_store = ows.WFSFeatureStore(u, '1.0.0',
                                             incoming='/var/www-tmp/')
        self.wfs_layer = mapping.Layer(self.wfs_store, 'users')
        self.wfs_layer.ows_srs = 'EPSG:4326'

        # Point symbolization
        circle = mapping.Mark('square')
        fn = mapping.TextSymbolizer('FNAME', {'font-family': 'VeraBd'})
        wfs_sym = mapping.PointSymbolizer(10, mark=circle)
        # Text symbolization
        wfs_text = mapping.TextSymbolizer('zco:fullname',{'font-family': 'VeraBd'})
        self.wfs_style = mapping.Style('wfs_style')
        self.wfs_style.rules.append(mapping.Rule('catch-all', wfs_sym))
        self.wfs_style.rules.append(mapping.Rule('catch-all', wfs_text))

        # Diagonal to test Jean-Denis's auto angle patch
        line = spatial.LineString(wkt='LINESTRING (-1.0 50.5, -0.5 51.2, 0.3 50.9)')
        atts = ('A Line',)
        source = memory.MemoryFeatureSource('inline', [['FNAME', 'string']], self.srs)
        source.add(line, atts)
        self.memline_store = memory.MemoryFeatureStore()
        self.memline_store['inline'] = source
        self.memline_layer = mapping.Layer(self.memline_store, 'inline')

        # --------------------------------------------------------------------
        # Map renderer
        self.mapper = rendering.MapRenderer('MAPSERVER', fontset=FONTSET,
                                          incoming='/var/www-tmp/')

