# $Id: testdata.py,v 1.8 2005/04/07 14:59:56 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import os
import sys
import unittest
import mapscript

from framework import cartography, TESTPATH

from cartography import spatial
from cartography.query import Query
from cartography.query import PythonFilter

from cartography.engine.mapserver.data import MapserverFilterGeometry
from cartography.engine.mapserver.rendering import _get_backend_mapserver_layerdef
from cartography.data.disk import DiskFeatureStore
from cartography.referencing.srs import SpatialReference

DISKSTORE = os.path.join(os.path.abspath(TESTPATH), 'diskstore.ovf')

# ============================================================================
# Utility functions
# ============================================================================

def pointEquality(p1, p2):
    """Return True if points have the same x/y, False otherwise"""
    if p1.x == p2.x and p1.y == p2.y:
        return True
    else:
        return False


# ============================================================================
# Tests
# ============================================================================

class GetFeaturesTestCase:
    
    """base class for testing feature access"""
   
    def testPropertiesSubset(self):
        # try FNAME, the second attribute
        props = ['FNAME']
        query = Query(props, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 1, len(features)
        assert features[0][1] == ['A Polygon'], features
        
        # Next, FID, the first attribute
        props = ['FID']
        query = Query(props, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 1
        assert features[0][1] == [1], features

    def testFalse(self):
        pyfilt = PythonFilter('False')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 0, len(features)

    def testTrue(self):
        pyfilt = PythonFilter('True')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        fiter = self.features(query)
        features = [f for f in self.features(query)]
        assert len(features) == 1, len(features)

    # attribute filtering

    def testAttrFalse(self):
        """expect 0 features when filter is globally False"""
        pyfilt = PythonFilter('f.FNAME.find("A B") == 0')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 0

    def testString(self):
        pyfilt = PythonFilter('f.FNAME.find("A P") == 0')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 1, len(features)

    def testNumeric(self):
        pyfilt = PythonFilter('f.FID*2 > 1.5')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 1
        assert isinstance(features[0][0], mapscript.shapeObj), features[0][0]

    def testFalseNumeric(self):
        pyfilt = PythonFilter('f.FID*2 > 3.5')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 0

    # bbox filtering
    
    def testBBOX(self):
        pyfilt = PythonFilter('g.BBOX(-20,25,60,75)')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 1

    def testBBOXDisjoint(self):
        """expect no features within a -120,25,-40,75 bounding box"""
        pyfilt = PythonFilter('g.BBOX(-120,25,-40,75)')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 0, features

    # combo filtering
    
    def testBBOXCombo(self):
        pyfilt = PythonFilter('f.FID*2 > 1.5 and g.BBOX(-20,25,60,75)')
        query = Query(filter=pyfilt, outputformat='MAPSERVER')
        features = [f for f in self.features(query)]
        assert len(features) == 1

    # spatial filtering

    def testPointIntersection(self):
        subject_p = spatial.Point(0.0, 51.5)
        pyfilt = PythonFilter('g.intersects(p)')
        query = Query(filter=pyfilt, outputformat='MAPSERVER', p=subject_p)
        features = [f for f in self.features(query)]
        assert len(features) == 1, len(features)
    
    def testPointNonIntersection(self):
        subject_p = spatial.Point(-10.0, 51.5)
        pyfilt = PythonFilter('g.intersects(p)')
        query = Query(filter=pyfilt, outputformat='MAPSERVER', p=subject_p)
        features = [f for f in self.features(query)]
        assert len(features) == 0, len(features)

    def testPointIntersectionCombo(self):
        subject_p = spatial.Point(0.0, 51.5)
        pyfilt = PythonFilter('f.FID*2 > 1.5 and g.intersects(p)')
        query = Query(filter=pyfilt, outputformat='MAPSERVER', p=subject_p)
        features = [f for f in self.features(query)]
        assert len(features) == 1, len(features)


class MapserverDiskFeaturesTestCase(GetFeaturesTestCase, unittest.TestCase):
    """testing getting features from a Shapefile FeatureStore"""
    def setUp(self):
        self.path = DISKSTORE
        self.srs = SpatialReference(epsg=4326)
        self.store = DiskFeatureStore(self.path)
        self.features = self.store.source('polygon').features

class MapserverDiskIterFeaturesTestCase(GetFeaturesTestCase, unittest.TestCase):
    """testing feature iterator from a Shapefile FeatureStore"""
    def setUp(self):
        self.path = DISKSTORE
        self.srs = SpatialReference(epsg=4326)
        self.store = DiskFeatureStore(self.path)
        self.features = self.store.source('polygon').iterfeatures

       
class MapserverFilterGeometryTestCase(unittest.TestCase):

    def setUp(self):
        ro = mapscript.rectObj(-1, -1, 1, 1)
        so = ro.toPolygon()
        self.fg = MapserverFilterGeometry(so)

    def testInstance(self):
        s = str(self.fg.f_geom.bounds)
        assert s == "{ 'minx': -1 , 'miny': -1 , 'maxx': 1 , 'maxy': 1 }", s

    def testIntersection(self):
        wkt = 'POLYGON ((-2 -2, -2 2, 2 2, 2 -2, -2 -2))'
        polygon = spatial.Polygon(wkt=wkt)
        assert self.fg.intersects(polygon)


LINE_SHP = os.path.abspath(os.path.join(TESTPATH, 'line.shp'))

class GetMapServerPathTest(unittest.TestCase):

    def setUp(self):
        f = open('test_template.ovf', 'r')
        xml = f.read()
        f.close()
        xml = xml.replace('TESTDATA_PATH', LINE_SHP)
        f = open('test.ovf', 'w')
        f.write(xml)
        f.close()
        
    def testOVF(self):
        linepath = _get_backend_mapserver_layerdef('test.ovf', 'line')[2]
        self.assert_(linepath == LINE_SHP, (LINE_SHP, linepath))
       
    def testShapefile(self):
        linepath = _get_backend_mapserver_layerdef(LINE_SHP, 'line')[2]
        self.assert_(linepath == LINE_SHP, (LINE_SHP, linepath))
        
    def testDirectory(self):
        linepath = _get_backend_mapserver_layerdef(os.path.abspath(TESTPATH), 'line')[2]
        self.assert_(linepath == LINE_SHP, (LINE_SHP, linepath))


# =============================================================================
if __name__ == '__main__':
    unittest.main()

