# $Id: testmapping.py,v 1.1 2005/03/29 22:13:47 sgillies Exp $

# =============================================================================
# Python Cartographic Library. Copyright (C) 2004 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import os
import unittest
import mapscript

from framework import cartography, TESTPATH, FONTSET
from cartography import data, spatial
from cartography.mapping import LabelPlacement, TextSymbolizer
from cartography.engine.mapserver import rendering

class EvalExpressionTestCase(unittest.TestCase):

    def testConvertFilterAtts(self):
        fields = [('FID', 'int'), ('FNAME', 'string')]
        values = (1, 'Foo')
        filter = 'str(f.FNAME) == "Foo" or str(f.FNAME) == "Bar" and f.FID > 1'
        ms_expr = rendering.convertFilterToMSLogical(filter, fields)
        assert ms_expr == '("[FNAME]" EQ "Foo" OR "[FNAME]" EQ "Bar" AND [FID] GT 1)', ms_expr

    def testConvertFilterDicts(self):
        fields = [('FID', 'int'), ('FNAME', 'string')]
        values = (1, 'Foo')
        filter = 'str(f["FNAME"]) == "Foo" or str(f["FNAME"]) == "Bar" and f["FID"] > 1'
        ms_expr = rendering.convertFilterToMSLogical(filter, fields)
        assert ms_expr == '("[FNAME]" EQ "Foo" OR "[FNAME]" EQ "Bar" AND [FID] GT 1)', ms_expr
        
    def testConvertFilterMixed(self):
        fields = [('FID', 'int'), ('FNAME', 'string')]
        values = (1, 'Foo')
        filter = 'str(f["FNAME"]) == "Foo" or str(f.FNAME) == "Bar" and f.FID > 1'
        ms_expr = rendering.convertFilterToMSLogical(filter, fields)
        assert ms_expr == '("[FNAME]" EQ "Foo" OR "[FNAME]" EQ "Bar" AND [FID] GT 1)', ms_expr

    def testSimpleStringExpression(self):
        fields = [('FID', 'int'), ('FNAME', 'string')]
        values = (1, 'Foo')
        filter1 = '("[FNAME]" EQ "Foo")'
        filter2 = '("[FNAME]" EQ "Bar")'
        filter3 = '("[BOGUS]" EQ "Foo")'
        assert rendering.evaluateMSLogical(filter1, fields, values) == True
        assert rendering.evaluateMSLogical(filter2, fields, values) == False
        assert rendering.evaluateMSLogical(filter3, fields, values) == False
   
    def testCompoundStringExpression(self):
        fields = [('FID', 'int'), ('FNAME', 'string'), ('FNICK', 'string')]
        values = (1, 'Foo', 'F')
        x1, r1 = '("[FNAME]" EQ "Foo") AND ("[FNICK]" EQ "F")', True
        x2, r2 = '("[FNAME]" EQ "Bar") OR ("[FNICK]" EQ "F")', True
        x3, r3 = '("[BOGUS]" EQ "Foo") AND ("[FNICK]" EQ "F")', False
        x4, r4 = '("[FNAME]" = "Foo") AND ("[FNICK]" = "F")', True
        x5, r5 = '("[FNAME]" eq "Bar") OR ("[FNICK]" = "F")', True
        x6, r6 = '("[BOGUS]" = "Foo") and ("[FNICK]" eq "F")', False
        assert rendering.evaluateMSLogical(x1, fields, values) == r1
        assert rendering.evaluateMSLogical(x2, fields, values) == r2
        assert rendering.evaluateMSLogical(x3, fields, values) == r3
        assert rendering.evaluateMSLogical(x4, fields, values) == r4
        assert rendering.evaluateMSLogical(x5, fields, values) == r5
        assert rendering.evaluateMSLogical(x6, fields, values) == r6
  
    def testSimpleNumericExpression(self):
        fields = [('FID', 'int'), ('FNAME', 'string')]
        values = (1, 'Foo')
        x1, r1 = '([FID] EQ 1)', True
        x2, r2 = '([FID] eq 1)', True
        x3, r3 = '([FID] = 1)', True
        x4, r4 = '([FID] > 0)', True
        x5, r5 = '([FID] > 1)', False
        x7, r7 = '("[FID]" = 1)', False
        assert rendering.evaluateMSLogical(x1, fields, values) == r1
        assert rendering.evaluateMSLogical(x2, fields, values) == r2
        assert rendering.evaluateMSLogical(x3, fields, values) == r3
        assert rendering.evaluateMSLogical(x4, fields, values) == r4
        assert rendering.evaluateMSLogical(x5, fields, values) == r5
        assert rendering.evaluateMSLogical(x7, fields, values) == r7

    def testCompoundNumericExpression(self):
        fields = [('FID', 'int'), ('WEIGHT', 'float'), ('FNAME', 'string')]
        values = (1, 12.0, 'Foo')
        x1, r1 = '([FID] EQ 1 AND [WEIGHT] > 10)', True
        x2, r2 = '([FID] eq 1 AND [WEIGHT] lt 20)', True
        x3, r3 = '([FID] = 2 OR [WEIGHT] <= 12.01)', True
        x5, r5 = '([FID] > 1 and [WEIGHT] < 100)', False
        assert rendering.evaluateMSLogical(x1, fields, values) == r1
        assert rendering.evaluateMSLogical(x2, fields, values) == r2
        assert rendering.evaluateMSLogical(x3, fields, values) == r3
        assert rendering.evaluateMSLogical(x5, fields, values) == r5

    def testCompoundMixedExpression(self):
        fields = [('FID', 'int'), ('WEIGHT', 'float'), ('FNAME', 'string')]
        values = (1, 12.0, 'Foo')
        x1, r1 = '([FID] EQ 1 AND "[FNAME]" = "Foo")', True
        x2, r2 = '("[FNAME]" eq "Bar" OR [WEIGHT] lt 20)', True
        assert rendering.evaluateMSLogical(x1, fields, values) == r1
        assert rendering.evaluateMSLogical(x2, fields, values) == r2
        
class MapServerTestCase(unittest.TestCase):

    def setUp(self):
        self.mo = mapscript.mapObj()
        li = self.mo.insertLayer(mapscript.layerObj())
        self.lo = self.mo.getLayer(li)
        ci = self.lo.insertClass(mapscript.classObj())
        self.co = self.lo.getClass(ci)
        si = self.co.insertStyle(mapscript.styleObj())
        self.so = self.co.getStyle(si)

class StylingTestCase(MapServerTestCase):
    pass
    
class LabelingTestCase(MapServerTestCase):

    def testLabelPlacementLL(self):
        """test setting label to LL"""
        from mapscript import MS_LL
        a = spatial.Point(0.0, 0.0)
        pos = rendering.LabelPlacement(a)
        text = rendering.TextSymbolizer(placement={'anchorpoint': a})
        rendering._labelize(self.co, self.lo, text)
        assert self.co.label.position == MS_LL

    def testLabelPlacementUR(self):
        """test setting label to UR"""
        from mapscript import MS_UR
        a = spatial.Point(0.8, 0.8)
        pos = rendering.LabelPlacement(a)
        text = rendering.TextSymbolizer(placement={'anchorpoint': a})
        rendering._labelize(self.co, self.lo, text)
        assert self.co.label.position == MS_UR


class LabelingTestCase(MapServerTestCase):

    def testLabelPlacementLL(self):
        """test setting label to LL"""
        from mapscript import MS_LL
        a = spatial.Point(0.0, 0.0)
        pos = LabelPlacement(a)
        text = TextSymbolizer(placement={'anchorpoint': a})
        rendering._labelize(self.co, self.lo, text)
        assert self.co.label.position == MS_LL

    def testLabelPlacementUR(self):
        """test setting label to UR"""
        from mapscript import MS_UR
        a = spatial.Point(0.8, 0.8)
        pos = LabelPlacement(a)
        text = TextSymbolizer(placement={'anchorpoint': a})
        rendering._labelize(self.co, self.lo, text)
        assert self.co.label.position == MS_UR


# ===========================================================================
if __name__ == '__main__':
    unittest.main()

