# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import sys
import cStringIO
from traceback import print_tb
import unittest

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
    pkg_resources.require('PCL-GDAL')
    pkg_resources.require('PCL-MapServer')
except:
    pass

from cartography.proj import SpatialReference
from cartography.geometry import Point, Geometry
from cartography.data.feature import Feature, PropertyTypeCollection
from cartography.data.feature import StringProperty, GeometryProperty
from cartography.data.memory import MemoryFeatureStore
from cartography.data.disk import DiskFeatureStore, DiskRasterStore
from cartography.data.ows import WMSRasterStore, WFSFeatureStore
from cartography.styles import *
from cartography.context import View, MapRenderer, MapContext, ContextLayer

from _testing import *
from cartography.engine.mapserver import rendering


class MapInternalTest(unittest.TestCase):
    def setUp(self):
        self.mapper = MapRenderer('MapServer', fontset=FONTSET)
    def test_map_in(self):
        # Replicate the MapServer test

        # raster
        store = DiskFeatureStore(TILEDSTORE)
        layer = Layer(store, 'tileindex')
        style = Style()
        style.rules.append(Rule('r', [RasterSymbolizer()]))
        raster = ContextLayer('raster', layer, style)

        # polygon
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'polygon')
        style = Style()
        rule = Rule('r', [PolygonSymbolizer()])
        style.rules.append(rule)
        polygon = ContextLayer('polygon', layer, style)

        # line
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style()
        rule = Rule('r', [LineSymbolizer()])
        style.rules.append(rule)
        line = ContextLayer('line', layer, style)

        # point
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'point')
        style = Style()
        graphic = Graphic(os.path.join(MS_FIXTURES, 'home.png'), size=16.0)
        rule = Rule('r', [PointSymbolizer(graphic), TextSymbolizer('FNAME', Font('VeraBd'))])
        style.rules.append(rule)
        point = ContextLayer('point', layer, style)

        map = MapContext([raster, polygon, line, point])
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-1, 50, 1, 52])
        size = [200, 200]
        image = self.mapper.render(map, view,
                                   'image/png', size, bgcolor='#EEEEFF',
                                   transparent=0, debug=True)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('test_map_in.png', 'wb')
        f.write(image.getdata())
        f.close()
        
class MapExternalTest(unittest.TestCase):
    def setUp(self):
        self.mapper = MapRenderer('MapServer', fontset=FONTSET, 
            incoming='/tmp/')
    def test_map_ex(self):
        # raster
        store = DiskFeatureStore(TILEDSTORE)
        layer = Layer(store, 'tileindex')
        style = Style()
        style.rules.append(Rule('r', [RasterSymbolizer()]))
        raster = ContextLayer('raster', layer, style)

        # WMS
        #url = "http://wms.jpl.nasa.gov/wms.cgi?"
        #store = WMSRasterStore(url, '1.1.1')
        #layer = Layer(store, 'global_mosaic_base/visual')
        #layer.ows_srs = 'EPSG:4326'
        #layer.ows_format = 'image/jpeg'
        #style = Style()
        #style.rules.append(Rule('r', [RasterSymbolizer()]))
        #wms = ContextLayer('wms', layer, style)

        # polygon
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'polygon', classifier=1)
        style = Style()
        rule = Rule('r', [PolygonSymbolizer()])
        style.rules.append(rule)
        polygon = ContextLayer('polygon', layer, style)

        # line
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line', classifier=1)
        style = Style()
        rule = Rule('r', [LineSymbolizer()])
        style.rules.append(rule)
        line = ContextLayer('line', layer, style)

        # point
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'point', classifier=1)
        style = Style()
        graphic = Graphic(os.path.join(MS_FIXTURES, 'home.png'), size=16.0)
        rule = Rule('r', [
                PointSymbolizer(graphic),
                TextSymbolizer('FNAME', Font('VeraBd'),
                    placement=Placement(anchorpoint=(0, -1))
                )
            ])
        style.rules.append(rule)
        point = ContextLayer('point', layer, style)

        # WFS
        store = WFSFeatureStore('http://zcologia.com:9001/mapserver/members/capabilities.rpy', version='1.0.0')
        layer = Layer(store, 'users', classifier=1)
        layer.ows_srs = 'EPSG:4326'
        style = Style()
        graphic = Graphic(mark=Mark('square'), size=10.0)
        rule = Rule('r', [PointSymbolizer(graphic), TextSymbolizer('zco:mid', Font('Vera'), halo=Halo(fill=Fill('#FFFFFF'), radius=1.0))])
        style.rules.append(rule)
        wfs = ContextLayer('point', layer, style)

        #map = MapContext([wms, polygon, line, point, wfs])
        map = MapContext([raster, polygon, line, point, wfs])
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-1, 50, 1, 52])
        size = [200, 200]
        image = self.mapper.render(map, view, 'image/jpeg', size, 
            transparent=0, debug=True)
        assert image.size == size
        assert image.mimetype == 'image/jpeg'
        f = open('test_map_ex.jpg', 'wb')
        f.write(image.getdata())
        f.close()        
   
class WorldMapTest(unittest.TestCase):
    def setUp(self):
        self.mapper = MapRenderer('MapServer', fontset=FONTSET,
            incoming='/tmp/')
    def test_world(self):
        # WMS
        url = "http://wms.jpl.nasa.gov/wms.cgi?"
        store = WMSRasterStore(url, '1.1.1')
        layer = Layer(store, 'modis')
        layer.ows_srs = 'EPSG:4326'
        layer.ows_format = 'image/jpeg'
        style = Style()
        style.rules.append(Rule('r', [RasterSymbolizer()]))
        wms = ContextLayer('wms', layer, style)

        # polygon
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'world_borders', classifier=1)
        style = Style()
        border = PolygonSymbolizer(stroke=Stroke('#FFFFFF', width=2, 
            opacity=0.5))
        border.fill = None
        rule = Rule('UK', [border], "str(f.FIPS_CNTRY) == 'UK'")
        style.rules.append(rule)
        world = ContextLayer('world', layer, style)

        map = MapContext([wms, world])
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-10, 30, 30, 70])
        size = [300, 300]
        image = self.mapper.render(map, view, 'image/jpeg', size,  
            bgcolor='#000000', transparent=0, debug=True)
        assert image.size == size
        assert image.mimetype == 'image/jpeg'
        f = open('test_world.jpg', 'wb')
        f.write(image.getdata())
        f.close()        

class MultiSymbolizerTest(unittest.TestCase):
    def setUp(self):
        self.mapper = MapRenderer('MapServer', fontset=FONTSET)
        class F(Feature):
            __schema__      = PropertyTypeCollection([
                                StringProperty('name'),
                                GeometryProperty('the_geom', homotype='Point'),
                                GeometryProperty('region', homotype='Polygon')
                                ])
            __typename__    = 'F'
            __srs__         = SpatialReference(epsg=4326)
        store = MemoryFeatureStore(F)
        sink = store.source('F')
        sink.addfeature(F('1', name='Point A', the_geom=Point(-0.5, 50.5),
            region=Point(-0.5, 50.5).buffer(0.5, 24)))
        sink.addfeature(F('2', name='Point B', the_geom=Point(0.5, 51.5),
            region=Point(0.5, 51.5).buffer(0.5, 24)))
        # Add a third feature which we'll eliminate by using layer.maxfeatures
        sink.addfeature(F('3', name='Point C', the_geom=Point(0.6, 51.4),
            region=Point(0.6, 51.4).buffer(0.5, 24)))
        self.layer = Layer(store, 'F', maxfeatures=2)
    def test_error(self):
        square_default = Mark('square')
        sym = PointSymbolizer(Graphic(mark=square_default, size=33.0),
            geometry='bogus')
        rule = Rule('c', [sym])
        style = Style('style')
        style.rules.append(rule)
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-1, 50, 1, 52])
        size = [200, 200]
        map = MapContext([ContextLayer('a', self.layer, style)])
        image = self.mapper.render(map, view,
                                   'image/png', size, bgcolor='#EEEEFF',
                                   transparent=0, debug=True)
        self.assertEqual(image.size, size)
        self.assertEqual(image.mimetype, 'image/png')
        self.assert_(isinstance(image.error, KeyError))
        self.assertEqual(str(image.error), "'bogus'")
        f = open('test_error.png', 'wb')
        f.write(image.getdata())
        f.close()
    def test_squares_and_text(self):
        # Two point symbolizers, same geometry
        square_default = Mark('square')
        square_white = Mark('square', fill=Fill('#FFFFFF'))
        sym_lo = PointSymbolizer(Graphic(mark=square_default, size=33.0))
        sym_hi = PointSymbolizer(Graphic(mark=square_white, size=25.0))
        text = TextSymbolizer('name', Font('VeraBd'))
        rule = Rule('c', [sym_lo, sym_hi, text])
        style = Style('style')
        style.rules.append(rule)
        layers = rendering._munge_layers(self.layer, style, 'image/png')
        self.assertEqual(len(layers), 1)
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-1, 50, 1, 52])
        size = [200, 200]
        map = MapContext([ContextLayer('a', self.layer, style)])
        image = self.mapper.render(map, view,
                                   'image/png', size, bgcolor='#EEEEFF',
                                   transparent=0, debug=True)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('test_squares_and_text.png', 'wb')
        f.write(image.getdata())
        f.close()
    def test_multi_geom(self):
        # polygon and point symbolizers, drawing on different geoms
        square_default = Mark('square')
        square_white = Mark('square', fill=Fill('#FFFFFF'))
        sym_lo = PolygonSymbolizer(fill=Fill('#00FF00'),
            stroke=Stroke('#009900'), geometry='region')
        sym_hi = PointSymbolizer(Graphic(mark=square_white, size=25.0))
        text = TextSymbolizer('name', Font('VeraBd'))
        rule = Rule('c', [sym_lo, sym_hi, text])
        style = Style('style')
        style.rules.append(rule)
        layers = rendering._munge_layers(self.layer, style, 'image/png')
        self.assertEqual(len(layers), 2)
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-1, 50, 1, 52])
        size = [200, 200]
        map = MapContext([ContextLayer('a', self.layer, style)])
        image = self.mapper.render(map, view,
                                   'image/png', size, bgcolor='#EEEEFF',
                                   transparent=0, debug=True)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('test_multi_geom.png', 'wb')
        f.write(image.getdata())
        f.close()

    def test_multi_geom_rules(self):
        # polygon and point symbolizers, drawing on different geoms
        square_default = Mark('square')
        square_white = Mark('square', fill=Fill('#FFFFFF'))
        sym_gr = PolygonSymbolizer(fill=Fill('#00FF00'),
            stroke=Stroke('#009900'), geometry='region')
        sym_bl = PolygonSymbolizer(fill=Fill('#0000FF'),
            stroke=Stroke('#000099'), geometry='region')
        sym_hi = PointSymbolizer(Graphic(mark=square_white, size=25.0))
        text = TextSymbolizer('name', Font('VeraBd'))
        ra = Rule('a', [sym_gr, sym_hi, text], "f.name.endswith('A')")
        rb = Rule('b', [sym_bl, sym_hi, text], "f.name.endswith('B')")
        style = Style('style')
        style.rules += [ra, rb]
        layers = rendering._munge_layers(self.layer, style, 'image/png')
        self.assertEqual(len(layers), 2)
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-1, 50, 1, 52])
        size = [200, 200]
        map = MapContext([ContextLayer('a', self.layer, style)])
        image = self.mapper.render(map, view,
                                   'image/png', size, bgcolor='#EEEEFF',
                                   transparent=0, debug=True)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('test_multi_geom_rules.png', 'wb')
        f.write(image.getdata())
        f.close()

class DrawSymbolizerTestCase(unittest.TestCase):
    def setUp(self):
        self.mapper = MapRenderer('MapServer', fontset=FONTSET)
    def testDrawMarkerPointSymbolizer(self):
        """draw a point symbolizer image"""
        circle = Mark('circle')
        sym = PointSymbolizer(Graphic(mark=circle, size=8))
        size = [16, 16]
        image = self.mapper.draw(sym, 'image/png', size)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('testDrawMarkerPointSymbolizer.png', 'wb')
        f.write(image.getdata())
        f.close()
    def testCrazyMarkSymbolizer(self):
        circle = Mark('circle')
        sym = PointSymbolizer(Graphic(mark=circle, size=12))
        sym.graphic.mark.stroke.width = 3
        sym.graphic.mark.fill.color = '#FF0000'
        sym.graphic.mark.fill.opacity = 0.5
        size = [16, 16]
        image = self.mapper.draw(sym, 'image/png', size)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('testCrazyMarkSymbolizer.png', 'wb')
        f.write(image.getdata())
        f.close()
    def testDrawLineSymbolizer(self):
        """draw a line symbolizer image"""
        s = Stroke()
        s.color = '#80FF80'
        s.width = 5
        sym = LineSymbolizer(s)
        size = [48, 48]
        image = self.mapper.draw(sym, 'image/png', size)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('testDrawLineSymbolizer.png', 'wb')
        f.write(image.getdata())
        f.close()
    def testDrawPolygonSymbolizer(self):
        """draw a polygon symbolizer image"""
        sym = PolygonSymbolizer(fill=Fill('#FF8080'))
        size = [48, 48]
        image = self.mapper.draw(sym, 'image/png', size)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('testDrawPolygonSymbolizer.png', 'wb')
        f.write(image.getdata())
        f.close()
    def testDrawTextSymbolizer(self):
        """draw a polygon symbolizer image"""
        text = TextSymbolizer('FNAME', Font('VeraBd'))
        size = [48, 48]
        image = self.mapper.draw(text, 'image/png', size)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('testDrawTextSymbolizer.png', 'wb')
        f.write(image.getdata())
        f.close()
    def test_symb_multi(self):
        wide = LineSymbolizer(stroke=Stroke('#CCCC99', width=7))
        thin = LineSymbolizer()
        text = TextSymbolizer('FNAME', Font('Vera'))
        size = [48, 48]
        image = self.mapper.draw([wide, thin, text], 'image/png', size, transparent=True)
        assert image.size == size
        assert image.mimetype == 'image/png'
        f = open('test_symb_multi', 'wb')
        f.write(image.getdata())
        f.close()
       
class DashTest(unittest.TestCase):
    def setUp(self):
        self.mapper = MapRenderer('MapServer', fontset=FONTSET)
        class F(Feature):
            __schema__      = PropertyTypeCollection([
                                StringProperty('name'),
                                GeometryProperty('region', homotype='Polygon')
                                ])
            __typename__    = 'F'
            __srs__         = SpatialReference(epsg=4326)
        store = MemoryFeatureStore(F)
        sink = store.source('F')
        sink.addfeature(F('1', name='Region A', 
            region=Geometry.fromWKT('POLYGON ((-0.5 50.5, -0.5 51.5, 0.5 51.5, 0.5 50.5, -0.5 50.5))')))
        self.layer = Layer(store, 'F')
    def test_dashed_line(self):
        stroke = Stroke(width=3, dasharray=[9, 3, 3, 3])
        sym = LineSymbolizer(stroke=stroke)
        rule = Rule('c', [sym])
        style = Style('style')
        style.rules.append(rule)
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-1, 50, 1, 52])
        size = [200, 200]
        map = MapContext([ContextLayer('a', self.layer, style)])
        image = self.mapper.render(map, view,
                                   'image/png', size, bgcolor='#EEEEFF',
                                   transparent=0, debug=True)
        self.assertEqual(image.size, size)
        self.assertEqual(image.mimetype, 'image/png')
        f = open('test_dashed.png', 'wb')
        f.write(image.getdata())
        f.close()
    def test_dashed_poly(self):
        stroke = Stroke(width=3, dasharray=[9, 3, 3, 3])
        sym = PolygonSymbolizer(stroke=stroke)
        rule = Rule('c', [sym])
        style = Style('style')
        style.rules.append(rule)
        srs = SpatialReference(epsg=4326)
        view = View(srs, [-1, 50, 1, 52])
        size = [200, 200]
        map = MapContext([ContextLayer('a', self.layer, style)])
        image = self.mapper.render(map, view,
                                   'image/png', size, bgcolor='#EEEEFF',
                                   transparent=0, debug=True)
        self.assertEqual(image.size, size)
        self.assertEqual(image.mimetype, 'image/png')
        f = open('test_dashed_poly.png', 'wb')
        f.write(image.getdata())
        f.close()
        
class ImageModeTestCase(unittest.TestCase):
    def setUp(self):
        self.mapper = MapRenderer('MapServer', fontset=FONTSET)
    def testDrawPNG24MarkerPointSymbolizer(self):
        circle = Mark('circle')
        sym = PointSymbolizer(Graphic(mark=circle, size=8))
        size = [16, 16]
        image = self.mapper.draw(sym, 'image/png; mode=RGB; interlace=OFF',size)
        self.assertEqual(image.size, size)
        self.assertEqual(image.mimetype, 'image/png')
        f = open('testDrawPNG24.png', 'wb')
        f.write(image.getdata())
        f.close()
    def testDrawBogusMarkerPointSymbolizer(self):
        """draw a point symbolizer image"""
        circle = Mark('circle')
        sym = PointSymbolizer(Graphic(mark=circle, size=8))
        size = [16, 16]
        self.assertRaises(Exception, self.mapper.draw,
                          sym, 'image/gif; mode=RGB', size)
        

if __name__ == '__main__':
    unittest.main()

