# =============================================================================
# Python Cartographic Library. Copyright (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

import unittest

try:
    import pkg_resources
    pkg_resources.require('PCL-Core')
    pkg_resources.require('PCL-GDAL')
    pkg_resources.require('PCL-MapServer')
except:
    pass

from cartography.proj import SpatialReference
from cartography.geometry import Point
from cartography.data.feature import Feature, PropertyTypeCollection
from cartography.data.feature import StringProperty, GeometryProperty
from cartography.data.memory import MemoryFeatureStore
from cartography.data.disk import DiskFeatureStore
from cartography.styles import Mark, PointSymbolizer, TextSymbolizer, \
    LineSymbolizer, Rule, Style, ExtrinsicStyle, Layer, \
    Fill, Font, Stroke, Graphic, PolygonSymbolizer
from cartography.context import View, MapRenderer, MapContext, ContextLayer

from _testing import *
from cartography.engine.mapserver import rendering

class LayerInitTest(unittest.TestCase):
    def test_easiest(self):
        # We expect one rule with one symbolizer to result in one MapServer
        # layer of one class with one symbolizer.
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        line_symb = LineSymbolizer()
        line_all = Rule('r0', [line_symb])
        style.rules.append(line_all)
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 1)
        self.assertEqual(layers, 
            [
            (   [layer, style, line_symb.__class__, None], # layer tags 
                [                                       # layers
                    [line_all, [line_symb], None],    # one class
                ],
            )
            ])
    def test_annotation(self):
        # We expect one rule with one symbolizer to result in one MapServer
        # annotation layer.
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        text_symb = TextSymbolizer('FNAME')
        text_all = Rule('r0', [text_symb])
        style.rules.append(text_all)
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 1)
        self.assertEqual(layers, 
            [
            (   [layer, style, text_symb.__class__, None], # layer tags 
                [                                   # layers
                    [text_all, [], text_symb],    # one class
                ],
            )
            ])
    def test_easylabel(self):
        # One rule with line and text symbolizer should result in one MapServer
        # layer with a single style, labeled, class.
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        line_symb = LineSymbolizer()
        text_symb = TextSymbolizer('FNAME')
        line_all = Rule('line', [line_symb, text_symb])
        style.rules.append(line_all)
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 1)
        self.assertEqual(layers, 
            [
            (   [layer, style, line_symb.__class__, None],
                [
                    [line_all, [line_symb], text_symb], # one class
                ],
            )
            ])
    def test_easiest_double_symb(self):
        # We expect one rule with two symbolizers to result in one MapServer
        # layer of one class with two styles.
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        line_symb = LineSymbolizer()
        line_all = Rule('r0', [line_symb, line_symb])
        style.rules.append(line_all)
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 1)
        self.assertEqual(layers, 
            [
            (   [layer, style, line_symb.__class__, None],
                [ 
                    [line_all, [line_symb, line_symb], None], # one class
                ],
            )
            ])
    def test_easiest_double_symb_label(self):
        # We expect one rule with two line symbolizers, and a text symbolizer
        # to result in one MapServer layer of one class with two styles and a
        # label
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        point_symb = PointSymbolizer()
        text_symb = TextSymbolizer(font=Font('VeraBd'), label='x')
        line_all = Rule('r0', [point_symb, point_symb, text_symb])
        style.rules.append(line_all)
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 1)
        self.assertEqual(layers, 
            [
            (   [layer, style, point_symb.__class__, None],
                [ 
                    [line_all, [point_symb, point_symb], text_symb], # one class
                ],
            )
            ])
    def test_layer_split_on_type(self):
        # We expect one rule with two differently typed symbolizers to result
        # in two MapServer layers of one class with one style.
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        line_symb = LineSymbolizer()
        point_symb = PointSymbolizer()
        line_all = Rule('r0', [line_symb, point_symb])
        style.rules.append(line_all)
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 2)
        self.assertEqual(layers, 
            [
            (   [layer, style, line_symb.__class__, None],
                [  
                    [line_all, [line_symb], None],
                ],
            ),
            (   [layer, style, point_symb.__class__, None],
                [ 
                    [line_all, [point_symb], None],
                ],
            )
            ])
    def test_layer_split_on_geom(self):
        # We expect one rule with two different symbolizer geometries to result
        # in two MapServer layers of one class with one style.
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        line_symb = LineSymbolizer(geometry='road')
        center_symb = LineSymbolizer(geometry='centerline')
        line_all = Rule('r0', [line_symb, center_symb])
        style.rules.append(line_all)
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 2)
        self.assertEqual(layers, 
            [
            (   [layer, style, line_symb.__class__, 'road'],
                [
                    [line_all, [line_symb], None],
                ],
            ),
            (   [layer, style, center_symb.__class__, 'centerline'],
                [
                    [line_all, [center_symb], None],
                ],
            )
            ])
    def test_easier(self):
        # We expect two rules with one symbolizer each to result in one
        # MapServer layer of two classes each with one symbolizer.
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        line_symb = LineSymbolizer()
        r0 = Rule('r0', [line_symb], 'is')
        r1 = Rule('r1', [line_symb], 'is not')
        style.rules += [r0, r1]
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 1)
        self.assertEqual(layers, 
            [
            (   [layer, style, line_symb.__class__, None],
                [ 
                    [r1, [line_symb], None],    # class 0
                    [r0, [line_symb], None],    # class 1
                ],
            )
            ])
    def test_hard(self):
        # We expect this mix of rules to result in two MapServer layers
        store = DiskFeatureStore(DISKSTORE)
        layer = Layer(store, 'line')
        style = Style('s')
        line_symb = LineSymbolizer()
        text_symb = TextSymbolizer('FNAME')
        r0 = Rule('r0', [line_symb, text_symb], 'is')
        r1 = Rule('r1', [line_symb], 'is not')
        r2 = Rule('r2', [text_symb], 'annotate')
        style.rules += [r0, r1, r2]
        layers = rendering._munge_layers(layer, style, 'image/png')
        self.assertEqual(len(layers), 2)
        self.assertEqual(layers, 
            [
            (   [layer, style, line_symb.__class__, None],
                [ 
                    [r1, [line_symb], None],    # class 1
                    [r0, [line_symb], text_symb],    # class 0
                ],
            ),
            (   [layer, style, text_symb.__class__, None],
                [  
                    [r2, [], text_symb],    # class 1
                ],
            )
            ])


# ============================================================================
if __name__ == '__main__':
    unittest.main()

