
try:
    # Try to do it first with setuptools
    import pkg_resources
    pkg_resources.require('PCL-Core')
    pkg_resources.require('PCL-MapServer')
except:
    # Fall back on the location specified in PYTHONPATH
    pass

from cartography.proj import SpatialReference
from cartography.geometry import Geometry, Point, LinearRing, Polygon
from cartography.data.feature import Feature, PropertyTypeCollection, \
    StringProperty, GeometryProperty
from cartography.data.memory import MemoryFeatureStore
from cartography.context import MapRenderer, View, MapContext, ContextLayer
from cartography.styles import TextSymbolizer, PolygonSymbolizer, \
    PointSymbolizer
from cartography.styles import Layer, Style, Rule, Font, Fill, Stroke, \
    Mark, Graphic, Halo

# ============================================================================
# Geometries
#
# from cartography.spatial.geometry import Geometry, Point, LinearRing, Polygon
#
# Create two triangular geometries, one from a well-known text string, the
# other from points. Next, get their intersection and union as two new 
# geometries.
# ============================================================================

wkt = 'POLYGON ((-1.0 50.5, -0.5 51.2, 0.3 50.9, -1 50.5))'
tri_a = Geometry.fromWKT(wkt)

shell = LinearRing([Point(-0.7, 50.3), Point(0.1, 51.0), Point(0.6, 50.1),
                    Point(-0.7, 50.3)])
tri_b = Polygon(shell)

# the intersection and union of these geoms
inter = tri_a.intersection(tri_b)
union = tri_a.union(tri_b)

# ===========================================================================
# Features
#
# from cartography.data.feature import Feature, PT_STRING, PT_GEOMETRY
# from cartography.data.memory import MemoryFeatureStore
#
# Create a new feature type and a store to hold instances of the new type.
# Open a 'sink' (inverse of 'source') on the store, wrap the geometries in
# new feature instances, and add them.
# ===========================================================================

EPSG4326 = SpatialReference(epsg=4326)

class ToyFeature(Feature):
    # Three property types in the schema: a string name, a homogeneous
    # geometry "the_geom", every instance of which is a Polygon, and a
    # "center" geometry.
    __schema__ = PropertyTypeCollection([
                    StringProperty('name'),
                    GeometryProperty('the_geom', homotype='Polygon'),
                    GeometryProperty('center'),
                    ])
    __typename__ = 'ToyFeature'
    __srs__ = EPSG4326


# Create a feature store based on the above feature type
store = MemoryFeatureStore(ToyFeature)

# Make and add new features. The first argument is a feature id, subsequent
# keyword arguments correspond to the elements of the feature type schema
source = store.source('ToyFeature')
source.addfeature(ToyFeature('tf:1', name='A Polygon', the_geom=tri_a,
    center=tri_a.centroid()))
source.addfeature(ToyFeature('tf:2', name='B Polygon', the_geom=tri_b,
    center=tri_b.centroid()))
source.addfeature(ToyFeature('tf:3', name='Intersection', the_geom=inter,
    center=inter.centroid()))
source.addfeature(ToyFeature('tf:4', name='Union', the_geom=union,
    center=union.centroid()))


# ===========================================================================
# Symbolization, Style, and Rendering
#
# from cartography.rendering import MapRenderer, View
# from cartography.rendering import MapContext, ContextLayer
# from cartography.mapping import TextSymbolizer, PolygonSymbolizer
# from cartography.mapping import PointSymbolizer
# from cartography.mapping import Layer, Style, Rule
# from cartography.mapping import Font, Halo, Fill, Stroke, Mark, Graphic
# ===========================================================================

# Setup the renderer. Point to a mapserver fontset file.
FONTSET = '../fonts.txt'
mapper = MapRenderer('MAPSERVER', fontset=FONTSET)

# Make a Polygon symbolizer. "the_geom" is the default geometry, and there's
# no need to specify it on the Symbolizer.
fill = Fill('#FF9999')
stroke = Stroke('#CC6666')
poly = PolygonSymbolizer(stroke, fill)

# Symbolize the "center" geometry with a square mark.
square = Graphic(mark=Mark('square'), size=6.0)
point = PointSymbolizer(graphic=square, geometry="center")

# Symbolize the "center" again with a text string, using the "name"
# property of each feature.
text = TextSymbolizer('name', Font('Vera', 10.0), fill=Fill('#333333'),
    geometry="center")


# Create a Layer that draws from the ToyFeature type in the in-memory store
# Wrap in a map context layer, and make a map context
layer = Layer(store, 'ToyFeature')
geoms = ContextLayer('geoms', layer, None)
map = MapContext([geoms])

# Render the initial geometries separately
# Create a rule using the symbolizers above, pop it into a style, and pass
# this style along to the "geoms" map context layer.

filter = "id in ['tf:1', 'tf:2']"
rule = Rule('A and B', [poly, point, text], filter)
style = Style('style')
style.rules.append(rule)
geoms.style = style

# define viewport, dimensions, render, and save
view = View(EPSG4326, [-1.2, 50.0, 1, 51.2])
size = [300, 300]
image = mapper.render(map, view, 'image/png', size, transparent=0)
f = open('separate.png', 'wb')
f.write(image.getdata())
f.close()

# Now, for the intersection
filter = "id in ['tf:3']"
style = Style('style')
style.rules.append(Rule('label', [poly, point, text], filter))
geoms.style = style

view = View(EPSG4326, [-1.2, 50, 1, 51.2])
size = [300, 300]
image = mapper.render(map, view, 'image/png', size, transparent=0)
f = open('intersection.png', 'wb')
f.write(image.getdata())
f.close()

# and union
filter = "id in ['tf:4']"
style = Style('style')
style.rules.append(Rule('label', [poly, point, text], filter))
geoms.style = style

view = View(EPSG4326, [-1.2, 50, 1, 51.2])
size = [300, 300]
image = mapper.render(map, view, 'image/png', size, transparent=0)
f = open('union.png', 'wb')
f.write(image.getdata())
f.close()


