# Example of using PostGIS from PCL's trunk
#
# To run this, you will need the world_borders data loaded into PostGIS 1.x.

try:
    # Try to use setuptools to get all dependencies. This is the most
    # reliable way.
    import pkg_resources
    pkg_resources.require('PCL-Core')
    pkg_resources.require('PCL-MapServer')
except:
    # Fall back on a temp site-packages location, that you specify through
    # the PYTHONPATH environment variable
    pass

from cartography.proj import SpatialReference
from cartography.geometry import Point
from cartography.data.postgis import PGFeatureStore, PGSelectionFeature
from cartography.data.feature import PropertyTypeCollection, \
    IntegerProperty, RealProperty, StringProperty, GeometryProperty
from cartography.context import MapRenderer, View, MapContext, ContextLayer
from cartography.styles import TextSymbolizer, PolygonSymbolizer
from cartography.styles import Layer, Style, Rule
from cartography.styles import Fill, Stroke

# ===========================================================================
# Feature store and a custom feature type
#
# from cartography.data.feature import Feature
# from cartography.data.feature import PT_STRING, PT_INT, PT_FLOAT, PT_GEOMETRY
# from cartography.data.memory import PGFeatureStore, PGSelectionFeature
#
# Create a PostGIS feature store, connected to a PostGIS database. This store
# will know about all tables and views with geometry columns. Next, create
# a new selection-based feature type and add to the store. 
# ===========================================================================

CONNECTION = 'host=localhost dbname=pcl_test user=postgres'
store = PGFeatureStore(CONNECTION)

# user-defined type
class WorldDensity(PGSelectionFeature):
    __typename__ = 'postgis.world_density'
    __schema__ = PropertyTypeCollection([
                    IntegerProperty('gid'),
                    StringProperty('fips_cntry'),
                    RealProperty('density'),
                    GeometryProperty('the_geom')
                    ])
    __sql_from__ = "postgis.world_borders"
    # same order as in the __schema__, very important!
    __sql_select__ = "gid, fips_cntry, (pop_cntry/area) AS density, the_geom"
    __gid__ = "gid"
    __srid__ = 4269
    
store.addfeaturetype(WorldDensity)

# ===========================================================================
# Symbolization, Style, and Rendering
#
# from cartography.rendering import MapRenderer, View
# from cartography.rendering import MapContext, ContextLayer
# from cartography.mapping import TextSymbolizer, PolygonSymbolizer
# from cartography.mapping import Layer, Style, Rule
#
# The "scheme" list below maps population density ranges to colors from 
# an orange ColorBrewer (tm) scheme, and from it we create a number of 
# symbolizers and style rules.
# ===========================================================================

# Setup the renderer. Point to a mapserver fontset file.
FONTSET ='../fonts.txt'
mapper = MapRenderer('MAPSERVER', fontset=FONTSET)

# Set layer to draw from the world_density feature type of the store
layer = Layer(store, 'postgis.world_density')
# uncomment line below for the faster, externally classified layer
#layer = Layer(store, 'postgis.world_density', classifier=1)
style = Style('orange')

# Or, if you prefer, the layer could draw from the custom feature type
#layer = Layer(store, 'postgis.world_density')

# color scheme for symbolizers, based on a computed population density
scheme = [
#    Min    Max     Fill color  Label
#    --------------------------------
    (0.0,   1.0,    '#fff7ec',  '> 0'),
    (1.0,   5.0,    '#fee8c8',  '> 1'),
    (5.0,   10.0,   '#fdd49e',  '> 5'),
    (10.0,  20.0,   '#fdbb84',  '> 10'),
    (20.0,  50.0,   '#fc8d59',  '> 20'),
    (50.0,  100.0,  '#ef6548',  '> 50'),
    (100.0, 200.0,  '#d7301f',  '> 100'),
    (200.0, 500.0,  '#b30000',  '> 200'),
    (500.0, 1e6,    '#7f0000',  '> 500'),
]

scheme.reverse()
for min, max, color, label in scheme:
    fs = "f.density > %f and f.density < %f" % (min, max)
    f = Fill(color=color)
    s = Stroke(color='#505050')
    rule = Rule(label, [PolygonSymbolizer(fill=f, stroke=s)], fs)
    style.rules.append(rule)
   
# Map context
map = MapContext([ContextLayer('borders', layer, style)])

# define viewport, dimensions, render, then save
def render():
    srs = SpatialReference('+proj=moll +lon_0=0.0')
    view = View(srs, [-18100000, -9050000, 18100000, 9050000])
    size = [600, 300]
    image = mapper.render(map, view, 'image/png', size)
    f = open("pg_%s.png" % (style.name), 'wb')
    f.write(image.getdata())
    f.close()

# ============================================================================
# Executing this with a -p option runs the Python profiler

if __name__ == '__main__':
    import sys
    import getopt
    do_profile = 0
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'p')
        if opts[0][0] == '-p':
            do_profile = 1
    except IndexError:
        pass
    except:
        raise

    if do_profile:
        import profile
        profile.run('render()')
    else:
        render()
        
