# =============================================================================
# Python Cartographic Library (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# =============================================================================

"""Check the installations of MapServer, GDAL, and PCL.

Note that the run_check function of this module can be used as a Zope
External Method. This is an ideal way to diagnose problems with your GIS
environment.
"""

import os
import sys

def check(f):
    """Perform a check of GIS environment.

    parameters
    ----------
    f : file
        A Python file-like object.
    """
    
    # =======================================================================
    # Checking cartography.referencing and PROJ.4
    # =======================================================================
    
    referencing_imported = 0
    
    # Try to import cartography.referencing
    f.write("1) Checking import of referencing modules ... ")
    try:
        from cartography import referencing
        from cartography.referencing.srs import SpatialReference
        from cartography.referencing.transform import _proj4
        from cartography.referencing.transform.proj4 import ProjTransform
        f.write("success\n")
        print >>f, "cartography.referencing @ %s" % (referencing.__file__)
        referencing_imported = 1
        
    except ImportError, e:
        f.write("failure\n")
        print >>f, "%s. Was PCL-Referencing installed? If it was installed to a non-standard location (other than '%s'), you will need to set the PYTHONPATH environment variable (current value is '%s') to that location." % (e, sys.path[1], os.environ.get('PYTHONPATH', None))
    
    # Try to transform a point
    f.write("\n2) Checking transformation ... ")
    if not referencing_imported:
        f.write("skipped\n")
        print >>f, "Skipped because modules could not be imported"
    else:
        try:
            utm13 = SpatialReference(epsg=26913)
            wgs84 = SpatialReference(epsg=4326)
            t = ProjTransform(utm13, wgs84)
            coords = [(-105, 39), (-106, 40)]
            new_coords = t.transform(coords)
            f.write("success\n")
            print >>f, "%s, EPSG:4326 -> %s, EPSG:26913" % (coords, new_coords)
        
        except SystemError, e:
            f.write("failure\n")
            print >>f, "%s. The most likely explanation for this failure is a missing or improperly set PROJ_LIB environment variable (current value is '%s'). This variable is required by the PROJ.4 library and should be set to the directory containing the PROJ.4 epsg file." % (e, os.environ.get('PROJ_LIB', None))
    
    # =======================================================================
    # Checking gdal.py, ogr.py, and cartography.data.disk
    # =======================================================================
    
    # Try to import gdal and ogr
    f.write("\n2) Checking import of gdal/ogr/osr ... ")
    try:
        import gdal
        import ogr
        import osr
        f.write("success\n")
        print >>f, "gdal @ %s" % (gdal.__file__)
        print >>f, "ogr @ %s" % (ogr.__file__)
        print >>f, "osr @ %s" % (osr.__file__)
        
    except ImportError, e:
        f.write("failure\n")
        if str(e).find("DLL load failed") > -1 \
        or str(e).find("cannot open shared object") > -1:
            print >>f, "%s The most likely explanation is that the GDAL DLL or SO is not on the binary or library search path. Has GDAL been installed? If you are using FWTools on windows, your PATH environment variable (current value is '%s') should include 'C:\Program Files\FWTools1.0.0\\bin'" % (e, os.environ.get('PATH', None))
        else:
            print >>f, "%s Has GDAL and its Python modules been installed? If it was installed to a non-standard location (other than '%s'), you will need to set the PYTHONPATH environment variable (current value is '%s') to that location. If you are using FWTools on windows, it is likely than PYTHONPATH should be set to 'C:\Program Files\FWTools1.0.0\pymod'." % (e, sys.path[1], os.environ.get('PYTHONPATH', None))
   
def run_check():
    """Wrapper to use as a Zope external method."""
    import StringIO
    f = StringIO.StringIO()
    check(f)
    return f.getvalue()

# Allow script-like execution
if __name__ == "__main__":
    check(sys.stdout)

