#!/bin/env python
#
# $Id: shpconfig.py 462 2006-07-23 01:06:52Z seang $

# ===========================================================================
# Python Cartographic Library (C) 2006 Sean C. Gillies
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#
# Contact email: sgillies@frii.com
# ===========================================================================

# ===========================================================================
# shpconfig.py
#
# Shapefile configuration tool for PCL
#
# Users will commonly find themselves holding shapefile format GIS data that
# defines no spatial referencing. You can check for spatial referencing of a
# shapefile by using the OGR utility 'ogrinfo' (see example 10-1 in "Web
# Mapping Illustrated" or the ogrinfo reference at
# http://ogr.maptools.org/ogrinfo.html)
#
#     $ ogrinfo /data/world/world_borders.shp -al -summary
#
# If an SRS WKT is reported, then this shapefile is spatially referenced and
# can be used directly with PCL/ZCO/PrimaGIS.  If an unknown SRS is reported,
# then file.shp is unreferenced and must be configured before use.
#
# Configuration begins with the creation of a config file, for example:
#
#     $ shpconfig.py -c world.ovf
#
# We are using the OGR virtual format specification from
#
#     http://ogr.maptools.org/drv_vrt.html
#
# for our config files. The benefit is that you can then inspect or manipulate
# this config file with the OGR utility functions.
#
# This world.ovf config file will be the disk-based backend for a ZCO DataStore.
# It currently contains no typenames (what OGR calls "layers"). These are added
# using the tool's -a option, for example:
#
#     $ shpconfig.py -a /data/world/world_borders.shp -t borders \
#                    --a_srs "EPSG:4326" world.ovf
#
# Adds shapefile to world.ovf with the typename "borders" and assigns to this
# the WGS84 spatial reference system known as "EPSG:4326". The --a_srs option
# works exactly like the -a_srs option of "ogr2ogr".
#
# You can use ogrinfo to inspect the config file:
#
#
#     $ ogrinfo world.ovf -al -summary
#     ERROR 4: Update access not supported for VRT datasources.
#     Had to open data source read-only.
#     INFO: Open of `world.ovf'
#     using driver `VRT' successful.
# 
#     Layer name: borders
#     Geometry: Polygon
#     Feature Count: 3784
#     Extent: (-180.000000, -90.000000) - (180.000000, 83.623596)
#     Layer SRS WKT:
#     GEOGCS["WGS 84",
#         DATUM["WGS_1984",
#             SPHEROID["WGS 84",6378137,298.257223563,
#                 AUTHORITY["EPSG","7030"]],
#             AUTHORITY["EPSG","6326"]],
#         PRIMEM["Greenwich",0,
#             AUTHORITY["EPSG","8901"]],
#         UNIT["degree",0.01745329251994328,
#             AUTHORITY["EPSG","9122"]],
#         AUTHORITY["EPSG","4326"]]
#     CAT: Real (16.0)
#     FIPS_CNTRY: String (80.0)
#     CNTRY_NAME: String (80.0)
#     AREA: Real (15.2)
#     POP_CNTRY: Real (15.2)
#
# This config file can be used as the path for a PCL or ZCO DataStore, and 
# offers one feature type, with the typename "borders".
#
# ===========================================================================

import getopt
import glob
import sys
import os
import ogr

# try to find elementtree or lxml
try:
    from cElementTree import Element, SubElement, tostring, fromstring
except:
    try:
        from elementtree.ElementTree import Element, SubElement
        from elementtree.ElementTree import tostring, fromstring
    except:
        from lxml.etree import Element, SubElement, tostring, fromstring
        
def usage():
    print """
PCL Shapefile Configuration Utility

Create a new configuration:
    
  shpconfig -c config-file
      
Add shapefile to configuration

  shpconfig -a path/to/shapefile.shp [-t typename] [--a_srs "srs"] config-file

The typename option overrides a shapefile's default typename (ogrinfo layer
name) and becomes the typename used with ZCO. The a_srs option assigns a
spatial reference system to a typename for unreferenced shapefiles. You can use
ogrinfo to verify a configuration file.
"""

a_srs = None
op = None
shapefile = None
new_typename = None

opts, args = getopt.getopt(sys.argv[1:], 'ca:t:', ['a_srs='])
for o, a in opts:
    if o == '-c':
        op = 'create'
    elif o == '-a':
        op = 'add'
        shapefile = a
    elif o == '-t':
        new_typename = a
    if o in ['--a_srs']:
        a_srs = a
        
try:
    config_file = args[0]
except:
    usage()
    print "! Use Error: Specify a configuration file"
    sys.exit(1)

if op == 'create':
    e_vrtstore = Element('OGRVRTDataSource')
    f = open(config_file, 'w')
    f.write(tostring(e_vrtstore))
    f.close()
    print """
* Created %s
""" % (config_file)
    
elif op == 'add':
    f = open(config_file, 'r+')
    e_vrtstore = fromstring(f.read())
    # get name and srs of srclayer
    ogr_ds = ogr.Open(shapefile)
    ogr_layer = ogr_ds.GetLayer(0)
    typename = ogr_layer.GetName()
    sref = ogr_layer.GetSpatialRef()
    srs = None
    if sref:
        srs = sref.ExportToWkt()
    ogr_ds.Destroy()
    if not srs and not a_srs:
        print """
! Use Error: %s is unreferenced,
  You must assign a spatial reference using the --a_srs option.
""" % (shapefile)
        sys.exit(1)
    e_vrttype = SubElement(e_vrtstore, 'OGRVRTLayer')
    e_vrttype.attrib['name'] = new_typename or typename
    e_srcstore = SubElement(e_vrttype, 'SrcDataSource')
    e_srcstore.attrib['relativeToVRT'] = '0'
    e_srcstore.text = os.path.abspath(shapefile)
    e_srctype = SubElement(e_vrttype, 'SrcLayer')
    e_srctype.text = typename
    if not srs:
        e_srcsrs = SubElement(e_vrttype, 'LayerSRS')
        e_srcsrs.text = a_srs
    f.seek(0)
    f.truncate()
    f.write(tostring(e_vrtstore))
    print """
* Added typename %s from %s
  as typename %s to %s
  with spatial reference %s
""" % (typename, shapefile, new_typename or typename, config_file, srs or a_srs)

else:
    print """
* No operation
"""

    
